// G. Miller, Numerical Analysis for Engineers and Scientists,
// Cambridge University Press, Cambridge, 2014.
// http://www.cambridge.org/9781107021082

#include <math.h>
#include "Matrix.H"
#define MIN(A,B) (((A<B)?(A):(B)))
#define MAX(A,B) (((A>B)?(A):(B)))

// the integrals of factors (x-x_k)*(x-x_(k-1)) etc.
// "gpred" are integrals used for predictor,
// and "gcorr" are integrals for corrector.
// recults include those from recursion, and the trivial cases.

void makeGs( Vector<double>& abscissas,
             Vector<double>& gpred,
             Vector<double>& gcorr )
{
  // assume abscissas is equipped for variable order, so 
  int s = abscissas.size();
  gpred.resize(s+1);
  gcorr.resize(s+1);
  Vector<double> tmp(s);

  // make g1j. assume most recent x is last
  double dx = abscissas[s-2] - abscissas[s-1];
  double t = dx*dx;
  for( int j=1; j<=s; j++ ) {
    tmp[j-1] = t/(j*(j+1.));
    t *= dx;
  }

  // the trivial cases, not from recursion
  gpred[0] = -dx;
  gcorr[0] = -dx;
  gcorr[1] = -0.5*dx*dx;

  // do recursion
  for( int i=1; i<s; i++ ) {
    if( i != 1 ) {
      for( int j=1; j<=s+1-i; j++ ) {
        tmp[j-1] *= (abscissas[s-1]-abscissas[s-1-i]);
        tmp[j-1] += tmp[j];
      }
    }
    gpred[i] = tmp[0];
    if( i<s-1 ) gcorr[i+1] = tmp[1];
  }
}

// construct only the bottom diagonal of the Newton tableau --
// assuming the rest have already been calculated.

void NewtonLastDiag( double& xnew, double& fnew,
                     Vector<double>& abscissas,
                     Vector<double>& tableau )
{
  int n = abscissas.size();
  int i = n-1;
  abscissas[i] = xnew;
  int j = (i*(i+1))/2;
  tableau[j] = fnew;
  for( int k=j+1; k<=j+i; k++ ) {
    tableau[k] = (tableau[k-1]-tableau[k-i-1])/(abscissas[i]-abscissas[i+j-k]);
  }
}

// roll data "up" in the Newton tableau so the last diagonal is empty
// for the next iteration.

void rollup( Vector<double>& abscissas, Vector<double>& tableau )
{
  int rows = abscissas.size();
  int k = 0;
  for( int r=0; r<rows-1; r++ ) {
    abscissas[r] = abscissas[r+1];
    int k2 = ((r+1)*(r+2))/2;
    for( int c=0; c<=r; c++ ) {
      tableau[k++] = tableau[k2++];
    }
  }
}

// compute the predictor.  on input, y is previous value.
// on output, next value by predictor routine of order "p"

void predictor( int p,
                const Vector<double>& abscissas,
                const Vector<double>& gpred,
                const Vector<double>& tableau,
                double& y )
{
  int s = abscissas.size();
  int k = ((s-2)*(s-1))/2;
  for( int j=0; j<p; j++ )
    y += tableau[k+j]*gpred[j];
}

// compute the corrector.  on input, y is previous value.
// on output, next value by corrector routine of order "p"

void corrector( int p,
                const Vector<double>& abscissas,
                const Vector<double>& gcorr,
                const Vector<double>& tableau,
                double& y )
{
  int s = abscissas.size();
  int k = (s*(s-1))/2;
  for( int j=0; j<p; j++ )
    y += tableau[k+j]*gcorr[j];
}

// input:
//   a,b            limits of x
//   y              y(a)
//   maxallocorder  max order method that can be used
//   h0             input guess for steps size
//   func           dy/dx = func(x,y)
// output:
//   y              y(b)
//   maxorder       max order actually used
//   step           number of steps taken

void adaptiveMSintegrator(  double a, double b, double relerr,
                            int maxallocorder, double h0,
                            double (*func)(double,double),
                            int& maxorder, int& step, double& y )
{

  double hold, ypred, ycorr, xp, beta, betam, betap;
  double steperr, err;
  Vector<double> abscissas(2), tableau(3), gpred, gcorr;
  double h = h0;              // an initial value
  double x = a;               // starting value
  int    trial=0;
  int    allocorder = 1;      // tableau has memory allocated for this order
  int    order = 1;           // initial order

  maxorder = 1;
  step     = 0;

  // put starting info into tableau.
  // NOTE:  we will arrange it so the most recent value of the absicssa is
  // always zero -- to avoid subtracting large numbers.
  abscissas[0] = 0;
  tableau[0]   = (*func)(x,y);

  do {
    trial = 1;
    if( order > maxorder ) maxorder = order;
    do {
      err = relerr * y;
      h = MIN(b - x, h);
      hold = h;
      xp = x+h;
      abscissas[abscissas.size()-1] = h;
      makeGs( abscissas, gpred, gcorr );

      // predictor
      ypred = y;
      predictor( order, abscissas, gpred, tableau, ypred );

      // fill out tableau
      double fp = (*func)(xp,ypred);
      NewtonLastDiag( h, fp, abscissas, tableau );

      // corrector
      ycorr = y;
      corrector( order, abscissas, gcorr, tableau, ycorr );

      // compute new step size, see if order change indicated
      int k = abscissas.size();
      k = (k*(k-1))/2;
      steperr = fabs(gcorr[order-1]*tableau[k+order]);
      beta = pow(err/steperr, 1./(order+1.));
      if( beta > 0.5 ) {
        betam = -1, betap = -1;
        if( order > 1 ) {
          steperr = fabs(gcorr[order-2]*tableau[k+order-1]);
          betam = pow(err/steperr, 1./order);
        }
        if( order < allocorder ) {
          steperr = fabs(gcorr[order]*tableau[k+order+1]);
          betap = pow(err/steperr, 1./(order+2));
        }
        if( betam > MAX(beta,betap) ) {
          // drop order
          h = h*MAX(MIN(16.,betam),1./16.);
          order--;
        } else if( betap > MAX(beta,betam) ) {
          // raise order
          h = h*MAX(MIN(16.,betap),1./16.);
          order++;
        } else {
          h = h*MAX(MIN(16.,beta),1./16.);
        }
        break;
      }
      // the 16 and 1/16 are limits to control crazy steps -- safety factors.
      h = h*MAX(MIN(16.,beta),1./16.);
    } while(++trial<20);
    ++step;
    double fp = (*func)(xp,ycorr);
    NewtonLastDiag( hold, fp, abscissas, tableau );

    // reset the abscissas in the tableau in order to mitigate
    // cancellation errros.
    for( int i=0; i<abscissas.size(); i++ ) abscissas[i] -= hold;

    x = xp;
    y = ycorr;

    if( (order+1 > allocorder) && (allocorder != maxallocorder) ) {
      // increase tableau size
      int s = abscissas.size();
      abscissas.resize(s+1);
      tableau.resize(((s+1)*(s+2))/2);
      allocorder++;
    } else {
      // keep size contant, roll values up
      rollup( abscissas, tableau );
    }
  }while(x<b);
}
