// G. Miller, Numerical Analysis for Engineers and Scientists,
// Cambridge University Press, Cambridge, 2014.
// http://www.cambridge.org/9781107021082

#include <math.h>

#define MIN(A,B) (((A<B)?(A):(B)))
#define MAX(A,B) (((A>B)?(A):(B)))

// a single step of the 4th order Runge-Kutta method.

void RK4( double x, double h, double y, double& ynew, double (*func)(double,double) )
{
  double k1 = (*func)(x,y);
  double k2 = (*func)(x+h/2, y+h*k1/2);
  double k3 = (*func)(x+h/2, y+h*k2/2);
  double k4 = (*func)(x+h, y+h*k3);
  ynew =  y + h*(k1 + k4 + 2*(k2+k3) )/6;
}

void adaptiveSSintegrator( double a, double b, double relerr, double h0,
                          double (*func)(double,double), int& step, double& y )
{
  double x, h, ycoarse, yfine, beta;
  double eps;

  x = a;
  y = 1./901;
  h = h0;
  step = 0;
  do {
    step++;

    do {
      eps = relerr*fabs(y);

      // coarse calculation
      RK4( x, h, y, ycoarse, func );

      // fine calculation
      RK4( x,     h/2, y,     yfine, func );
      RK4( x+h/2, h/2, yfine, yfine, func );

      // find new h
      beta = pow( eps*15/MAX(fabs(yfine-ycoarse),1.e-16), 0.2 );
      beta = MIN(beta,2); // prevent too big of a step

      if( beta > 0.5 ) {
        y = yfine;
        x += h;
        h = h*beta;
        break;
      }
      h = h*beta;
    } while(1);
    if( b-x > h ) h = b-x;
  } while( x < b );
}
