// G. Miller, Numerical Analysis for Engineers and Scientists,
// Cambridge University Press, Cambridge, 2014.
// http://www.cambridge.org/9781107021082

#include <math.h>

#ifndef MAX
#define MAX(A,B) (((A)>(B))?(A):(B))
#endif
  
// recursive adaptive Simpson
//  inputs:
//    a,b:       limits of integration
//    f0,f3,f6:  f(a), f( (a+b)/2 ), f(b)  
//    icoarse:   coarse estimate of integral
//    depth:     useful for debugging.
//    iest:      estimate of abs of overall integral
//  output:  return value of integral over range a,b

double adaptiveSimpson2( double& a, double& b,
                         double& f0, double& f3, double& f6,
                         double& icoarse, int depth,
                         double (*func)(double), double iest )
{
  double w = (b-a)/3;
  double x1 = a+w;
  double x2 = b-w;

  double xl = 0.5*(a+x1);
  double f1 = (*func)(xl);
  double f2 = (*func)(x1);
  double f4 = (*func)(x2);
  double xr = 0.5*(x2+b);
  double f5 = (*func)(xr);

  double is1 = w*( f0 + 4*f1 + f2 )/6;
  double is2 = w*( f2 + 4*f3 + f4 )/6;
  double is3 = w*( f4 + 4*f5 + f6 )/6;
  double ifine = is1 + is2 + is3;

  // (1) if the local integral is insignificant relative to the integral estimate.
  bool stop1 = (iest + (ifine - icoarse) == iest);
  // (2) if the interval is too small to contain unique representable quad pts.
  bool stop2 = ( (xl <= a) || (xr >= b) );

  if( stop1 || stop2 ) {
    return ifine;
  } else {
    return
      adaptiveSimpson2( a,  x1, f0, f1, f2, is1, depth+1, func, iest ) +
      adaptiveSimpson2( x1, x2, f2, f3, f4, is2, depth+1, func, iest ) +
      adaptiveSimpson2( x2, b,  f4, f5, f6, is3, depth+1, func, iest );
  }
}

// a front end to adaptiveSimspon2 to simplify useage

double adaptiveSimpson( double a, double b, double tolerance,
                        double (*func)(double) )
{
  double f0 = (*func)(a);
  double f3 = (*func)(0.5*(a+b));
  double f6 = (*func)(b);
  double icoarse = (b-a)*(f0 + 4*f3 + f6)/6;

  double w = (b-a)/3;
  double x1 = a+w;
  double x2 = b-w;
  double xl = 0.5*(a+x1);
  double f1 = (*func)(xl);
  double f2 = (*func)(x1);
  double f4 = (*func)(x2);
  double xr = 0.5*(x2+b);
  double f5 = (*func)(xr);

  double is1 = w*( f0 + 4*f1 + f2 )/6;
  double is2 = w*( f2 + 4*f3 + f4 )/6;
  double is3 = w*( f4 + 4*f5 + f6 )/6;
  double ifine = is1 + is2 + is3;

  double iest = MAX( fabs(icoarse), fabs(ifine) );
  // rescale per target error
  double machine = 1.1e-16;
  iest *= tolerance/machine;

  // (1) if the local integral is insignificant relative to the integral estimate.
  bool stop1 = (iest + (ifine - icoarse) == iest);
  // (2) if the interval is too small to contain unique representable quad pts.
  bool stop2 = ( (xl <= a) || (xr >= b) );

  if( stop1 || stop2 ) {
    return ifine;
  } else {
    return
      adaptiveSimpson2( a,  x1, f0, f1, f2, is1, 2, func, iest ) +
      adaptiveSimpson2( x1, x2, f2, f3, f4, is2, 2, func, iest ) +
      adaptiveSimpson2( x2, b,  f4, f5, f6, is3, 2, func, iest );
  }
}
