// G. Miller, Numerical Analysis for Engineers and Scientists,
// Cambridge University Press, Cambridge, 2014.
// http://www.cambridge.org/9781107021082

#include <iostream>
#include <iomanip>
#include <math.h>
#include "Matrix.H"
#include <assert.h>

// perform 1 step of conjugate gradient method for symmetric
// positive definite matrix A.  return the norm of residual.

// r   residual
// p   search direction
// x   solution
// tmp spare memory

double cgStep( const Matrix<double>&A,
               Vector<double>& x,
               Vector<double>& p,
               Vector<double>& r,
               Vector<double>& tmp )
{
  int n = A.rows();
  assert( n == A.columns() );
  assert( n == x.size() );
  assert( n == p.size() );
  assert( n == tmp.size() );

  // vector A.p
  for(  int i=0; i<n; i++ ) {
    tmp[i] = 0;
    for( int j=0; j<n; j++ ) tmp[i] += A[i][j]*p[j];
  }

  // alpha, denom of gamma
  double gamma = 0;
  double d=0;
  for( int i=0; i<n; i++ ) {
    gamma += r[i]*r[i];
    d     += p[i]*tmp[i];
  }
  double alpha = gamma/d;

  // x update and r update
  d = 0;
  for( int i=0; i<n; i++ ) {
    x[i] += alpha*p[i];
    r[i] -= alpha*tmp[i];
    d    += r[i]*r[i];
  }

  // finish gamma
  gamma = - d/gamma;

  // p
  d = 0;
  for( int i=0; i<n; i++ ) {
    p[i] = r[i] - gamma*p[i];
    d    += p[i]*p[i];
  }

  // L2 norm of p is L2 norm of r
  return sqrt(d);
}

