// G. Miller, Numerical Analysis for Engineers and Scientists,
// Cambridge University Press, Cambridge, 2014.
// http://www.cambridge.org/9781107021082

#include <iostream>
#include <iomanip>
#include "Matrix.H"
#include <math.h>
using namespace std;

// function to be interpolated
double f(double x)
{
  return 1./(1.+12.*x*x);
}

// given support points, find the coefficients of a polynomial
// by Newton interpolation, for evaluation by Horner scheme.

// tableau:  packed into a vector, increasing index moves up and to right.
// This is an order that permits addition of a new support point at little cost.
// e.g.,
//
//   0
//      2
//   1     5
//      4     9
//   3     8
//      7
//   6


void Newton( Vector<double>& abscissas,
             Vector<double>& ordinates,
             Vector<double>& tableau )
{
  int n = abscissas.size();
  assert( ordinates.size() == n );
  assert( tableau.size() == (n*(n+1))/2 );

  int j = 0;
  for( int i=0; i<n; i++ ) {
    tableau[j] = ordinates[i];
    for( int k=j+1; k<=j+i; k++ ) {
      tableau[k] = (tableau[k-1]-tableau[k-i-1])/(abscissas[i]-abscissas[i+j-k]);
    }
    j += i+1;
  }
}

// evaluate a polynomial constructed by Newton tableau
// using Horner's scheme.  This implementation assumes x0<x1<...<xn
// and uses implicit permutation.

double Horner( Vector<double>& abscissas,
               Vector<double>& tableau,
               double x )
{
  int n = abscissas.size();
  Vector<double>xval(n-1);
  Vector<double>fval(n);

  // find the index of the abscissa greater closest to x
  int r0=0;
  double d0 = fabs(abscissas[r0]-x);
  for( int r=1; r<n; r++ )  {
    double d1 = fabs(abscissas[r]-x);
    if( d1 < d0 ) {
      d0 = d1;
      r0 = r;
    }
  }
  xval[0] = abscissas[r0];
  int indx = (r0*(r0+1))/2;
  fval[0] = tableau[indx];

  // find the rest of the control points, assume x0<x1<x2...
  int rm = r0-1;
  int rp = r0+1;
  for( int c=1; c<n-1; c++ ) {
    if( rp < n ) {
      if( rm >= 0 ) {
        if( fabs(abscissas[rp]-x) < fabs(abscissas[rm]-x) ) {
          xval[c] = abscissas[rp++];
          indx += rp;
          fval[c] = tableau[indx]; 
        } else {
          xval[c] = abscissas[rm--];
          indx ++;
          fval[c] = tableau[indx];
        }
      } else {
        xval[c] = abscissas[rp++];
        indx += rp;
        fval[c] = tableau[indx];
      }
    } else {
      xval[c] = abscissas[rm--];
      indx ++;
      fval[c] = tableau[indx];
    }
  }
  fval[n-1] = tableau[(n*(n+1))/2-1];

  double p = fval[n-1];
  for( int c=n-2; c>=0; c-- ) {
    p *= (x-xval[c]);
    p += fval[c];
  }

  return p;
}

int main()
{
  // interpolation function for regularly space points.
  int n = 11; // number of points, = poly order + 1
  Vector<double> abs_reg(n), ord_reg(n), tab_reg((n*(n+1))/2);
  for( int i=0; i<n; i++ ) {
    abs_reg[i] = -1 + 0.2*i;
    ord_reg[i] = f(abs_reg[i]);
  }
  Newton( abs_reg, ord_reg, tab_reg );

  // same for Chebyshev spacing
  Vector<double> abs_cbs(n), ord_cbs(n), tab_cbs((n*(n+1))/2);
  for( int i=0; i<n; i++ ) {
    abs_cbs[i] = cos(M_PI*(2*i+1)/(2.*n));
    ord_cbs[i] = f(abs_cbs[i]);
  }
  Newton( abs_cbs, ord_cbs, tab_cbs );

  for( int i=0; i<=1000; i++ ) {
    double x = -1 + 0.002*i;
    double f1 = Horner( abs_reg, tab_reg, x );
    double f2 = Horner( abs_cbs, tab_cbs, x );
    cout << x << " " << f1 << " " << f2 << endl;
  }
}
