// G. Miller, Numerical Analysis for Engineers and Scientists,
// Cambridge University Press, Cambridge, 2014.
// http://www.cambridge.org/9781107021082

#include <iostream>
#include <iomanip>
#include <math.h>
#include "Matrix.H"
using namespace std;

// Replace Matrix "A" with L of L LH decomposition.

int Cholesky( Matrix<double>& A )
{
  int n = A.rows();
  int m = A.columns();
  assert( n == m );

  for( int i=0; i<n; i++ ) {
    for( int j=i; j<n; j++ ) {
      double p;
      double x = A[i][j];
      for( int k=i-1; k>=0; k-- )
        x -= A[j][k]*A[i][k];
      if( i==j ) {
        if( x <= 0 ) return -1;
        A[i][j] = sqrt(x);
        p = 1./sqrt(x);
      } else {
        A[j][i] = x*p;
      }
    }
  }
  return 0;
}

// solve Ax=b after "A" has been replaced by "L" in place.
// The resut "x" is returned in the vector "b".

void solveLLH( const Matrix<double>& A, Vector<double>& b )
{
  assert( A.columns() == b.size() );
  int n = A.rows();
  Vector<double> y(b.size());

  // solve Ly = b for y
  for( int i=0; i<n; i++ ) {
    y[i] = b[i];
    for( int j=0; j<i; j++ ) y[i] -= A[i][j]*y[j];
    y[i] /= A[i][i];
  }

  // solve Rx = y for x -- store result in b
  for( int i=n-1; i>=0; i-- ) {
    b[i] = y[i];
    for( int j=i+1; j<n; j++ ) b[i] -= A[j][i]*b[j];
    b[i] /= A[i][i];
  }
}

void printLLH( const Matrix<double>&A )
{
  ios_base::fmtflags ff = cout.flags();
  streamsize prec = cout.precision();

  cout << "L:" << endl << scientific << setprecision(16) << showpos;
  for( int r=0; r<A.rows(); r++ ) {
    for( int c=0; c<=r;  c++ )
      cout << " " <<  A[r][c];
    for( int c=r+1; c<A.columns();  c++ )
      cout << " " <<  0.0;
    cout << endl;
  }

  cout << "LH:" << endl;
  for( int r=0; r<A.rows(); r++ ) {
    for( int c=0; c<r;  c++ )
      cout << " " <<  0.0;
    for( int c=r; c<A.columns();  c++ )
      cout << " " <<  A[c][r];
    cout << endl;
  }

  cout.flags(ff);
  cout << setprecision(prec);
}

