// G. Miller, Numerical Analysis for Engineers and Scientists,
// Cambridge University Press, Cambridge, 2014.
// http://www.cambridge.org/9781107021082

#include <cmath>
#include <complex>
#include "Matrix.H"
#include <iostream>
#include <iomanip>
using namespace std;

#ifndef MAX
#define MAX(A,B) (((A)>(B))?(A):(B))
#endif

// c    s
// s*  -c*

void CGivens( const complex<double>& y1, const complex<double>& y2,
              complex<double>& c, complex<double>& s, complex<double>& v )
{
  double mu = MAX(abs(y1),abs(y2));
  complex<double> s1 = y1/mu;
  complex<double> s2 = y2/mu;
  double k = mu*abs(sqrt( conj(s1)*s1 + conj(s2)*s2 ));
  complex<double> one;
  one.real() = 1;
  one.imag() = 0;
  if( !k ) {
    c = one;
    s = 0;
    v = 0;
    return;
  }

  c = conj(y1)/k;
  if( c.real() < 0 ) { c *= -1; k *= -1; }
  s = conj(y2)/k;
  v = s/(c+one);

  cout << "Givens to op on " << y1 << " " << y2 << endl;
  cout << "  test " << (conj(s)*y1 - conj(c)*y2) << endl;
}

void makeHessenberg( Matrix<complex<double> >& A )
{
  int N = A.rows();
  assert( A.columns() == N );

  Vector<complex<double> > w(N);

  for( int c=0; c<N-2; c++ ) {
    // Householder to put in Hessenberg form

    complex<double> len = 0;
    for( int r=c+2; r<N; r++ )
      len += conj( A[r][c] ) * A[r][c];

    if( abs(len) != 0 ) {
      complex<double> K = sqrt( len + conj( A[c+1][c] ) * A[c+1][c] );
      // want Re( conj(A[c+1][c]) * K ) < 0
      complex<double> tmp = conj(A[c+1][c])*K;
      if( tmp.real() > 0 ) K *= -1;

      for( int r=c+1; r<N; r++ ) w[r] = A[r][c];
      w[c+1] -= K;
      len = sqrt( len + conj(w[c+1])*w[c+1] );

      for( int r=c+1; r<N; r++ ) w[r] /= len;

      A[c+1][c] = K;
      for( int r=c+2; r<N; r++ ) A[r][c] = 0;

      for( int c2=c+1; c2<N; c2++ ) {
        complex<double> d = 0;
        for( int r=c+1; r<N; r++ ) d += conj(w[r])*A[r][c2];
        d *= 2;
        for( int r=c+1; r<N; r++ ) {
          A[r][c2] -= d*w[r];
        }
      }

      for( int c2=0; c2<N; c2++ ) {
        complex<double> d = 0;
        for( int r=c+1; r<N; r++ ) d += A[c2][r]*w[r];
        d *= 2;
        for( int r=c+1; r<N; r++ ) {
           A[c2][r] -= d*conj(w[r]);
        }
      }
    }
  }
}

int oneQRstep( Matrix<complex<double> >& A, int N )
{
  double machine = 1.e-16;
  if( N==1 ) {
    cout << "N=1 return." << endl;
    return 1;
  }

  int z = N-1;
  do {
    if( z == 0 ) break;
    z--;
    double sum = abs(A[z][z]) + abs(A[z+1][z+1]);
    if( abs(A[z+1][z]) <= machine*sum ) {
      ++z;
      break;
    }
  } while( 1 );

  if( z >= N-1 ) {
    cout << "z too big " << z << " " << N << " return." << endl;
    return 1;
  }

  cout << "oneQRstep let Z=" << z << endl;

  complex<double> shift = 0;

  complex<double> qa = 1;
  complex<double> qb = -(A[N-1][N-1]+A[N-2][N-2]);
  complex<double> qc = A[N-1][N-1]*A[N-2][N-2]-A[N-1][N-2]*A[N-2][N-1];
  
  complex<double> sq = qa*qc;
                  sq *= -4.;
                  sq += qb*qb;
                  sq = sqrt(sq);
  complex<double> tmp = conj(qb)*sq;
  if( tmp.real() < 0 ) sq *= -1;


  cout << "qb " << qb << endl;
  cout << "sq " << sq << endl;
  complex<double> mu1 = (qb+sq)/qa; mu1 /= -2.;
  complex<double> mu2 = qc/(qb+sq); mu2 *= -2.;
  cout << "choices " << mu1 << " " << mu2 << endl;
  shift = mu2;
  if( abs(A[N-1][N-1]-mu1) < abs(A[N-1][N-1]-mu2) ) shift = mu1;

  cout << "shift " << shift << endl;

  complex<double> c, s, v;

  CGivens( A[z][z]-shift, A[z+1][z], c, s, v );

  for( int col=z; col<N; col++ ) {

    complex<double> z1 = c*A[z][col]       + s*A[z+1][col];
    complex<double> z2 = conj(s)*A[z][col] - conj(c)*A[z+1][col];
    A[z  ][col] = z1;
    A[z+1][col] = z2;
  }

  cout << "A after pass.1 " << z << endl << A << endl;

  for( int col=z; col<N; col++ ) {
    complex<double> z1 = A[col][z]*conj(c) + A[col][z+1]*conj(s);
    complex<double> z2 = A[col][z]*s       - A[col][z+1]*c;
    A[col][z  ] = z1;
    A[col][z+1] = z2;
  }

  cout << "A after pass.2 " << z << endl << A << endl;

  for( int r=z+1; r<N-1; r++ ) {

    CGivens( A[r][r-1], A[r+1][r-1], c, s, v );

    for( int col=r-1; col<N; col++ ) {

      complex<double> z1 = c*A[r][col]       + s*A[r+1][col];
      complex<double> z2 = conj(s)*A[r][col] - conj(c)*A[r+1][col];
      A[r  ][col] = z1;
      A[r+1][col] = z2;
    }

    cout << "A after pass.1 " << r << endl << A << endl;

    for( int col=z; col<N; col++ ) {
      complex<double> z1 = A[col][r]*conj(c) + A[col][r+1]*conj(s);
      complex<double> z2 = A[col][r]*s       - A[col][r+1]*c;
      A[col][r  ] = z1;
      A[col][r+1] = z2;
    }

    cout << "A after pass.2 " << r << endl << A << endl;
  }

  cout << "matrix " << endl << A << endl;

  return 0;
    
}

int main()
{

  Matrix<complex<double> > A(3,3);

  A[0][0].real() = 1; A[0][0].imag() = 0;
  A[0][1].real() = 1; A[0][1].imag() = 0;
  A[0][2].real() = 0; A[0][2].imag() = 0;

  A[1][0].real() = 1; A[1][0].imag() = 0;
  A[1][1].real() = 0; A[1][1].imag() = 0;
  A[1][2].real() = 1; A[1][2].imag() = 0;

  A[2][0].real() = 1; A[2][0].imag() = 0;
  A[2][1].real() = 0; A[2][1].imag() = 0;
  A[2][2].real() = 0; A[2][2].imag() = 0;

  cout << A << endl;

  makeHessenberg(A);
  cout << A << endl;

  int curN = A.rows();
  Vector<complex<double> > lambda(curN);
  for( int step=0; step<30; step++ ) {
    int got = oneQRstep( A, curN );
    if( got ) {
      curN--;
      lambda[curN] = A[curN][curN];
      if( curN == 0 ) break;
    }
  }

  cout << "lambda" << endl << lambda << endl;
  
  return 1;
}
