// G. Miller, Numerical Analysis for Engineers and Scientists,
// Cambridge University Press, Cambridge, 2014.
// http://www.cambridge.org/9781107021082

#include <math.h>
#include "Matrix.H"
#include "Givens.H"

#ifndef MIN
#define MIN(A,B) (((A)<(B))?(A):(B))
#endif

// return false of it appears that R is singular.
bool qp_modify_append( const Vector<double>& np, 
                       const Vector<double>& d,
                       Matrix<double>& J,
                       Matrix<double>& R )
{
  int n = R.rows();
  int m = R.columns();

  // now fix with a single Householder step
  Vector<double> w(n);

  double len = 0;
  w[m] = d[m];
  for( int i=m+1; i<n; i++ ) {
    w[i] = d[i];
    len += d[i]*d[i];
  }
  double K = len + d[m]*d[m];
  if( K <= 0 ) return false;
  K = sqrt( K );
  if( d[m] > 0 ) K = -K;
  w[m] -= K;
  len = sqrt( len + w[m]*w[m] );
  for( int i=m; i<n; i++ ) w[i] /= len;

  // fix J by right multiplication by (1-2*w wT)
  for( int r=0; r<n; r++ ) {
    double dot = 0;
    for( int c=m; c<n; c++ ) dot += J[r][c]*w[c];
    dot *= -2;
    for( int c=m; c<n; c++ ) J[r][c] += dot*w[c];
  }

  // and fix R
  R.resize(n,m+1);
  for( int i=0; i<m; i++ ) R[i][m] = d[i];
  R[m][m] = K;
  for( int i=m+1; i<n; i++ ) R[i][m] = 0;

  return true;
}

void qp_modify_delete( Matrix<double>& J,
                       Matrix<double>& R,
                       Vector<int>& A,
                       Vector<double>& lambda,
                       int cdel )
{
  for( int i=0; i<A.size(); i++ ) {
    if( A[i] == cdel ) {
      A[i] = -1;
    } else if( A[i] > cdel ) {\
      A[i] --;
    }
  }

  // fix A and lambda, and begin on R
  int n = R.rows();
  int m = R.columns();

  for( int c=cdel+1; c<m; c++ ) {
    lambda[c-1] = lambda[c];
    for( int r=0; r<n; r++ ) {
      R[r][c-1] = R[r][c];
    }
  }
  lambda[m-1] = 0;

  R.resize(n,--m);

  // change from Hessenberg to right triangular with Givens reflections
  for( int r=cdel; r<m; r++ ) {
    double cf, sf, vf;
    Givens( R[r][r], R[r+1][r], cf, sf, vf );
    for( int c=r; c<m; c++ ) {
      double y1 = R[r  ][c];
      double y2 = R[r+1][c];
      R[r  ][c] = cf*y1 + sf*y2;
      R[r+1][c] = vf*( R[r][c] + y1 ) - y2;
    }
    for( int c=0; c<n; c++ ) {
      double y1 = J[c][r  ];
      double y2 = J[c][r+1];
      J[c][r  ] = cf*y1 + sf*y2;
      J[c][r+1] = vf*( J[c][r] + y1 ) - y2;
    }
  }
}

// the optimum of f(x) = (1/2) x^T G x + b^T x , with G = L L^T positive definite,
// subject to constraints Ce^T x + ce =  0, Ce dimension (n x n_equality)
// and                    Ci^T x + ci >= 0, Ci dimension (n x n_inequality)
// on input, x is the solution to the unconstrained problem.
// it is assumed the constraints are independent.

// return value is 0 if successful, or 1 if no feasible solution exists.

// df is change to quadratic function upon imposition of constraints.

// tolerance is the tolerance with which constraints are enforced:
// a reasonable value is a few times epsilon * condition number of G,
// if available.

int ConvexQP( const Matrix<double>& L,
              Vector<double>& x,
              const Matrix<double>& Ce, const Vector<double>& ce,
              const Matrix<double>& Ci, const Vector<double>& ci,
              double& df,
              const double& tolerance )
{
  int n            = x.size();
  int n_equality   = ce.size();
  int n_inequality = ci.size();
  Matrix<double> J(n,n);
  Matrix<double> R(n,n);
  Vector<double> np(n);
  Vector<double> d(n);
  Vector<double> z(n);
  Vector<double> r(n);
  Vector<double> s(n_inequality);
  Vector<double> lambda(n);

  // represent "A" list as a mapping of inequality # to column # in matrices N,R
  Vector<int>    A(n_inequality);

  for( int i=0; i<n; i++ ) lambda[i] = 0;
  for( int i=0; i<n_inequality; i++ ) A[i] = -1;
  
  // at start, Q=I and J = L^{-T}
  // LT J = I
  // make J
  for( int c=0; c<n; c++ ) {
    for( int k=n-1; k>c; k-- ) J[k][c] = 0;
    J[c][c] = 1./L[c][c];
    for( int k=c-1; k>=0; k-- ) {
      double sum = 0;
      for( int l=k+1; l<n; l++ ) sum += L[l][k]*J[l][c];
      J[k][c] = -sum/L[k][k];
     }
  }

  // initialize R=0
  for( int r=0; r<n; r++ )
    for( int c=0; c<n; c++ ) R[r][c] = 0;
  R.resize(n,0);

  int n_constraints = 0;

  // first pass - add equality constraints
  for( int p=0; p<n_equality; p++ ) {

    for( int i=0; i<n; i++ ) np[i] = Ce[i][p];

    // d is (J1,J2)^T n_p
    for( int j=0; j<n; j++ ) {
      d[j] = 0;
      for( int i=0; i<n; i++ )
         d[j] += J[i][j]*np[i];
    }

    // z = H n_p = J2 * d
    // step direction in primal (x) space
    for( int i=0; i<n; i++ ) {
     z[i] = 0;
     for( int j=n_constraints; j<n; j++ )
       z[i] += J[i][j]*d[j];
    }

    // r = Ndagger n_p = R^-1 d
    // neg of step direction in dual (lambda) space
    for( int i=n_constraints-1; i>=0; i-- ) {
      double sum = 0;
      for( int j=i+1; j<n_constraints; j++ ) sum += R[i][j]*r[j];
      r[i] = ( d[i] - sum ) / R[i][i];
    }

    // t_p is full step length to make constraint np feasible
    double dot1 = 0, dot2 = 0;
    for( int i=0; i<n; i++ ) {
      dot1 += np[i]*x[i];
      dot2 += np[i]*z[i];
    }
    double t_p = - ( dot1 + ce[p] )/dot2;

    // update x assuming full step works
    for( int i=0; i<n; i++ ) x[i] += t_p*z[i];

    // update lambda, assuming same
    for( int i=0; i<n_constraints; i++ ) lambda[i] -= t_p*r[i];
    lambda[n_constraints] = t_p;

    // change to function value
    df += 0.5*t_p*t_p*dot2;

    n_constraints++;

    if( !qp_modify_append( np, d, J, R ) ) {
      // R singular -- constraints must be dependent
      return 1;
    }
  }


  // now the inequality constraints
  do {

    // find a violated inequality constraint
    double worst_s = 1;
    int p        = -1;
    for( int i=0; i<n_inequality; i++ ) {
      s[i] = ci[i];
      for( int j=0; j<n; j++ ) s[i] += Ci[j][i]*x[j];
      if( A[i] < 0 && s[i] < worst_s ) {
        worst_s = s[i]; 
        p       = i;
      }
    }

    if( worst_s >= -tolerance ) {
      // no constraint has been violated.  done!.
      return 0;
    }

    for( int i=0; i<n; i++ ) np[i] =  Ci[i][p];
    double lambda_p = 0;

    do {

      // d is (J1,J2)^T n_p
      for( int j=0; j<n; j++ ) {
        d[j] = 0;
        for( int i=0; i<n; i++ )
           d[j] += J[i][j]*np[i];
      }

      // z = H n_p = J2 * d
      // step direction in primal (x) space
      for( int i=0; i<n; i++ ) {
       z[i] = 0;
       for( int j=n_constraints; j<n; j++ )
         z[i] += J[i][j]*d[j];
      }

      // r = Ndagger n_p = R^-1 d
      // neg of step direction in dual (lambda) space
      for( int i=n_constraints-1; i>=0; i-- ) {
        double sum = 0;
        for( int j=i+1; j<n_constraints; j++ ) sum += R[i][j]*r[j];
        r[i] = ( d[i] - sum ) / R[i][i];
      }

      // find a step length
      bool t_d_bounded = false;
      bool t_p_bounded = false;
      double t, t_d, t_p;

      // search for step length in dual (lambda) space.
      // "l" is first constraint to be infeasible.
      int l;
      for( int k=n_equality; k<n_constraints; k++ ) {
        if( r[k] > 0.0 ) {
          if( (!t_d_bounded) || (lambda[k] <  r[k]*t_d) ) {
            t_d = lambda[k]/r[k];
            t_d_bounded = true;
            l = k;
          }
        }
      }

      // search for step length in primal (x) space.
      double dot1 = 0, dot2 = 0;
      for( int i=0; i<n; i++ ) {
        dot1 += z[i]*z[i];
        dot2 += np[i]*z[i];
      }
      if( dot1 > 0 && dot2 != 0 ) {
        t_p = -s[p]/dot2;
        t_p_bounded = true;
      }

      if( t_d_bounded ) t = t_d;
      if( t_p_bounded ) t = t_p;
      if( t_d_bounded && t_p_bounded ) t = MIN(t_d,t_p);

      // if unbounded, then failure
      if( !(t_d_bounded || t_p_bounded) ) {
        return 1;
        break;
      }

      if( !t_p_bounded ) {
        // constraint "l" is infeasible.

        for( int k=0; k<n_constraints; k++ ) lambda[k] -= t*r[k];
        lambda_p += t;
        
        qp_modify_delete( J, R, A, lambda, l );
        n_constraints--;

      } else {

        for( int k=0; k<n; k++ ) x[k] += t*z[k];
        df += t*dot2*( 0.5*t + lambda_p );
        for( int k=0; k<n_constraints; k++ ) lambda[k] -= t*r[k];
        lambda_p += t;

        if( t == t_p ) {

          // a full step was taken.
          if( !qp_modify_append( np, d, J, R ) ) {
            // R singular.
            return 1;
          }
          A[p] = n_constraints;
          lambda[n_constraints] = lambda_p;
          n_constraints++;
          break;  // break out of do loop

        } else  {

          for( int i=0; i<n_inequality; i++ ) {
            s[i] = ci[i];
            for( int j=0; j<n; j++ ) s[i] += x[j]*Ci[j][i];
          } 
          qp_modify_delete( J, R, A, lambda, l );
          n_constraints--;

        }
      }

    } while(1);
  } while(1);
  return 0;
}
