// G. Miller, Numerical Analysis for Engineers and Scientists,
// Cambridge University Press, Cambridge, 2014.
// http://www.cambridge.org/9781107021082

#include <math.h>
#include "Matrix.H"
#include <assert.h>

// re-order data according to bit reversal algorithm of Rodriguez 1988.

void reverse( Vector<double>& fR,
              Vector<double>& fI,
              int N, int m )
{
  int i, j, k , NV2, last;
  double tmpr, tmpi;

  NV2=N>>1;
  last = (N-1)-(1<<((m+1)>>1));
  j = 0;
  for( i=1; i<=last; i++ ) {
    for( k=NV2; k<=j; k>>=1 ) j-= k;
    j += k;
    if( i < j ) {
      tmpr  = fR[i]; tmpi  = fI[i];
      fR[i] = fR[j]; fI[i] = fI[j];
      fR[j] = tmpr;  fI[j] = tmpi;
    }
  }
}

// construct the integer address j_0,...,j_{p-1},k_{m-1-p},...,k_0
// convention:  k_0 is lowest-order bit
// alo has bit j_{p-1} clear,
// ahi has bit j_{p-1} set

inline void address( int k, int j, unsigned int twomm1, int p,
                     int& alo, int& ahi )
{
  unsigned int jmask = 1;
  unsigned int r=twomm1;  // 2^{m-1}
  alo = ahi = 0;

  // skip the p-1 bit
  for( int b=0; b<p-1; b++ ) {
    if(j & jmask) alo |= r;
    jmask <<= 1;  // shift left 1 = multiply by 2
    r >>= 1;      // shift right 1 = divide by 2
  }
  ahi = r;
  // now r is 2^(m-1-p)
  r <<= 1;  // now 2^(m-p)
  r -= 1;   // now 1 in every place up to and including  m-1-p
  alo |= ( ((unsigned int)k) & r );
  ahi |= alo ;
}

// fast fourier transform
// if forward=true convert input data to phase amplitutdes
// if forward=false convert phase amplitues to data

void fft( Vector<double>& fR,
          Vector<double>& fI,
          bool forward )
{
  // find "m"
  int m = 0;
  int N = fR.size();
  assert( fR.size() == fI.size() );

  while( N /=2  ) { m++; }

  N = fR.size();
  int jmax=1;
  int kmax=N/2;
  for( int p=0; p<m; p++ ) {
    for( int j=0; j<jmax; j++ ) {
      double ang = M_PI*j/((double)jmax);
      if( forward ) ang = -ang;
      double c = cos(ang);
      double s = sin(ang);
      for( int k=0; k<kmax; k++ ) {

        // addresses
        // alo has bit j_(p-1) clear
        // ahi has bit j_(p-1) set

        int alo, ahi;
        address( k, j, N/2, p+1, alo, ahi );

        double ur = fR[alo];
        double ui = fI[alo];

        double vr = fR[ahi];
        double vi = fI[ahi];

        fR[alo] = ur + c*vr - s*vi;
        fI[alo] = ui + c*vi + s*vr;

        fR[ahi] = ur - c*vr + s*vi;
        fI[ahi] = ui - c*vi - s*vr;
      }
    }
    jmax *= 2;
    kmax /= 2;
  }

  if( forward ) {
    for( int i=0; i<N; i++ ) {
      fR[i] /= N;
      fI[i] /= N;
    }
  }

  reverse( fR, fI, N, m );
}
