// G. Miller, Numerical Analysis for Engineers and Scientists,
// Cambridge University Press, Cambridge, 2014.
// http://www.cambridge.org/9781107021082

#include <iostream>
#include <iomanip>
#include <math.h>
#include "Matrix.H"
#include "GE.H"
using namespace std;

// Perform Gaussian Elimination with pivoting:
//  -- pivot type is specified by variable "type".
//  -- permutations are represented by vectors of indices "PermL" and "PermR".
//  -- matrix "A" is replaced by "LR" in place.
// Return 0 on success, -1 of there's a problem.

int ge( Matrix<double>& A,
        Vector<int>& PermL,
        Vector<int>& PermR,
        GEPivotType type )
{
  int n = A.rows();
  int m = A.columns();
  assert( n == m );

  // initialize the permutation list
  for( int i=0; i<n; i++ ) PermL[i] = PermR[i] = i;

  Vector<double> rowscale;
  if( type == IMPLICITSCALEDPP ) {
    rowscale.resize(n);
    for( int r=0; r<n; r++ ) {
      double s = 0;
      for( int c=0; c<n; c++ ) s += fabs(A[r][c]);
      rowscale[r] = s;
    }
  }

  for( int d=0; d<n-1; d++ ) {

    int pivot_c, pivot_r;
    double pval;

    // find pivot
    if( type == TRIVIAL ) {
      // Trival pivoting:  find first nonzero element in column
      pivot_c = d;
      for( int r=d; r<n; r++ ) {
        if( A[r][d] != 0 ) {
          pivot_r = r;
          pval = fabs(A[r][d]);
          break;
        }
      }
    } else if( type == PARTIAL ) {
      // Partial pivoting:  find largest element in column
      pivot_c = d;
      pivot_r = d;
      pval = fabs(A[d][d]);
      for( int r=d+1; r<n; r++ ) {
        double t = fabs(A[r][d]);
        if( t > pval ) {
          pval = t;
          pivot_r = r;
        }
      }
    } else if( type == IMPLICITSCALEDPP ) {
      pivot_c = d;
      pivot_r = d;
      pval = fabs(A[d][d])/rowscale[d];
      for( int r=d+1; r<n; r++ ) {
        double t = fabs(A[r][d])/rowscale[r];
        if( t > pval ) {
          pval = t;
          pivot_r = r;
        }
      }
    } else if( type == ROOK ) {
      // find first that is largest in row and column
      int  num_pass = 0;       // number of searches where max passes 
      bool rowcol   = true;    // begin search in row.
      pivot_c = pivot_r = d;   // starting guess
      int r=pivot_r, c=pivot_c;
      pval = fabs(A[r][c]);
      do {
        bool change = false;
        c = pivot_c;
        r = pivot_r;
        for( int i=d; i<n; i++ ) {
          if( rowcol ) {
            r = i;
          } else {
            c = i;
          }
          double t = fabs(A[r][c]);
          if( t > pval ) {
            pval = t;
            pivot_r = r;
            pivot_c = c;
            change = true;
          }
        }
        if( !change ) {
          num_pass ++;
        } else {
          num_pass = 1;
        }
        rowcol = !rowcol;
      } while( num_pass != 2 );
    } else if( type == COMPLETE ) {
      // find largest in all rows and columns.
      pval = 0;
      for( int r=d; r<n; r++ ) {
        for( int c=d; c<n; c++ ) {
          double t = fabs(A[r][c]);
          if( t > pval ) {
            pval = t;
            pivot_c = c;
            pivot_r = r;
          }
        }
      }
    } else {
      cout << "unknown pivot type" << endl;
      return -1;
    }

    if( pval <= 0 ) {
      cout << "bad pivot" << endl;
      return -1;
    }

    if( pivot_r != d ) {
      // a row permutation occurred.  Do it.
      int k = PermL[pivot_r]; PermL[pivot_r] = PermL[d]; PermL[d] = k;
      for( int i=0; i<n; i++ ) {
        double t = A[d][i];  A[d][i] = A[pivot_r][i]; A[pivot_r][i] = t;
      }

      if( type == IMPLICITSCALEDPP ) {
        double t = rowscale[d];
        rowscale[d] = rowscale[pivot_r];  rowscale[pivot_r] = t;
      }
    }

    if( pivot_c != d ) {
      // a column permutation occurred.  Do it.
      int k = PermR[pivot_c]; PermR[pivot_c] = PermR[d]; PermR[d] = k;
      for( int i=0; i<n; i++ ) {
        double t = A[i][d]; A[i][d] = A[i][pivot_c]; A[i][pivot_c] = t;
      }
    }

    // do the elimination:  everything above this was pivoting overhead.
    for( int r=d+1; r<n; r++ ) {
      double f = A[r][d]/A[d][d];
      A[r][d] = f;
      for( int c=d+1; c<n; c++ ) {
        A[r][c] -= f*A[d][c];
      }
    }
  }

  return 0;
}

// solve Ax=b after "A" has been replaced by "LR" in place, with
// associated permutations stored in index lists "PermL" and "PermR".
// The result "x" is returned in the vector "b".

void solveLR( const Matrix<double>& A,
              const Vector<int>& PermL,
              const Vector<int>& PermR,
              Vector<double>& b )
{
  assert( A.columns() == b.size() );
  int n = A.rows();
  Vector<double> y(b.size());


  // let y = Pleft b
  for( int i=0; i<n; i++ ) y[i] = b[PermL[i]];

  // solve Lb = y for b
  for( int i=0; i<n; i++ ) {
    b[i] = y[i];
    for( int j=0; j<i; j++ )
      b[i] -= A[i][j]*b[j];
  }

  // solve Ry = b for y 
  for( int i=n-1; i>=0; i-- ) {
    y[i] = b[i];
    for( int j=i+1; j<n; j++ ) y[i] -= A[i][j]*y[j];
    y[i] /= A[i][i];
  }

  // let b = Pright y
  for( int i=0; i<n; i++ ) b[i] = y[PermR[i]];
}

void printLRP( const Matrix<double>&A,
               const Vector<int>& PermL,
               const Vector<int>& PermR )
{
  ios_base::fmtflags ff = cout.flags();
  streamsize prec = cout.precision();

  cout << "L:" << endl << scientific << setprecision(16) << showpos;
  for( int r=0; r<A.rows(); r++ ) {
    for( int c=0; c<r;  c++ )
      cout << " " <<  A[r][c];
    cout << " " <<  1.0;
    for( int c=r+1; c<A.columns();  c++ )
      cout << " " <<  0.0;
    cout << endl;
  }

  cout << "R:" << endl;
  for( int r=0; r<A.rows(); r++ ) {
    for( int c=0; c<r;  c++ )
      cout << " " <<  0.0;
    for( int c=r; c<A.columns();  c++ )
      cout << " " <<  A[r][c];
    cout << endl;
  }

  cout << "PL:" << endl << noshowpos;
  for( int r=0; r<A.rows(); r++ ) {
    int rr = PermL[r];
    for( int c=0; c<A.columns(); c++ )
      cout << " " << ((c==rr)?1:0);
     cout << endl;
  }

  cout << "PR:" << endl << noshowpos;
  for( int r=0; r<A.rows(); r++ ) {
    int rr = PermR[r];
    for( int c=0; c<A.columns(); c++ )
      cout << " " << ((c==rr)?1:0);
     cout << endl;
  }

  cout.flags(ff);
  cout << setprecision(prec);
}
