// G. Miller, Numerical Analysis for Engineers and Scientists,
// Cambridge University Press, Cambridge, 2014.
// http://www.cambridge.org/9781107021082

#include <iostream>
#include <iomanip>
#include <math.h>
#include "Matrix.H"
#include "GE.H"
using namespace std;

inline double rndten(double t)
{
  double v = t;
  if( v == 0 ) return v;

  int s = ((v>0) ? 1 : -1);
  v *= s;

  // round to 2 decimal places.

  // first make number lie between 10 and 99
  double e=1;
  while( v > 99 ) { v /= 10.; e*= 10.; }
  while( v < 10 ) { v *= 10.; e/= 10.; }

  // now round v
  int i = round(v);

//cout << t << "      ------------------> " << i*e*s << endl;
  return i*e*s;
}

int main()
{
  int code, n = 3;
  Matrix<double> A(n,n), A0(n,n);
  Vector<double> b(n), b0(n), r(n);
  Vector<int>    pl(n), pr(n);

  for( int k=0; k<5; k++ ) {

    cout << "2 decimal calculation from start with method " << k << endl;

    if( n == 2 ) {
      A[0][0] = 5;    A[0][1] = 1000;
      A[1][0] = 1000; A[1][1] = 1000;

      b[0] = 500;
      b[1] = 1000;
    } else if( n == 3 ) {
      A[0][0] = 1; A[0][1] = 1; A[0][2] = 2e9;
      A[1][0] = 2; A[1][1] =-1; A[1][2] = 1e9;
      A[2][0] = 1; A[2][1] = 2; A[2][2] = 0;
      b[0]    = 2e9 + 2;
      b[1]    = 1e9 + 1;
      b[2]    = 3;
    }

    for( int i=0; i<n; i++ ) {
      b0[i] = b[i];
      for( int j=0; j<n; j++ ) A0[i][j] = A[i][j];
    }

    cout << "A" << endl << A << endl;
    cout << "b" << endl << b << endl;


    switch(k) {
    case 0: cout << "trivial pivoting" << endl;
            code = ge( A,pl,pr, TRIVIAL);
            break;
    case 1: cout << "partial pivoting" << endl;
            code = ge( A,pl,pr, PARTIAL);
            break;
    case 2: cout << "implicit scaled pivoting" << endl;
            code = ge( A,pl,pr, IMPLICITSCALEDPP);
            break;
    case 3: cout << "rook pivoting" << endl;
            code = ge( A,pl,pr, ROOK );
            break;
    case 4: cout << "complete pivoting" << endl;
            code = ge( A,pl,pr, COMPLETE );
            break;
    }
    if( code != 0 ) {
      cout << "error" << endl;
    } else {
      printLRP(A,pl,pr);
      solveLR( A, pl, pr, b );
      cout << "x= " << scientific  << setprecision(16) << endl << b << endl;
    }

    // iterative refinement
    //   r = b - Ax
    //   A dx = r
    //   x += dx

    // check:  A(x+dx) = (b-r) + r = b.
    for( int iter=0; iter<10; iter++ ) {
      cout << "iterative refinement pass " << iter+1 << endl;
      // make residual
      double l2=0;
      for( int i=0; i<n; i++ ) {
        r[i] = b0[i];
        for( int j=0; j<n; j++ ) r[i] = rndten(r[i] - rndten(A0[i][j]*b[j]));
        l2 += r[i]*r[i];
      }
      l2 = sqrt(l2);
      cout << "r " << endl << r << endl;
      if( l2 < 1.e-9 ) break;
      // solve for dx
      solveLR( A, pl, pr, r );
      cout << "dx " << endl << r << endl;
      // and update x
      for( int i=0; i<n; i++ ) b[i] = rndten(b[i] + r[i]);
      cout << "x " << endl << b << endl;
    }
    
          
      
  }
}
