// G. Miller, Numerical Analysis for Engineers and Scientists,
// Cambridge University Press, Cambridge, 2014.
// http://www.cambridge.org/9781107021082

#define VERBOSE
#include <iostream>
#include <iomanip>
#include <sstream>
#include <math.h>
#include "Matrix.H"
#include "SVD.H"
#include "QR.H"
#include "HQsort.H"
#include "rysfit.H"
using namespace std;

// McMurchie-Davidson d_0^{nmu,nlambda} factors for overlap and KE.
// alpha = alpha_mu + alpha_lambda
// a_amu = x_a - x_mu
// a_alambda = x_a - x_lambda

double MD( const double& alpha,
           const int&    a_nmu, const int&    a_nlambda,
           const double& a_dx_amu, const double& a_dx_alambda )
{
  // order so what we call mu has the bigger n
  int nmu, nlambda;
  double dx_amu, dx_alambda;
  if( a_nmu >= a_nlambda ) {
    nmu     = a_nmu;
    nlambda = a_nlambda;
    dx_amu     = a_dx_amu;
    dx_alambda = a_dx_alambda;
  } else {
    nmu     = a_nlambda;
    nlambda = a_nmu;
    dx_amu     = a_dx_alambda;
    dx_alambda = a_dx_amu;
  }

  Vector<double> d(nmu+nlambda+1);
  d[0] = 1;
  for( int n=1; n<=nmu+nlambda; n++ ) {
    d[n] = dx_amu*d[n-1];
    if( n>=2 ) d[n] += (n-1)*d[n-2]/(2*alpha);
  }
  for( int m=1; m<=nlambda; m++ ) {
    for( int n=0; n<=nmu+nlambda-m; n++ ) {
      d[n] = (dx_alambda-dx_amu)*d[n] + d[n+1];
    }
  }

  return d[nmu];
}

double Rys_A( const int& nmu, const int& nlam,
              const double& tsq, const double& alpha,
              const double& xmu, const double& xlam,
              const double& xa,  const double& xi )
{
  Vector<double>A (nmu+nlam+1);

  A[0] = 1;
  for( int n=1; n<=nmu+nlam; n++ ) {
    A[n] = ( tsq*(xi-xa)  + (xa-xmu) )*A[n-1];
    if( n >= 2 ) A[n] += (n-1)*(1-tsq)*A[n-2]/(2.*alpha);
  }

  for( int m=1; m<= nlam; m++ ) {
    for( int n=0; n<=nmu+nlam-m; n++ ) {
      A[n] =  A[n]*(xmu-xlam) + A[n+1];
    }
  }

  return A[nmu];
}

// Rys Dupuis King "G" factors, relative to G_00 = 1
double RDK_G( const int& na, const int& nb,
              const double& alpha, const double& beta, const double& tsq,
              const double& xa, const double& xb,
              const double& xi, const double& xk )
{
  double b00 = tsq/(2.*(alpha+beta));
  double b10 = 1./(2.*alpha)  - beta*b00/alpha;
  double c00 = (xa-xi) + beta*tsq*(xb-xa)/(alpha+beta);
  double cprime = (xb-xk) + alpha*tsq*(xa-xb)/(alpha+beta);
  double bprime = 1./(2.*beta) - alpha*b00/beta;

  Matrix<double> G(na+1,nb+1);

  for( int ia=0; ia<=na; ia++ ) {
    for( int ib=0; ib<=nb; ib++ ) {
      if( ib==0 && ia==0 ) {
        G[ia][ib] = 1;
      } else if( ib==0 ) {
        G[ia][ib] = c00*G[ia-1][ib];
        if( ia > 1 ) G[ia][ib] += (ia-1)*b10*G[ia-2][ib];
//      if( ib > 0 ) G[ia][ib] += ib*b00*G[ia-1][ib-1];
      } else {
        G[ia][ib] = cprime*G[ia][ib-1];
        if( ib > 1 ) G[ia][ib] += (ib-1)*bprime*G[ia][ib-2];
        if( ia > 0 ) G[ia][ib] += ia*b00*G[ia-1][ib-1];
      }
    }
  }
  return G[na][nb];
}

// mu,lambda on electron 1; nu,sigma on electron 2
double RDK_Gxfer( const int& nmu, const int& nnu,
                 const int& nlambda, const int& nsigma,
                 const double& A, const double& B, const double& tsq,
                 const double& xa, const double& xb,
                 const double& xmu, const double& xnu ,
                 const double& xlambda, const double& xsigma )
{
  Matrix<double> Ii(nlambda+1,nlambda+1);
  Matrix<double> Ik(nsigma+1,nsigma+1);

  for( int n2=nnu; n2<=nnu+nsigma; n2++ ) {
    // for each "n" on electron 2 ...

    // find the G_{n1,n2} factor
    for( int n1=nmu; n1<=nmu+nlambda; n1++ ) {
      // first index:  related to n1
      // second index:  nlambda, initially zero
      Ii[n1-nmu][0] = RDK_G( n1, n2, A, B, tsq, xa, xb, xmu, xnu );
    }


    // transfer momentum from mu to lambda
    for( int j=1; j<=nlambda; j++ ) {
      for( int n=nmu; n<=nmu+nlambda-j; n++ ) {
        Ii[n-nmu][j] = Ii[n-nmu+1][j-1] + (xmu-xlambda)*Ii[n-nmu][j-1];
      }
    }


    // save result in matrix Ik for nu/sigma transfer.
    Ik[n2-nnu][0] = Ii[0][nlambda];
  }


  // transfer from nu to sigma
  for( int n2=1; n2<=nsigma; n2++ ) {
    for( int n1=nnu; n1<=nnu+nsigma-n2; n1++ ) {
      Ik[n1-nnu][n2] = Ik[n1-nnu+1][n2-1] + (xnu-xsigma)*Ik[n1-nnu][n2-1];
    }
  }


  return Ik[0][nsigma];
}

void Rys( const double& alpha, Vector<double>& rys_t, Vector<double>& rys_W )
{
  static double rHermite[4] = {
     2.93063742025724401922350270524,
     1.98165675669584292585463063976,
     1.15719371244678019472076577906,
     0.38118699020732211685471888558 };

   static double wHermite[4] = {
     0.0001996040722113676192060904525,
     0.0170779830074134754562030564364,
     0.2078023258148918795432586202857,
     0.6611470125582412910304159744958 };

  int order = 4;
  double *tp[4], *wp[4];

  double m, hw;
  if( alpha <= 1 ) {
    m=0.5; hw=0.5;
    tp[0] = s1t0; tp[1] = s1t1; tp[2] = s1t2; tp[3] = s1t3;
    wp[0] = s1W0; wp[1] = s1W1; wp[2] = s1W2; wp[3] = s1W3;
  } else if( alpha <= 5 ) {
    m=3; hw=2;
    tp[0] = s2t0; tp[1] = s2t1; tp[2] = s2t2; tp[3] = s2t3;
    wp[0] = s2W0; wp[1] = s2W1; wp[2] = s2W2; wp[3] = s2W3;
  } else if( alpha <= 10 ) {
    m = 7.5; hw=2.5;
    tp[0] = s3t0; tp[1] = s3t1; tp[2] = s3t2; tp[3] = s3t3;
    wp[0] = s3W0; wp[1] = s3W1; wp[2] = s3W2; wp[3] = s3W3;
  } else if( alpha <= 15 ) {
    m = 12.5; hw=2.5;
    tp[0] = s4t0; tp[1] = s4t1; tp[2] = s4t2; tp[3] = s4t3;
    wp[0] = s4W0; wp[1] = s4W1; wp[2] = s4W2; wp[3] = s4W3;
  } else if( alpha <= 20 ) {
    m = 17.5; hw=2.5;
    tp[0] = s5t0; tp[1] = s5t1; tp[2] = s5t2; tp[3] = s5t3;
    wp[0] = s5W0; wp[1] = s5W1; wp[2] = s5W2; wp[3] = s5W3;
  }

  if( alpha <= 20 ) {
    int nc=20; // number of terms to include
    // Clenshaw algorithm
    Vector<double> y(nc+2);
    y[nc+1] = y[nc] = 0;
    // put in range -1,1
    double alpham = (alpha-m)/hw;
    for( int o=0; o<order; o++ ) {
      for( int k=nc-1; k>=0; k-- ) {
        int f = (k==0)?1:2;
        y[k] = f*alpham*y[k+1] - y[k+2] + tp[o][k];
      }
      rys_t[o] = y[0];
    }
    for( int o=0; o<order; o++ ) {
      for( int k=nc-1; k>=0; k-- ) {
        int f = (k==0)?1:2;
        y[k] = f*alpham*y[k+1] - y[k+2] + wp[o][k];
      }
      rys_W[o] = y[0];
    }
  } else { // a Laurent expansion fit
    tp[0] = t0coef; tp[1] = t1coef; tp[2] = t2coef; tp[3] = t3coef;
    wp[0] = W0coef; wp[1] = W1coef; wp[2] = W2coef; wp[3] = W3coef;
    double alphamax = 50.;
    int kmin = -3;
    int nk = 11;
    for( int o=0; o<order; o++ ) {
      double q=0;
      for( int j=nk-1; j>=0; j-- ) {
        q *= alpha/alphamax;
        q += tp[o][j];
      }
      q *= pow(alpha/alphamax,(double)kmin);
      double rsq = rHermite[o]; rsq *= rsq;
      rys_t[o] = rsq/alpha + exp(-alpha)*q;
    }
    for( int o=0; o<order; o++ ) {
      double q=0;
      for( int j=nk-1; j>=0; j-- ) {
        q *= alpha/alphamax;
        q += wp[o][j];
      }
      q *= pow(alpha/alphamax,(double)kmin);
      rys_W[o] = wHermite[o]/sqrt(alpha) + exp(-alpha)*q;
    }
  }
}

void integrals( Vector<double>& fourcenter, 
                Matrix<double>& H1,
                Matrix<double>& S,
                Matrix<double>& nucleix,
                Vector<int>& nucleiZ,
                Matrix<double>& basisx,
                Matrix<int>&    basisn,
                Vector<double>& basisa )
{
  double small = 1.e-9;
  int Nbasis = basisa.size();
  int N2 = (Nbasis*(Nbasis+1))/2;
  int N4 = (N2*(N2+1))/2;
  int Nnuclei = nucleiZ.size();

  // the H1 terms and overlap
  Vector<double> xa(3);

  for( int mu=0; mu<Nbasis; mu++ ) {
    double amu = basisa[mu];
    const Vector<double>& xmu = basisx[mu];
    for( int lambda=0; lambda<=mu; lambda++ ) {
      double alambda = basisa[lambda];
      const Vector<double>& xlambda = basisx[lambda];

      double alpha = amu+alambda;
      double rho = amu*alambda/alpha;
      double dxsq=0;
      for( int d=0; d<3; d++ ) {
        double tmp = xmu[d]-xlambda[d];
        dxsq += tmp*tmp;
        xa[d] = (amu*xmu[d] + alambda*xlambda[d])/alpha;
      }

      // overlap:  this is the S-orbital result
      double olap = M_PI/alpha;
      olap *= sqrt(olap)*exp(-rho*dxsq);

      // McMurchie Davidson --- modify "S" overlap to generate other L states
      double md_olap = 1;
      for( int dir=0; dir<3; dir++ ) {
        int nmu = basisn[mu][dir];
        int nlambda = basisn[lambda][dir];
        double d_a_mu = xa[dir] - xmu[dir];
        double d_m_lambda = xa[dir] - xlambda[dir];
        double d = MD(alpha,nmu,nlambda,d_a_mu,d_m_lambda);
        md_olap *= d;
      }
      md_olap *= olap;
      S[mu][lambda] = S[lambda][mu] = md_olap;

      cout << "<" << mu << "|" << lambda << ">=" << md_olap << endl;

      // if just S, KE is olap*rho*(3 - 2*rho*dxsq);

      Vector<double> td(3);
      td[0]=td[1]=td[2]=1;
      for( int dir=0; dir<3; dir++ ) {
        int nmu = basisn[mu][dir];
        int nlambda = basisn[lambda][dir];
        double d_a_mu = xa[dir] - xmu[dir];
        double d_m_lambda = xa[dir] - xlambda[dir];
        double tdd = 0;
        tdd += 2.*amu*alambda*MD( alpha, nmu+1, nlambda+1, d_a_mu, d_m_lambda);
        if( nmu > 0 ) {
          tdd -= nmu*alambda*MD( alpha, nmu-1, nlambda+1, d_a_mu, d_m_lambda );
        }
        if( nlambda > 0 ) {
          tdd -= nlambda*amu*MD( alpha, nmu+1, nlambda-1, d_a_mu, d_m_lambda );
        }
        if( nmu > 0 && nlambda > 0 ) {
          tdd += 0.5*nmu*nlambda*MD( alpha, nmu-1, nlambda-1, d_a_mu, d_m_lambda );
        }
        td[dir] *= tdd;
        double d = MD( alpha, nmu, nlambda, d_a_mu, d_m_lambda );
        td[(dir+1)%3] *= d;
        td[(dir+2)%3] *= d;
      }
      double md_ke = 0;
      for( int dir=0; dir<3; dir++ ) md_ke += td[dir];
      md_ke *= olap;
  
      // KE
      cout << "<" << mu << "|KE|" << lambda << ">=" << md_ke << endl;

      H1[mu][lambda] = H1[lambda][mu] =  md_ke;



      Vector<double> rys_t(4), rys_W(4);
      double total_rys_nuc=0;
      for( int k=0; k<Nnuclei; k++ ) {
        double rys_nuc = 0;
        const Vector<double>& xn = nucleix[k];
        int Z = nucleiZ[k];

        double xaisq=0;
        double xmlsq=0;
        for( int dir=0; dir<3; dir++ ) {
          double tmp = xmu[dir]-xlambda[dir];
          xmlsq += tmp*tmp;
          tmp = xa[dir] - xn[dir];
          xaisq += tmp*tmp;
        }
        Rys( alpha*xaisq, rys_t, rys_W );
        for( int root=0; root<4; root++ ) {
          double tsq = rys_t[root];
          double P = 1;
          for( int dir=0; dir<3; dir++ )  {
            int nmu = basisn[mu][dir];
            int nlam = basisn[lambda][dir];
            double A = Rys_A( nmu, nlam, tsq, alpha,
		xmu[dir], xlambda[dir], xa[dir], xn[dir] );
            P *= A;
          }
          rys_nuc += P*rys_W[root];
        }
        rys_nuc *= 2*M_PI*exp(-amu*alambda*xmlsq/alpha)/alpha;
        total_rys_nuc -= Z*rys_nuc;
      }

      // nuclear
      double s_nuc_tot=0;
      for( int k=0; k<Nnuclei; k++ ) {
        int Z = nucleiZ[k];
        const Vector<double>& xn = nucleix[k];

        double tmp, r = 0;
        for( int d=0; d<3; d++ ) {
          tmp = xn[d] - xa[d];
          r += tmp*tmp;
        }
        tmp = alpha*r/3;
        if( tmp < small ) {
          // treat erf(x)/x in limit x-> 0
          tmp = 1.-tmp;
          tmp *= 2.*sqrt(alpha/M_PI);
        } else {
          r = sqrt(r);
          tmp = erf(sqrt(alpha)*r)/r;
        }
        s_nuc_tot -= Z*olap*tmp;
      }

      H1[mu][lambda] += total_rys_nuc;
      if( mu!=lambda ) H1[lambda][mu] += total_rys_nuc;

      cout << "<" << mu << "|NUC|" << lambda << ">=" << total_rys_nuc << endl;

    }
  }

  // the four-center terms
  // packed in funny way to exploit symmetry
  double p3 = M_PI*M_PI*M_PI;

  int mu=0, nu=0, lambda=0, sigma=0, eone=0, etwo=0;
  Vector<double> xb(3);
  for( int indx=0; indx<N4; indx++ ) {
    const Vector<double>& xmu     = basisx[mu];
    const Vector<double>& xnu     = basisx[nu];
    const Vector<double>& xlambda = basisx[lambda];
    const Vector<double>& xsigma  = basisx[sigma];

    double amu     = basisa[mu];
    double anu     = basisa[nu];
    double alambda = basisa[lambda];
    double asigma  = basisa[sigma];

    double alpha = amu+alambda;
    double beta  = anu+asigma;

    double tmp, s_4c, dxaasq=0, dxbbsq=0, dxabsq=0;
    for( int d=0; d<3; d++ ) {
      xa[d] = (amu*xmu[d] + alambda*xlambda[d])/alpha;
      xb[d] = (anu*xnu[d] + asigma*xsigma[d])/beta;
      tmp = xmu[d] - xlambda[d];
      dxaasq += tmp*tmp;
      tmp = xnu[d] - xsigma[d];
      dxbbsq += tmp*tmp;
      tmp = xa[d] - xb[d];
      dxabsq += tmp*tmp;
    }
    double Ca = exp( - amu*alambda*dxaasq/alpha );
    double Cb = exp( - anu*asigma*dxbbsq/beta );
    double c = alpha*beta/(alpha+beta);
    tmp = dxabsq*c/3.;
    if( tmp < small ) {
      s_4c = 1.-tmp;
      s_4c *= 2*sqrt(c/M_PI);
    } else {
     tmp = sqrt(dxabsq);
     s_4c = erf(sqrt(c)*tmp)/tmp;
    }
    s_4c *= p3*Ca*Cb/(alpha*beta*sqrt(alpha*beta));

// rys version
    double rys_rep = 0;
    Vector<double> rys_t(4), rys_W(4);
    Rys( alpha*beta*dxabsq/(alpha+beta), rys_t, rys_W );
    for( int root=0; root<4; root++ ) {
      double tsq = rys_t[root];
      double P = 1;
      for( int dir=0; dir<3; dir++ ) {
        int nmu        = basisn[mu][dir];
        int nlambda    = basisn[lambda][dir];
        int nnu        = basisn[nu][dir];
        int nsigma     = basisn[sigma][dir];
        P *= RDK_Gxfer( nmu, nnu, nlambda, nsigma, alpha, beta, tsq,
             xa[dir], xb[dir], xmu[dir], xnu[dir], xlambda[dir], xsigma[dir] );
      }
      rys_rep += P*rys_W[root];
    }
    rys_rep *= 2*sqrt(M_PI)*M_PI*M_PI*Ca*Cb;
    rys_rep /= alpha*beta*sqrt(alpha+beta);
    fourcenter[indx] = rys_rep;

    cout << "<" << mu << "," << nu << "|1/r|" << lambda << "," << sigma << ">="
         << rys_rep << endl;


    if( eone == etwo ) {
      etwo = nu = sigma = 0;
      if( lambda == mu ) {
         mu++;
         lambda=0;
       } else {
         lambda++;
       }
       eone++;
    } else {
      if( sigma == nu ) {
        nu++;
        sigma=0;
      } else {
        sigma++;
      }
      etwo++;
    }
  }
}

void orthogonalize( Matrix<double>& occOrbCoef,
                    Matrix<double>& U,
                    Vector<double>& sigma )
{
  int Nbasis = occOrbCoef.columns();
  int Npairs = occOrbCoef.rows();
  Matrix<double> tmp(Npairs,Nbasis);

  for( int i=0; i<Npairs; i++ ) {
    for( int j=0; j<Nbasis; j++ ) {
      tmp[i][j] = 0;
      for( int k=0; k<Nbasis; k++ )
        tmp[i][j] += sigma[j]*U[k][j]*occOrbCoef[i][k];
    }
  }

  double d;
  for( int i=0; i<Npairs; i++ ) {
    for( int j=0; j<i; j++ ) {
      d = 0;
      for( int k=0; k<Nbasis; k++ )
        d += tmp[i][k]*tmp[j][k];
      for( int k=0; k<Nbasis; k++ )
        tmp[i][k] -= d*tmp[j][k];
    }
    d = 0;
    for( int k=0; k<Nbasis; k++ )
      d += tmp[i][k]*tmp[i][k];
    d = 1./sqrt(d);
    for( int k=0; k<Nbasis; k++ )
      tmp[i][k] *= d;
  }

  for( int i=0; i<Npairs; i++ ) {
    for( int j=0; j<Nbasis; j++ ) {
      occOrbCoef[i][j] = 0;
      for( int k=0; k<Nbasis; k++ )
        occOrbCoef[i][j] += U[j][k]*tmp[i][k]/sigma[k];
    }
  }
}

void solveH2O( double angle)
{
  int Nbasis  = 24;
  int Nnuclei = 3;
  int Npairs = 5;
  int N2 = (Nbasis*(Nbasis+1))/2;
  int N4 = (N2*(N2+1))/2;
  Vector<double> fourcenter(N4);
  Matrix<double> H1(Nbasis,Nbasis);
  Matrix<double> S(Nbasis,Nbasis);
  Matrix<double> nucleix(Nnuclei,3);
  Vector<int>    nucleiZ(Nnuclei);
  Matrix<double> basisx(Nbasis,3);
  Matrix<int>    basisn(Nbasis,3);
  Vector<double> basisa(Nbasis);

  Matrix<double> J(Nbasis,Nbasis);
  Matrix<double> K(Nbasis,Nbasis);

  double Etot=0, deltaE, deltaC;

  // oxygen
  nucleix[0][0] = nucleix[0][1] = nucleix[0][2] = 0;
  nucleiZ[0] = 8;
  // hydrogen
  double rad = 1.811096;
  nucleix[1][0] = rad*cos(angle/2);
  nucleix[1][1] = rad*sin(angle/2);
  nucleix[1][2] = 0;
  nucleiZ[1] = 1;
  // hydrogen
  nucleix[2][0] = rad*cos(angle/2);
  nucleix[2][1] = -rad*sin(angle/2);
  nucleix[2][2] = 0;
  nucleiZ[2] = 1;


  // STO-3G components for hydrogen 1
  for( int b=0; b<3; b++ ) {
    for( int i=0; i<3; i++ ) {
      basisx[b][i] = nucleix[1][i];
      basisn[b][i] = 0;
    }
  }
  basisa[0] = 0.16885540;
  basisa[1] = 0.62391373;
  basisa[2] = 3.42525091;

  // STO-3G components for hydrogen 2
  for( int b=3; b<6; b++ ) {
    for( int i=0; i<3; i++ ) {
      basisx[b][i] = nucleix[2][i];
      basisn[b][i] = 0;
    }
  }
  basisa[3] = 0.16885540;
  basisa[4] = 0.62391373;
  basisa[5] = 3.42525091;

  // STO-3G components for oxygen
  for( int b=6; b<21; b++ ) {
    for( int i=0; i<3; i++ ) {
      basisx[b][i] = nucleix[0][i];
    }
  }
  // S
  for( int b=6; b<15; b++ ) {
    for( int i=0; i<3; i++ ) {
      basisn[b][i] = 0;
    }
  }
  // Px
  for( int b=15; b<18; b++ ) {
    basisn[b][0] = 1; basisn[b][1] = basisn[b][2] = 0;
  }
  // Py
  for( int b=18; b<21; b++ ) {
    basisn[b][1] = 1; basisn[b][0] = basisn[b][2] = 0;
  }
  // Pz
  for( int b=21; b<24; b++ ) {
    basisn[b][2] = 1; basisn[b][0] = basisn[b][1] = 0;
  }
  // 1S
  basisa[6] = 0.16885540;
  basisa[7] = 0.62391373;
  basisa[8] =  3.42525091;
  // 2S
  basisa[9] = 0.3803890;
  basisa[10] = 1.1695961;
  basisa[11] = 5.0331513;
  basisa[12] = 6.4436083;
  basisa[13] = 23.808861;
  basisa[14] = 130.70932;
  // 2P
  basisa[15] = 0.3803890;
  basisa[16] = 1.1695961;
  basisa[17] = 5.0331513;
  // 2P
  basisa[18] = 0.3803890;
  basisa[19] = 1.1695961;
  basisa[20] = 5.0331513;
  // 2P
  basisa[21] = 0.3803890;
  basisa[22] = 1.1695961;
  basisa[23] = 5.0331513;

  // initialize occupied Orbital coeffients
  Matrix<double> occOrbCoef(Npairs,Nbasis);
  Matrix<double> newOccOrbCoef(Npairs,Nbasis);

  for( int p=0; p<Npairs; p++ )
    for( int i=0; i<Nbasis; i++ ) occOrbCoef[p][i] = 0;
  // 1S of O
  occOrbCoef[0][7] = 1;
  // 2S of O
  occOrbCoef[1][10] = 1;
  // Px-H 
  occOrbCoef[2][13] = 0.1;
  occOrbCoef[2][15] = 0.5;
  occOrbCoef[2][1]  = 0.2;
  // Px-H 
  occOrbCoef[3][13] = 0.1;
  occOrbCoef[3][15] = -0.5;
  occOrbCoef[3][4]  = 0.2;
  // Pz 
  occOrbCoef[4][18] = 1;

  integrals( fourcenter, H1, S, nucleix, nucleiZ, basisx, basisn, basisa );

  Matrix<double> U(Nbasis,Nbasis);
  Matrix<double> VT(Nbasis,Nbasis);
  Vector<double> sigma(Nbasis);
  Matrix<double> Scpy(Nbasis,Nbasis);
  for( int i=0; i<Nbasis; i++ ) {
    for( int j=0; j<Nbasis; j++ ) Scpy[i][j] = S[i][j];
  }
  SVD( S, U, VT, sigma );
  for( int i=0; i<Nbasis; i++ ) sigma[i] = sqrt(sigma[i]);

  Matrix<double> P(Nbasis,Nbasis);
  Matrix<double> F(Nbasis,Nbasis);
  Matrix<double> X(Nbasis,Nbasis);
  Vector<double> lambda(Nbasis);
  Vector<int> perm(Nbasis);

  int iteration = 0;
  double Etotal;
  do {
    ++iteration;
    cout << "iteration " << iteration << endl;

    // needed if limiting was employed, or for bad IC
    orthogonalize( occOrbCoef, U, sigma );

    // P matrix
    for( int i=0; i<Nbasis; i++ ) {
      for( int j=0; j<=i; j++ ) {
        double  t = 0;
        for( int k=0; k<Npairs; k++ ) {
          t += occOrbCoef[k][i]*occOrbCoef[k][j];
        }
        P[i][j] = P[j][i] = t;
      }
    }


    // initialize the Fock matrix
    for( int i=0; i<Nbasis; i++ )
      for( int j=0; j<Nbasis; j++ ) F[i][j] = H1[i][j];

    for( int i=0; i<Nbasis; i++ )
      for( int j=0; j<Nbasis; j++ ) J[i][j] = K[i][j] = 0;

    int l1=0, l2=0, r1=0, r2=0, eone=0, etwo=0;
    for( int indx=0; indx<N4; indx++ ) {
      double fc = fourcenter[indx];

      F[l1][r1] += 2*fc*P[l2][r2];
      F[l1][r2] -=   fc*P[l2][r1];

      J[l1][r1] += fc*P[l2][r2];
      K[l1][r2] += fc*P[l2][r1];

      if( l1!=r1 ) {
        F[r1][l1] += 2*fc*P[l2][r2];
        F[r1][r2] -=   fc*P[l2][l1];

        J[r1][l1] += fc*P[l2][r2];
        K[r1][r2] += fc*P[l2][l1];
      }

      if( l2!=r2 ) {
        F[l1][r1] += 2*fc*P[r2][l2];
        F[l1][l2] -=   fc*P[r2][r1];

        J[l1][r1] += fc*P[r2][l2];
        K[l1][l2] += fc*P[r2][r1];
      }

      if( l1!=r1 && l2!=r2 ) {
        F[r1][l1] += 2*fc*P[r2][l2];
        F[r1][l2] -=   fc*P[r2][l1];

        J[r1][l1] += fc*P[r2][l2];
        K[r1][l2] += fc*P[r2][l1];
      }

      if( eone != etwo ) {

        F[l2][r2] += 2*fc*P[l1][r1];
        F[l2][r1] -=   fc*P[l1][r2];

        J[l2][r2] += fc*P[l1][r1];
        K[l2][r1] += fc*P[l1][r2];

        if( l2!=r2 ) {
          F[r2][l2] += 2*fc*P[l1][r1];
          F[r2][r1] -=   fc*P[l1][l2];

          J[r2][l2] += fc*P[l1][r1];
          K[r2][r1] += fc*P[l1][l2];
        }

        if( l1!=r1 ) {
          F[l2][r2] += 2*fc*P[r1][l1];
          F[l2][l1] -=   fc*P[r1][r2];

          J[l2][r2] += fc*P[r1][l1];
          K[l2][l1] += fc*P[r1][r2];
        }

        if( l2!=r2 && l1!=r1 ) {
          F[r2][l2] += 2*fc*P[r1][l1];
          F[r2][l1] -=   fc*P[r1][l2];

          J[r2][l2] += fc*P[r1][l1];
          K[r2][l1] += fc*P[r1][l2];
        }
      }

      if( eone == etwo ) {
        etwo = l2 = r2 = 0;
        if( r1 == l1 ) {
           l1++;
           r1=0;
         } else {
           r1++;
         }
         eone++;
      } else {
        if( r2 == l2 ) {
          l2++;
          r2=0;
        } else {
          r2++;
        }
        etwo++;
      }
    }

    // evaluate total energy
    double Ee = 0, Enn = 0;
    for( int i=0; i<Nbasis; i++ ) {
      for( int j=0; j<Nbasis; j++ ) {
         Ee += P[i][j]*(F[i][j]+H1[i][j]);
      }
    }
    Ee /= 2;
    for( int i=0; i<Nnuclei; i++ ) {
      const Vector<double>& xi = nucleix[i];
      cout << "particle " << i << " " << xi << endl;
      int zi = nucleiZ[i];
      for( int j=0; j<i; j++ ) {
        const Vector<double>& xj = nucleix[j];
        int zj = nucleiZ[j];
        double r = 0;
        for( int dir=0; dir<3; dir++ ) 
          r += (xi[dir]-xj[dir])*(xi[dir]-xj[dir]);
        cout << "radius " << i << " " << j << " " << r << endl;
        Enn += zi*zj/sqrt(r);
      }
    }

    // F C = E S C
    // F C = E U sigma UT C
    // 1/sqrt(sigma) UT F U 1/sqrt(sigma) sqrt(sigma) UT C = E sqrt(sigma) UT C

    // replace VT with F U/sqrt(sigma)
    for( int i=0; i<Nbasis; i++ ) {
      for( int j=0; j<Nbasis; j++ ) {
        VT[i][j] = 0;
        for( int k=0; k<Nbasis; k++ ) VT[i][j] += F[i][k]*U[k][j]/sigma[j];
      }
    }

    // replace F with 1/sqrt(sigma) UT VT
    for( int i=0; i<Nbasis; i++ ) {
      for( int j=0; j<Nbasis; j++ ) {
        F[i][j] = 0;
        for( int k=0; k<Nbasis; k++ ) F[i][j] += U[k][i]*VT[k][j]/sigma[i];
      }
    }

    // eigenvalue
    QReig( F, X, lambda );

    // X = sqrt(sigma) UT C
    // C = U 1/sqrt(sigma) X

    // but collect only lowest energy pairs
    hqsort0( lambda, perm );

    double Eorb = 0;
    for( int i=0; i<Npairs; i++ ) Eorb += 2*lambda[Nbasis-1-i];
    double Ej = 0, Ek = 0;
    for( int p=0; p<Nbasis; p++ ) {
       for( int q=0; q<Nbasis; q++ ) {
          Ej  += P[p][q]*J[p][q];
          Ek  += P[p][q]*K[p][q];
       }
    }
    Etotal = Eorb-(2*Ej-Ek) + Enn;
    cout << "Eorb " << Eorb << endl;
    cout << "Enuc " << Enn  << endl;
    cout << "EJ   " << Ej  << endl;
    cout << "EK   " << Ek  << endl;
    cout << angle << " test  " << Etotal  << endl;

    double Etmp = Eorb-(2*Ej-Ek) + Enn;
    deltaE = fabs(Etmp-Etot);
    Etot = Etmp;
    
    for( int i=0; i<Npairs; i++ ) {
      for( int j=0; j<Nbasis; j++ ) {
        newOccOrbCoef[i][j] = 0;
        for( int k=0; k<Nbasis; k++ )
          newOccOrbCoef[i][j] += U[j][k]*X[k][perm[Nbasis-1-i]]/sigma[k];
      }
    }

    // relax
    double f = 0.0;
    deltaC = 0;
    for( int i=0; i<Npairs; i++ ) {
      for( int j=0; j<Nbasis; j++ ) {
        double tmp = (1-f)*newOccOrbCoef[i][j] + f*occOrbCoef[i][j];
        double diff = occOrbCoef[i][j] - tmp;
        deltaC += diff*diff;
        occOrbCoef[i][j] = tmp;
      }
    }
    deltaC = sqrt(deltaC);

  } while( (iteration < 2) || deltaE > 1.e-4  || deltaC > 1.e-4 );
  cout << angle*180/M_PI << " converged  " << Etotal  << endl;
}

int main()
{
  cout << scientific << setprecision(14);
  for( int m=-5; m<=5; m++ ) {
   double ang = 104.44993 + m;
   ang *= M_PI/180.;
   solveH2O( ang );
  }
}
