// G. Miller, Numerical Analysis for Engineers and Scientists,
// Cambridge University Press, Cambridge, 2014.
// http://www.cambridge.org/9781107021082

#include <iostream>
#include <math.h>
#include "Matrix.H"
#include "SVD.H"
using namespace std;

#ifndef MAX
#define MAX(A,B) (((A)>(B))?(A):(B))
#endif

#ifndef MIN
#define MIN(A,B) (((A)<(B))?(A):(B))
#endif

void makeHilbert( int n, Matrix<double>& A, Vector<double>& b )
{
  for( int i=0; i<n; i++ ) {
    b[i] = 0;
    for( int j=0; j<n; j++ ) {
      double t = 1./(i+j+1.);
      b[i] += t;
      A[i][j] = t;
    }
  }
}

int main()
{
  Matrix<double> A(20,20), U(20,20), VT(20,20);
  Vector<double> b(20),    bcpy(20), sigma(20), UTb(20);

  for( int n=5; n<=20; n+=5 ) {
    cout << "n=" << n << endl;
    A.resize(n,n);  U.resize(n,n);  VT.resize(n,n);
    b.resize(n);    bcpy.resize(n); sigma.resize(n); UTb.resize(n);

    // make test matrix and rhs, make a copy of rhs
    makeHilbert(n, A, b);
    for( int i=0; i<n; i++ ) bcpy[i] = b[i];

    // perform SVD
    SVD( A, U, VT, sigma );

    // min and max singular values; condition number
    double smax = sigma[0], smin = sigma[0];
    for( int i=1; i<A.columns(); i++ ) {
      smax = MAX(smax,sigma[i]);
      smin = MIN(smin,sigma[i]);
    }
    cout << "  condition number in L2 " << smax/smin << endl;

    for( double cutoff = 1.e-20; cutoff <= 1.e-10; cutoff *= 1.e10 ) {
      cout << "cutoff " << cutoff << endl;

      // solve U sigma VT x = b;   x = V 1/sigma UT b
      for( int i=0; i<n; i++ ) {
        UTb[i] = 0;
        for( int j=0; j<n; j++ ) UTb[i] += U[j][i]*bcpy[j];
      }
      // divide by sigma only where sigma is greater than cutoff
      for( int i=0; i<n; i++ ) UTb[i] *= ((sigma[i]>=cutoff*smax) ? (1/sigma[i]) : 0 );
      for( int i=0; i<n; i++ ) {
         b[i] = 0;
         for( int j=0; j<n; j++ ) b[i] += VT[j][i]*UTb[j];
      }
      double norm = 0;
      for( int i=0; i<n; i++ ) norm += (b[i]-1.)*(b[i]-1.);
      norm = sqrt(norm);
      cout << "    norm = " << norm << endl;
    }
  }
}
