// G. Miller, Numerical Analysis for Engineers and Scientists,
// Cambridge University Press, Cambridge, 2014.
// http://www.cambridge.org/9781107021082

#include <iostream>
#include <iomanip>
#include <math.h>
#include "Matrix.H"
#include "Householder.H"
#include "HQsort.H"
using namespace std;

#ifndef MAX
#define MAX(A,B) (((A)>(B))?(A):(B))
#endif

// Replace A with composite of R (without diagonal)
// with columns w stored on and below diagonal.
// Diagonal of R goes into "diag".
// Return 0 on success, -1 if there's a problem (singular)

int Householder( Matrix<double>& A, Vector<int>& Pl, Vector<int>& Pr,
                 Vector<double>& diag, QRPivotType type )
{
  int n_eqn = A.rows();
  int n_var = A.columns();
  assert( n_eqn >= n_var );
  assert( diag.size() >= n_var );

  int    i, j, l, s=0;
  double K, len, dot;
  
  // initialize the permutation
  for( int i=0; i<n_eqn; i++ ) Pl[i] = i;
  for( int i=0; i<n_var; i++ ) Pr[i] = i;

  // if Bjorck pivoting, first sort so that row inf-norm is ordered
  if( type == BJORCK ) {
    Vector<double> norm(n_eqn);
    for( int i=0; i<n_eqn; i++ ) {
      double m = fabs(A[i][0]);
      for( int j=1; j<n_var; j++ ) m = MAX(m,fabs(A[i][j]));
      norm[i] = m;
    }

    // sort with customized Quicksort.
    hqsort( norm, Pl, 0, n_eqn-1 );

    // rearrange rows
    for( int i=n_eqn-1; i>=0; i-- ) {
      int k = Pl[i];
      while( k > i ) k = Pl[k];
      if( k != i ) {
        for( int j=0; j<n_var; j++ ) {
          double d=A[k][j]; A[k][j]=A[i][j]; A[i][j]=d;
        }
      }
    }
  }

  // for each column of A
  for( j=0; j<n_var; j++ ) {

    if( type != NOTHING ) {
      // perform column pivoting
      int bestcol    = j-1;
      double bestval = -1;
      for( int jj=j; jj<n_var; jj++ ) {
        double val = 0;
        for( int k=j; k<n_eqn; k++ ) val += A[k][jj]*A[k][jj];
        if( val > bestval ) {
          bestval = val;
          bestcol = jj;
        }
      }
      if( bestcol != j ) {
        // column pivot
        int p = Pr[j]; Pr[j]=Pr[bestcol]; Pr[bestcol]=p;
        for( int kk=0; kk<n_eqn; kk++ ) {
          double d = A[kk][j]; A[kk][j]=A[kk][bestcol]; A[kk][bestcol]=d;
        }
      }
    }

    if( type == POWELLREID ) {
      // perform row pivoting
      int bestrow = j-1;
      double bestval = -1;
      for( int jj=j; jj<n_eqn; jj++ ) {
        double val = fabs(A[jj][j]);
        if( val > bestval ) {
          bestval = val;
          bestrow = jj;
        }
      }
      if( bestrow != j ) {
        // row pivot
        int p = Pl[j]; Pl[j]=Pl[bestrow]; Pl[bestrow]=p;
        for( int kk=0; kk<n_var; kk++ ) {
          double d = A[j][kk]; A[j][kk]=A[bestrow][kk]; A[bestrow][kk]=d;
        }
      }
    }

    // here begins the reduction:  everything above this is pivoting overhead.
    // find the length squared of the vector we want to transform into (K e_1)
    // omit the first entry ...
    len = 0.0;
    for( i=j+1; i<n_eqn; i++ )
      len += A[i][j]*A[i][j];

    // make K -- sign for numerical stability
    K = sqrt( len + A[j][j]*A[j][j] );
    if( A[j][j] > 0 ) K = -K;

    // K is the new diagonal entry
    diag[j] = K;
    if( !diag[j] ) s= -1; // singular

    // replace the column of A with w = a - K e_1
    // not yet normalized.
    A[j][j] -= K;

    // now update len to be length of w
    len = sqrt( len + A[j][j]*A[j][j] );
    if( len <= 0 ) {
      // this means that the corresponding part of Q is part of I,
      // w is zero.  diag is zero.
      continue;
    }

    // normalize so |w|_2 =1
    for( i=j; i<n_eqn; i++ )
      A[i][j] /= len;

    // now A[i][j], i=j to n_eqn is normalized w

    // use w on subsequent columns of A
    // a := a - 2 w (w.a)
    for( l=j+1; l<n_var; l++ ) {

      // make 2(w.a)
      dot = 0.0;
      for( i=j; i<n_eqn; i++ )
        dot += A[i][l]*A[i][j];
      dot *= 2;

      // a := a - dot w
      for( i=j; i<n_eqn; i++ )
        A[i][l] -= dot*A[i][j];
    }

  }
  return s;
}

// Replace b with Q^T b
// Everything needed to make Q is stored in w form in A.

void QTb( const Matrix<double>&A, Vector<double>& b )
{
  int n_var = A.columns();
  int n_eqn = A.rows();
  int    i, j;
  double dot;

  // for each w
  for( j=0; j<n_var; j++ ) {

    // b := b - 2 w (w.b)
    // first make 2*(w.b)
    dot = 0;
    for( i=j; i<n_eqn; i++ )
      dot += A[i][j]*b[i];
    dot *= 2;

    // b := b - w*dot
    for( i=j; i<n_eqn; i++ )
      b[i] -= dot*A[i][j];
  }
}

// replace b  with x, solving Ax=b after  A has been converted
// to QR, with diagonals of R in "diag".  Return L2 norm of residual
// (for overdetermined systems only).

double solveQR( const Matrix<double>& A, const Vector<int>& Pl, const Vector<int>& Pr,
                const Vector<double>& diag, Vector<double>& b )
{
  int n_var = A.columns();
  int n_eqn = A.rows();
  int    i, j;

  // first, make u = Q^T b
  Vector<double> u(n_eqn);
  for( j=0; j<n_eqn; j++ ) u[j] = b[Pl[j]];

  // replace u with QT u
  QTb( A, u );

  double residual = 0;
  for( int i=n_var; i<n_eqn; i++ )
    residual += u[i]*u[i];
  residual = sqrt(residual);

  // now solve Rx=u (save x in b):  back substitution

  for( i=n_var-1; i>=0; i-- ) {
    b[i] = u[i];
    for( j=i+1; j<n_var; j++ )
      b[i] -= A[i][j]*b[j];
    b[i] /= diag[i];
  }

  // and permute
  for( i=n_var-1; i>=0; i-- ) {
    int k = Pr[i];
    while( k < i ) k = Pr[k];
    double d = b[k]; b[k] = b[i]; b[i] = d;
  }

  return residual;
}

// print the QR form of the matrix in readable form.

void printQR( const Matrix<double>&A , const Vector<int>&Pl,
              const Vector<int>&Pr, const Vector<double>& diag )
{
  int n_eqn = A.rows();
  ios_base::fmtflags ff = cout.flags();
  streamsize prec = cout.precision();
  cout << scientific << setprecision(16) << endl;

  cout << "Pl: " << endl << noshowpos << Pl << showpos << endl;

  cout << "Q:" << endl;
  Vector<double> Qrow(n_eqn);
  for( int i=0; i<n_eqn; i++ ) {
    for( int j=0; j<n_eqn; j++ ) Qrow[j] = 0;
    Qrow[i] = 1;
    QTb(A,Qrow);
    for( int j=0; j<n_eqn; j++ )
      cout << " " <<  Qrow[j];
    cout << endl;
  }

  cout << "R:" << endl;
  for( int r=0; r<A.rows(); r++ ) {
    int rmax = A.columns();
    rmax = (rmax > r) ? r : rmax;
    for( int c=0; c<rmax;  c++ )
      cout << " " <<  0.0;
    for( int c=rmax; c<A.columns();  c++ )
      cout << " " <<  ((c==r) ? diag[c] : A[r][c]);
    cout << endl;
  }

  cout << "Pr: " << endl << noshowpos << Pr << endl;

  cout.flags(ff);
  cout << setprecision(prec);
}
