// G. Miller, Numerical Analysis for Engineers and Scientists,
// Cambridge University Press, Cambridge, 2014.
// http://www.cambridge.org/9781107021082

#include <iostream>
#include <iomanip>
#include <math.h>
#include <Matrix.H>
#include <Householder.H>
#include <QR.H>
using namespace std;

int main()
{
  cout << setprecision(10) << scientific;

  // read inputs:
  int npts;
  cin >> npts;
  cout << "number of data points " << npts << endl;

  Vector<double> x(npts), y(npts), X(npts);
  for( int i=0; i<npts; i++ ) {
    cin >> x[i];
    X[i] = x[i];
    cin >> y[i];
  }
  cout << "x input" << endl << x << endl;
  cout << "y input" << endl << y << endl;

  Matrix<double> Sigma(2*npts,2*npts);
  for( int i=0; i<2*npts; i++ ) {
    for( int j=0; j<=i; j++ ) {
      cin >> Sigma[i][j];
      Sigma[j][i] = Sigma[i][j];
    }
  }


  double beta; // nonlinear term mixing
  double zeta; // damping
  cin >> beta;
  cin >> zeta;

  // replace Sigma with Sigma^(-1/2)

  Matrix<double> P(2*npts,2*npts);
  Vector<double> lambda(2*npts);
  QReig( Sigma, P, lambda );

  for( int i=0; i<2*npts; i++ ) {
    for( int j=0; j<2*npts; j++ ) {
      Sigma[i][j] = 0;
      for( int k=0; k<2*npts; k++ ) Sigma[i][j] += P[i][k]*P[j][k]/sqrt(lambda[k]);
    }
  }


  Matrix<double> A(2*npts,npts+2);
  Vector<double> b(2*npts), diag(npts+2);
  Vector<int> Pl(2*npts), Pr(npts+2);

  double dbi;
  int iteration=0;

  double slope = 0;
  double intercept = 0;

  Vector<double> oldb(npts+2);
  for( int i=0; i<npts; i++ ) oldb[i] = x[i];
  oldb[npts] = slope;
  oldb[npts+1] = intercept;

  do {
    ++iteration;

    // solve weighted least squares problem
    for( int i=0; i<npts; i++ ) {
      A[i     ][npts] = A[i     ][npts+1] = 0;
      A[i+npts][npts] = A[i+npts][npts+1] = 0;
      b[i] = b[i+npts] = 0;
      for( int j=0; j<npts; j++ ) {
        A[i     ][j] = Sigma[i     ][j] + slope*beta*Sigma[i     ][j+npts];
        A[i+npts][j] = Sigma[i+npts][j] + slope*beta*Sigma[i+npts][j+npts];
        A[i     ][npts  ] += (1.-beta)*Sigma[i     ][j+npts]*X[j];
        A[i+npts][npts  ] += (1.-beta)*Sigma[i+npts][j+npts]*X[j];
        A[i     ][npts+1] += Sigma[i     ][j+npts];
        A[i+npts][npts+1] += Sigma[i+npts][j+npts];
        b[i     ] += Sigma[i     ][j]*x[j] + Sigma[i     ][j+npts]*y[j];
        b[i+npts] += Sigma[i+npts][j]*x[j] + Sigma[i+npts][j+npts]*y[j];
      }
    }

    Householder( A, Pl, Pr, diag, BJORCK );
    solveQR( A, Pl, Pr, diag, b );

    // measure change (L inf) before relaxation
    dbi = 0;
    for( int i=0; i<npts+2; i++ ) {
      double t = fabs(b[i] - oldb[i]);
      dbi = max(dbi,t);
    }

    // relax
    for( int i=0; i<npts+2; i++ ) {
      b[i] = (1.-zeta)*b[i] + zeta*oldb[i];
      oldb[i] = b[i];
    }
      
    // replace X with new value

    slope = b[npts];
    intercept = b[npts+1];
    for( int i=0; i<npts; i++ ) {
      X[i] = b[i];
    }
            
    cout << iteration << " change dbi=" << dbi
                      << " m=" << slope 
                      << " b=" << intercept << endl;
  } while( dbi > 1.e-6 && iteration<100 );

  cout << "after " << iteration << " iterations " << endl;

  // now find covariance

  cout << "slope m      " << slope << endl;
  cout << "intercept b  " << intercept << endl;

  for( int i=0; i<npts; i++ ) {
    A[i     ][npts] = A[i     ][npts+1] = 0;
    A[i+npts][npts] = A[i+npts][npts+1] = 0;
    for( int j=0; j<npts; j++ ) {
      A[i     ][j] = Sigma[i     ][j] + slope*Sigma[i     ][j+npts];
      A[i+npts][j] = Sigma[i+npts][j] + slope*Sigma[i+npts][j+npts];
      A[i     ][npts  ] += Sigma[i     ][j+npts]*X[j];
      A[i     ][npts+1] += Sigma[i     ][j+npts];
      A[i+npts][npts  ] += Sigma[i+npts][j+npts]*X[j];
      A[i+npts][npts+1] += Sigma[i+npts][j+npts];
    }
  }
  Householder( A, Pl, Pr, diag, NOTHING );
  // make Q
  Matrix<double> Q(2*npts,2*npts);
  for( int i=0; i<2*npts; i++ ) {
    Vector<double>& v = Q[i];
    for( int j=0; j<2*npts; j++ ) v[j] = 0;
    v[i] = 1;
    QTb(A,v);
  }
  Matrix<double> Bdag(npts+2,2*npts);
  // solve R Bdag = QT for Bdag
  for( int i=0; i<2*npts; i++ ) {
    for( int j=npts+1; j>=0; j-- ) {
      double t = Q[i][j];
      for( int k=j+1; k<npts+2; k++ ) 
        t -= A[j][k]*Bdag[k][i];
      Bdag[j][i] = t/diag[j];
    }
  }

  double varm=0, varb=0, covmb=0;

  for( int i=0;  i<2*npts; i++ ) {
    varm  += Bdag[npts  ][i]*Bdag[npts  ][i];
    varb  += Bdag[npts+1][i]*Bdag[npts+1][i];
    covmb += Bdag[npts  ][i]*Bdag[npts+1][i];
  }
  cout << "var m " << varm << endl;
  cout << "sig m " << sqrt(varm) << endl;
  cout << "var b " << varb << endl;
  cout << "sig b " << sqrt(varb) << endl;
  cout << "covar mb " << covmb << endl;
}
