// G. Miller, Numerical Analysis for Engineers and Scientists,
// Cambridge University Press, Cambridge, 2014.
// http://www.cambridge.org/9781107021082

#ifndef __Matrix_H__
#define __Matrix_H__
#include <iostream>
#include <assert.h>
#include "RCP.H"

template <class T> class Vector
{

  public:

    Vector()
    {
      m_size = m_alloc_size = 0;
    }

    Vector(int a_size)
    {
      m_size = m_alloc_size = 0;
      resize( a_size );
    }

    ~Vector()
    {
    }

    void resize( int a_size )
    {
      assert( a_size>=0 );

      if( a_size == m_size ) return;

      if( a_size < m_alloc_size ) {
        m_size = a_size;
        return;
      }

      // The following few lines are why we need the RCP class.
      // if "T" has allocated memory, then "spare[]=m_data[]"
      // copies pointers, but the subsequent delete of m_data
      // will remove allocated memory.  "spare" will end up with
      // (formerly) valid pointers that point to freed memory,
      // not allocated memory.

      T* spare = new T[a_size];
      for( int i=0; i<m_size; i++ )
        spare[i] = m_data[i];
      // RCP<> operator= will handle the deletion of any memory
      // allocated to m_data (that is not referenced elsewhere).
      m_data = RCP<T>(spare);

      m_alloc_size = m_size = a_size;
    }

    const T& operator[] (int n) const
    {
      assert( n>=0 );
      assert( n<m_size );
      return m_data[n];
    }

    T& operator[] (int n)
    {
      assert( n>=0 );
      assert( n<m_size );
      return m_data[n];
    }

    int size() const
    {
      return m_size;
    }

    friend std::ostream& operator<< (std::ostream& os, const Vector<T>&v )
    {
      for( int i=0; i<v.m_size; i++ ) {
        os << " " << v[i] << std::endl;
      }
      return os;
    }

    friend std::istream& operator>> (std::istream& is, Vector<T>&v )
    {
      for( int i=0; i<v.m_size; i++ ) {
        is >> v[i];
      }
      return is;
    }
  

  private:

  RCP<T> m_data;
  int  m_size;       // declared size
  int  m_alloc_size; // allocated size
};

template <class T> class Matrix
{
  public:

    Matrix()
    {
      m_rows = m_columns = 0;
    }

    Matrix(int a_rows, int a_columns)
    {
      m_rows = m_columns = 0;
      resize( a_rows, a_columns );
    }

    ~Matrix()
    {
    }
  
    void resize( int a_rows, int a_columns )
    {
      assert( a_rows >= 0 );
      assert( a_columns >= 0 );

      // first, rows
      if( a_rows != m_rows ) {
         m_data.resize( a_rows );
         for( int i=m_rows; i<a_rows; i++ ) {
           m_data[i].resize( m_columns );
           for( int j=0; j<m_columns; j++ ) m_data[i][j] = 0;
         }
      }
      m_rows = a_rows;

      // next, columns
      if( a_columns != m_columns ) {
        for( int i=0; i<m_rows; i++ ) {
           m_data[i].resize( a_columns );
        }
      }
      m_columns = a_columns;
    }

    const Vector<T>& operator[] (int n) const
    {
      assert( n >= 0 );
      assert( n < m_rows );
      return m_data[n];
    }

    Vector<T>& operator[] (int n)
    {
      assert( n >= 0 );
      assert( n < m_rows );
      return m_data[n];
    }

    int rows() const
    {
      return m_rows;
    }

    int columns() const
    {
      return m_columns;
    }

    friend std::ostream& operator<< (std::ostream& os, const Matrix<T>&m )
    {
      for( int i=0; i<m.m_rows; i++ ) {
        for( int j=0; j<m.m_columns; j++ ) {
          os << " " << m[i][j];
        }
        os << std::endl;
      }
      return os;
    }

    friend std::istream& operator>> (std::istream& is, Matrix<T>&m )
    {
      for( int i=0; i<m.m_rows; i++ ) {
        is >> m[i];
      }
      return is;
    }
  
  private:
    
  Vector<Vector<T> > m_data;
  int m_rows;
  int m_columns;
};    
#endif
