// G. Miller, Numerical Analysis for Engineers and Scientists,
// Cambridge University Press, Cambridge, 2014.
// http://www.cambridge.org/9781107021082

#include <math.h>
#include "Matrix.H"
#include "Givens.H"
#include <assert.h>
#include <iostream>

// input:  a symmetric real "A", and "P" is identity matrix.
// output: "A" is tridiagonal, and "P" is the similarity transform.

void makeTridiag(  Matrix<double>& A,
                   Matrix<double>& P )
{
  int N = A.rows();

  // first step -- make tridiagonal.  Do N-2 Householder steps
  Vector<double> w(N);
  for( int c=0; c<N-2; c++ ) {

    // a_i starts 1 below diagonal.

    double len = 0;
    for( int r=c+2; r<N; r++ ) len += A[r][c]*A[r][c];

    if( len != 0 ) {
      double K = sqrt( len + A[c+1][c]*A[c+1][c] );
      if( A[c+1][c] > 0 ) K = -K;

      // make W explicitly
      for( int r=c+1; r<N; r++ ) w[r] = A[r][c];
      w[c+1] -= K;
      len = sqrt( len + w[c+1]*w[c+1] );

      for( int r=c+1; r<N; r++ ) w[r] /= len;

      // now process "A" -- convert column "c" directly, and others numerically
      A[c+1][c] = K;
      for( int r=c+2; r<N; r++ ) A[r][c] = 0;

      for( int c2=c+1; c2<N; c2++ ) {
        // d = w . a
        double d=0;
        for( int r=c+1; r<N; r++ ) d += w[r]*A[r][c2];
        // a -= 2*d*w
        d *= 2;
        for( int r=c+1; r<N; r++ ) {
          A[r][c2] -= d*w[r];
        }
      }

      // Next operate from right -- similarity transformation.
      // same steps as above.

      A[c][c+1] = K;
      for( int r=c+2; r<N; r++ ) A[c][r] = 0;

      for( int c2=c+1; c2<N; c2++ ) {
        double d = 0;
        for( int r=c+1; r<N; r++ ) d += w[r]*A[c2][r];
        d *= 2;
        for( int r=c+1; r<N; r++ ) {
          A[c2][r] -= d*w[r];
        }
      }

      // and change P.
      // at start I A I, then introduce a Q
      //          (I Q) Q A Q (Q I)
      // so P := P Q from right

      for( int c2=0; c2<N; c2++ ) {
        double d = 0;
        for( int r=c+1; r<N; r++ ) d += w[r]*P[c2][r];
        for( int r=c+1; r<N; r++ ) P[c2][r] -= 2.*d*w[r];
      }
    }
  }

  // symmetrize
  for( int i=0; i<N; i++ ) {
    for( int j=0; j<i; j++ )
       A[i][j] = A[j][i] = 0.5*( A[i][j] + A[j][i] );
  }
}

// Perform one QR step assuming a symmetric tridiagonal matrix A.
// The "A" information is represented by a vector of diagonal
// values, and a vector of subdiagonal values.
// "P" collects the Q's and therefore becomes the eigenvectors.

// return "1" if last element of diag is an eigenvalue.

int oneQRstep( Vector<double>& diag,
                Vector<double>& subdiag,
                Matrix<double>& P,
                int             N )
{
  int NP = P.rows();
  double machine = 1e-16;

  if( N == 1 ) return 1;  // A is a scalar

  // check for reducibility:  z will be the starting index
  // for a reduced block.
  int z = N-1;
  do {
    if( z == 0 ) break;
    z--;
    double sum = fabs(diag[z]) + fabs(diag[z+1]);
    if( fabs(subdiag[z]) <= machine*sum ) {
      // the matrix is reducible.  set z to be the
      // "zero" position of last partition.
      ++z;
      break;
    }
  } while(1);

  // if partition contains only 1 diagonal element,
  // no work needs to be done.
  if( z >= N-1 ) return 1;

  double shift = 0;

  double qa = 1;
  double qb = -(diag[N-1]+diag[N-2]);
  double qc =  diag[N-1]*diag[N-2]-subdiag[N-2]*subdiag[N-2];
  double qs = ((qb<0)?-1:1);

  double mu1 = -qs*( fabs(qb)+sqrt(qb*qb-4*qa*qc) )/( 2 * qa );
  double mu2 = -qs*( 2 * qc )/( fabs(qb)+sqrt(qb*qb-4*qa*qc) );
  shift = mu2;
  if( fabs(diag[N-1]-mu1) < fabs(diag[N-1]-mu2) ) shift = mu1;

  // find first Givens factors to eliminate A[z+1][z] using A[z][z]
  double c,s,v;
  Givens( diag[z]-shift, subdiag[z], c, s, v );

  // apply this Givens from left, AND right.  the elements
  // affected are  0,0  0,1  0,2
  //               1,0  1,1  1,2
  //               2,0  2,1
  // or diag[z], subdiag[z], diag[z+1], subdiag[z+1], and
  // "spare" = A[z+2][z] which lies outside tridiagonal form.

  double di = diag[z];
  double dip1 = diag[z+1];
  double sdi = subdiag[z];
  double sdip1=0;
  if( N-z > 2 ) sdip1 = subdiag[z+1];
  double dip2;
  double sdip2=0;

  diag[z  ] = c*(c*(di-dip1) + 2*s*sdi) + dip1;
  diag[z+1] = c*(c*(dip1-di) - 2*s*sdi) + di;
  subdiag[z] = c*(s*(di-dip1) - 2*c*sdi) + sdi;
  if( N-z > 2 ) subdiag[z+1] = -c*sdip1;
  double spare = s*sdip1;

  // and apply the Givens reflection to the eigenvectors;
  for( int j=0; j<NP; j++ ) {
    double u1 = P[j][z  ];
    double u2 = P[j][z+1];
    P[j][z  ] = c*u1 + s*u2;
    P[j][z+1] = v*(u1 + P[j][z]) - u2;
  }

  // select next Givens to eliminate this "spare" value,
  // but operation on right makes a new spare ...

  for( int i=z; i<N-2; i++ ) {

    Givens( subdiag[i], spare, c, s, v );

    dip1 = diag[i+1];
    dip2 = diag[i+2];
    sdi  = subdiag[i];
    sdip1 = subdiag[i+1];
    if( i != N-3 ) sdip2 = subdiag[i+2];

    diag[i+1] = c*(c*(dip1-dip2) + 2*s*sdip1) + dip2;
    diag[i+2] = c*(c*(dip2-dip1) - 2*s*sdip1) + dip1;
    subdiag[i  ] = s*spare + c*sdi;
    subdiag[i+1] = c*(s*(dip1-dip2) - 2*c*sdip1) + sdip1;
    if( i != N-3 ) {
      subdiag[i+2] = -c*sdip2;
      spare = s*sdip2;
    }

    // and apply the Givens reflection to the eigenvectors;
    for( int j=0; j<NP; j++ ) {
      double u1 = P[j][i+1];
      double u2 = P[j][i+2];
      P[j][i+1] = c*u1 + s*u2;
      P[j][i+2] = v*(u1 + P[j][i+1]) - u2;
    }
  }

  return 0;
}

// On input, A is a symmetric real matrix, and P is the identity matrix.
// On output, P is the matrix of eigenvectors and lambda is the vector
// of corresponding eigenvalues.  "A" is modified to tridiagonal form.
// return 0 on success

int QReig( Matrix<double>& A,
           Matrix<double>& P,
           Vector<double>& lambda )
{
  int N=A.rows();
  assert( A.columns() == N );
  assert( P.rows() == N );
  assert( P.columns() == N );
  assert( lambda.size() >= N );

  for( int i=0; i<N; i++ ) {
    for( int j=0; j<N; j++ ) P[i][j] = 0;
    P[i][i] = 1;
  }

  // make tridiagnonal

  makeTridiag( A, P );

  // make a more compact data representation:  save only the diagnonal
  // ans subdiagonal (it has been assumed that A was symmetric).

  Vector<double> subdiag(N-1);

  for( int i=0; i<N;   i++ ) lambda[i]    = A[i][i];
  for( int i=0; i<N-1; i++ ) subdiag[i] = A[i+1][i];

  int curN = N;

  // limit on loop prevents accidentally going forever
  for( int step=0; step<10*N; step++ ) {

    int got = oneQRstep( lambda, subdiag, P, curN );

    if( got ) {
      curN--;
      if( curN == 0 ) return 0;
    }
  }

  // failure to converge ...
  return 1;
}
