// G. Miller, Numerical Analysis for Engineers and Scientists,
// Cambridge University Press, Cambridge, 2014.
// http://www.cambridge.org/9781107021082

#include <iostream>
#include <iomanip>
#include <math.h>
#include "Matrix.H"
#include "QR.H"
#include "Cholesky.H"
#include "AdaptiveSimpson.H"
using namespace std;

// Stieltjes values are global for benefit of integrator
Vector<double> s_delta;
Vector<double> s_gammasq;

// more global info for integrator
int  s_pwr;
bool s_extra_x;

// evaluate polynomial degree "deg" with Stieltjes recursion
double polyfn( double x, int deg )
{
  double pm2=0, pm1=0, p=1;
  for( int i=1; i<=deg; i++ ) {
    pm2 = pm1;
    pm1 = p;
    p *= x-s_delta[i-1];
    if( i!=1 ) p -= s_gammasq[i-2]*pm2;
  }
  return p;
}

// the function to be evaluated for Stieltjes form
double orth_rys_f( double x )
{
  double pi = polyfn(x*x,s_pwr);
  double retval = 2.*pi*pi*exp(-x*x/2);
  if( s_extra_x ) retval *= x*x;
  return retval;
}

void coefficients_from_Stieltjes( int maxorder, double& p0p0 )
{
  s_delta.resize(maxorder);       // in case
  s_gammasq.resize(maxorder-1);

  double oldIpipi;
  for( int i=0; i<maxorder; i++ ) {
    s_pwr   = i;      // global variable passes degree to orth_rys_f
    s_extra_x = false;
    double Ipipi = adaptiveSimpson( 0., 1., 1.e-15, orth_rys_f );
    s_extra_x = true;
    double Ixpipi = adaptiveSimpson( 0., 1., 1.e-15, orth_rys_f );

    if( i == 0 ) p0p0 = Ipipi;
    s_delta[i] = Ixpipi/Ipipi;
    if( i != 0 ) s_gammasq[i-1] = Ipipi/oldIpipi;
    oldIpipi = Ipipi;
  }
}

// global power for benefit of moment integrator
int f_pwr;

// integrand for moment calculation
double moment_rys_f( double x )
{
  return 2*exp(-x*x/2)*pow(x,2.*f_pwr);
}

void coefficients_from_moments( int maxorder, double& p0p0,
                                Vector<double> delta,
                                Vector<double> gamma  )
{
  Matrix<double> M(maxorder+1,maxorder+1);

  for( int i=0; i<maxorder+1; i++ ) {
    for( int j=i; j<maxorder+1; j++ ) {
      f_pwr = i+j;  // global variable passes degree to moment_rys_f
      M[j][i] = M[i][j] = adaptiveSimpson( 0., 1., 1.e-15, moment_rys_f );
    }
  }

  // save (p0,p0) for later
  p0p0 = M[0][0];

  // replace M with tildeL
  Cholesky(M);

  delta.resize(maxorder); // in case
  delta[0] = M[1][0]/M[0][0];
  for( int j=1; j<maxorder; j++ )
    delta[j] = M[j+1][j]/M[j][j] - M[j][j-1]/M[j-1][j-1];

  gamma.resize(maxorder-1); // in case
  for( int j=0; j<maxorder-1; j++ )
    gamma[j] = M[j+1][j+1]/M[j][j];
}

int main()
{
  cout << scientific << setprecision(15);

  int maxorder = 5;
  double p0p0;
  Vector<double> delta(maxorder), gamma(maxorder-1);

  for( int method=0; method<2; method++ ) {
    switch(method) {
      case 0:
        coefficients_from_moments( maxorder, p0p0, delta, gamma );
        cout << "compute Jacobi matrix coefs with moments" << endl;
        break;
      default:
        coefficients_from_Stieltjes( maxorder, p0p0 );
        cout << "compute Jacobi matrix coefs with Stieltjes method" << endl;
        // copy from global array
        for( int i=0; i<maxorder-1; i++ ) {
          delta[i] = s_delta[i];
          gamma[i] = sqrt(s_gammasq[i]);
        }
        delta[maxorder-1] = s_delta[maxorder-1];
        break;
    }
    cout << "delta" << endl << delta << endl;
    cout << "gamma" << endl << gamma << endl;

    for( int order=1; order<=maxorder; order++ ) {

      // J
      Matrix<double> J(order,order);
      for( int i=0; i<order; i++ ) {
        for( int j=0; j<order; j++ ) J[i][j] = 0;
        J[i][i] = delta[i];
      }
      for( int i=0; i<order-1; i++ )
        J[i][i+1] = J[i+1][i] = gamma[i];

      // eigenvalues and vectors
      Matrix<double> V(order,order);
      Vector<double> x(order);

      // make eigensystem
      QReig( J, V, x );

      // weights
      Vector<double> w(order);
      for( int i=0; i<order; i++ ) {
        double len = 0;
        for( int j=0; j<order; j++ ) len += V[j][i]*V[j][i];
        w[i] = (p0p0/len)*V[0][i]*V[0][i];
      }

      cout << "Rys quadrature method with " <<  order << " abscissas:" << endl;
      cout << "   quadrature points" << endl << x << endl;
      cout << "   quadrature weights" << endl << w << endl;
    }
  }
}
