// G. Miller, Numerical Analysis for Engineers and Scientists,
// Cambridge University Press, Cambridge, 2014.
// http://www.cambridge.org/9781107021082

#include <iostream>
#include <fstream>
#include <iomanip>
#include <math.h>
#include "Matrix.H"
#include "QR.H"
#include "AdaptiveSimpson.H"
#include "Householder.H"
#include "HQsort.H"
using namespace std;

fstream io;

// Stieltjes values are global for benefit of integrator
Vector<double> s_delta;
Vector<double> s_gammasq;

// more global info for integrator
int  s_pwr;
bool s_extra_x;
double g_alpha;

// evaluate polynomial degree "deg" with Stieltjes recursion
double polyfn( double x, int deg )
{
  double pm2=0, pm1=0, p=1;
  for( int i=1; i<=deg; i++ ) {
    pm2 = pm1;
    pm1 = p;
    p *= x-s_delta[i-1];
    if( i!=1 ) p -= s_gammasq[i-2]*pm2;
  }
  return p;
}

// the function to be evaluated for Stieltjes form
double orth_rys_f( double x )
{
  double pi = polyfn(x*x,s_pwr);
  double retval = pi*pi*exp(-g_alpha*x*x);
  if( s_extra_x ) retval *= x*x;
  return retval;
}

void coefficients_from_Stieltjes( int maxorder, double& p0p0 )
{
  s_delta.resize(maxorder);       // in case
  s_gammasq.resize(maxorder-1);

  double oldIpipi;
  for( int i=0; i<maxorder; i++ ) {
    s_pwr   = i;      // global variable passes degree to orth_rys_f
    s_extra_x = false;
    double Ipipi = adaptiveSimpson( 0., 1., 1.e-15, orth_rys_f );
    s_extra_x = true;
    double Ixpipi = adaptiveSimpson( 0., 1., 1.e-15, orth_rys_f );

    if( i == 0 ) p0p0 = Ipipi;
    s_delta[i] = Ixpipi/Ipipi;
    if( i != 0 ) s_gammasq[i-1] = Ipipi/oldIpipi;
    oldIpipi = Ipipi;
  }
}

// global power for benefit of moment integrator
int f_pwr;

// integrand for moment calculation
void main2( const int& order, const double& a_alpha, Vector<double>& x, Vector<double>& W )
{

  g_alpha = a_alpha;
  double p0p0;
  Vector<double> delta(order), gamma(order-1);

  coefficients_from_Stieltjes( order, p0p0 );
  // copy from global array
  for( int i=0; i<order-1; i++ ) {
    delta[i] = s_delta[i];
    gamma[i] = sqrt(s_gammasq[i]);
  }
  delta[order-1] = s_delta[order-1];

  // J
  Matrix<double> J(order,order);
  for( int i=0; i<order; i++ ) {
    for( int j=0; j<order; j++ ) J[i][j] = 0;
    J[i][i] = delta[i];
  }
  for( int i=0; i<order-1; i++ )
    J[i][i+1] = J[i+1][i] = gamma[i];

  // eigenvalues and vectors
  Matrix<double> V(order,order);
  QReig( J, V, x );

  // weights
  for( int i=0; i<order; i++ ) {
    double len = 0;
    for( int j=0; j<order; j++ ) len += V[j][i]*V[j][i];
    W[i] = (p0p0/len)*V[0][i]*V[0][i];
  }

  // sort
  Vector<int> perm(order);
  Vector<double> wtmp(order);
  hqsort0( x, perm );
  for( int i=0; i<order; i++ ) wtmp[i] = W[perm[i]];
  for( int i=0; i<order; i++ ) W[i] = wtmp[i];
}


// Evaluate Chebychev poly
double Tn( int n, double x)
{
  if( n==0 ) return 1.0;
  if( n==1 ) return x;
  double T=x, Tm = 1.;
  int k=1;
  do {
    double tmp = T;
    T = 2*x*T - Tm;
    Tm = tmp;
  }while( ++k < n );
  return T;
}

void Chebyfit( const double& a, const double& b, int npts, const string& prefix )
{
  io << showpos << scientific << setprecision(15);
  io << "// a=" << a << " b=" << b << endl;
  int order = 4;
  Vector<double> t(order);
  Vector<double> W(order);

  Matrix<double> Ws(order,npts);
  Matrix<double> ts(order,npts);

  // npts sampled -> roots of T_npts -> fit in 0,...,npts-1

  for( int i=0; i<npts; i++ ) {
    double xk = cos(M_PI*(2.*i+1)/(2.*npts));
    double alpha = (b+a)/2 + xk*(b-a)/2;
    main2( order, alpha, t, W );
    for( int j=0; j<order; j++ ) {
      Ws[j][i] = W[j];
      ts[j][i] = t[j];
    }
  }


  Vector<double> bj(npts);

  for( int wt=0; wt<2; wt++ ) {
    char c = wt ? 'W' : 't';
    for( int o=0; o<order; o++ ) {
      for( int j=0; j<npts; j++ ) bj[j] = 0;
      for( int k=0; k<npts; k++ ) {
        double xk = cos(M_PI*(2.*k+1)/(2.*npts));
        double Tjold=0, Tj=1;
        double val = (wt ? Ws[o][k] : ts[o][k] );
        for( int j=0; j<npts; j++ ) {
          bj[j] += Tj*val;
          if( j==0 ) {
            Tj = xk;
            Tjold = 1;
          } else {
            double foo = Tj;
            Tj = 2*xk*Tj - Tjold;
            Tjold = foo;
          }
        }
      }
      for( int j=0; j<npts; j++ )
        bj[j] *= ((j==0)?1.:2.)/npts;
        io << "static double " << prefix << c << noshowpos << o
           << "[" << npts << showpos << "]={" << endl;
      for( int i=0; i<npts; i++ ) {
        io << bj[i];
        if( i != npts-1 ) io << ", ";
        if( i%3 == 2 && i != npts-1 ) io << endl;
      }
      io << "};" << endl;
    }
  }
}

void LaurentFit( const double& xmin, const double& xmax,
                 int npts, int kmin, int kmax )
{
// Model:  W = wHermite/sqrt(x) + exp(-x)*Qw;  Qw a poly
//         u = rHermite^2/(x-rhermite^2) + exp(-x)*Qu; Qu a poly
// so ... need roots of H_2n = H8


  static double rHermite[4] = {
     2.93063742025724401922350270524,
     1.98165675669584292585463063976,
     1.15719371244678019472076577906,
     0.38118699020732211685471888558 };

   static double wHermite[4] = {
     0.0001996040722113676192060904525,
     0.0170779830074134754562030564364,
     0.2078023258148918795432586202857,
     0.6611470125582412910304159744958 };

  int order = 4;
  Vector<double> t(order);
  Vector<double> W(order);
  Matrix<double> Ws(order,npts);
  Matrix<double> ts(order,npts);

  io << showpos << scientific << setprecision(15);
  io << "// a=" << xmin << " b=" << xmax << endl;

  Vector<double> xpts(npts);
  int n1 = npts/3;
  int n2 = npts*2/3;
  double x1 = xmin + 0.1*(xmax-xmin);
  double x2 = x1 + 0.25*(xmax-x1);
  for( int i=0; i<n1; i++ ) {
    xpts[i] = xmin + i*(x1-xmin)/n1;
  }
  for( int i=n1; i<n2; i++ ) {
    xpts[i] = x1 + (i-n1)*(x2-x1)/(n2-n1);
  }
  for( int i=n2; i<npts; i++ ) {
    xpts[i] = x2 + (i-n2)*(xmax-x2)/(npts - n2 -1);
  }


  for( int i=0; i<npts; i++ ) {
    main2( order, xpts[i], t, W );
    for( int j=0; j<order; j++ ) {
      Ws[j][i] = W[j];
      ts[j][i] = t[j];
    }
  }

  int nk = kmax+1-kmin;
  Matrix<double> A(npts,nk);
  Vector<double> b(npts);
  Vector<int> Pl(npts);
  Vector<int> Pr(nk);
  Vector<double> diag(nk);

  for( int wt=0; wt<2; wt++ ) {
    char c = wt ? 'W' : 't';
    for( int o=0; o<order; o++ ) {

      for( int i=0; i<npts; i++ ) {
        for( int j=0; j<nk; j++ ) A[i][j] = 0;
        b[i] = 0;
      }

      for( int i=0; i<npts; i++ ) {
        double x = xpts[i];
        double xs = x/xmax; // use scaled variable
        double dat = wt ? Ws[o][i] : ts[o][i];
        if( wt ) {
          dat -= wHermite[o]/sqrt(x);
        } else {
          double rsq = rHermite[o]; rsq *= rsq;
          dat -= rsq/x;
        }
        b[i] = dat;
        double xk = pow(xs,(double)kmin);
        for( int j=0; j<nk; j++ ) {
          // weight by exp(-x)
          A[i][j] = xk*exp(-x);
          // A[i][j] = xk;
          xk *= xs;
        }
      }

      Householder( A, Pl, Pr, diag, POWELLREID );
      solveQR( A, Pl, Pr, diag, b );

      io << "static double " << c << noshowpos << o << "coef[" << nk
         << showpos << "]={" << endl;
      for( int i=0; i<nk; i++ ) {
        io << b[i];
        if( i != nk-1 ) io << ", ";
        if( i%3 == 2 && i != nk-1 ) io << endl;
      }
      io << "};" << endl;
    }
  }
}

void fit()
{
 
  io.open("rysfit.H", fstream::out );

  // fit to Laurent series 20,50
  LaurentFit( 20, 50, 150, -3, 7 );
  // fit to Chebyshev in range 0,1
  Chebyfit( 0.0, 1.0, 32, "s1" );
  // fit to Chebyshev in range 1,5
  Chebyfit( 1.0, 5.0, 32, "s2" );
  // fit to Chebyshev in range 5,10
  Chebyfit( 5.0, 10., 32, "s3" );
  // fit to Chebyshev in range 10,15
  Chebyfit( 10., 15., 32, "s4" );
  // fit to Chebyshev in range 15,20
  Chebyfit( 15., 20., 32, "s5" );

  io.close();
}

#ifdef TEST
void test()
{
#include "rysfit.H"

  fstream errplot("rplot", fstream::out );

  static double rHermite[4] = {
     2.93063742025724401922350270524,
     1.98165675669584292585463063976,
     1.15719371244678019472076577906,
     0.38118699020732211685471888558 };

   static double wHermite[4] = {
     0.0001996040722113676192060904525,
     0.0170779830074134754562030564364,
     0.2078023258148918795432586202857,
     0.6611470125582412910304159744958 };

  int order = 4;
  double *tp[4], *wp[4];
  Vector<double> u(order), W(order);
  for( int i=0; i<200; i++ ) {
    double x = 0.5*i;
    main2( order, x, u, W );
    errplot << x;
    double m, hw;
    if( x <= 1 ) {
      m=0.5; hw=0.5;
      tp[0] = s1t0; tp[1] = s1t1; tp[2] = s1t2; tp[3] = s1t3;
      wp[0] = s1W0; wp[1] = s1W1; wp[2] = s1W2; wp[3] = s1W3;
    } else if( x <= 5 ) {
      m=3; hw=2;
      tp[0] = s2t0; tp[1] = s2t1; tp[2] = s2t2; tp[3] = s2t3;
      wp[0] = s2W0; wp[1] = s2W1; wp[2] = s2W2; wp[3] = s2W3;
    } else if( x <= 10 ) {
      m = 7.5; hw=2.5;
      tp[0] = s3t0; tp[1] = s3t1; tp[2] = s3t2; tp[3] = s3t3;
      wp[0] = s3W0; wp[1] = s3W1; wp[2] = s3W2; wp[3] = s3W3;
    } else if( x <= 15 ) {
      m = 12.5; hw=2.5;
      tp[0] = s4t0; tp[1] = s4t1; tp[2] = s4t2; tp[3] = s4t3;
      wp[0] = s4W0; wp[1] = s4W1; wp[2] = s4W2; wp[3] = s4W3;
    } else if( x <= 20 ) {
      m = 17.5; hw=2.5;
      tp[0] = s5t0; tp[1] = s5t1; tp[2] = s5t2; tp[3] = s5t3;
      wp[0] = s5W0; wp[1] = s5W1; wp[2] = s5W2; wp[3] = s5W3;
    }
    if( x <= 20 ) {
      int nc=20;
      // Clenshaw algorithm
      Vector<double> y(nc+2);
      y[nc+1] = y[nc] = 0;
      // put in range -1,1
      double xm = (x-m)/hw;
      for( int o=0; o<order; o++ ) {
        for( int k=nc-1; k>=0; k-- ) {
  	  int f = (k==0)?1:2;
          y[k] = f*xm*y[k+1] - y[k+2] + tp[o][k];
        }
        errplot << " " << y[0]-u[o] << " " << u[o];
      }
      for( int o=0; o<order; o++ ) {
        for( int k=nc-1; k>=0; k-- ) {
  	  int f = (k==0)?1:2;
          y[k] = f*xm*y[k+1] - y[k+2] + wp[o][k];
        }
        errplot << " " << y[0]-W[o] << " " << W[o];
      }
      errplot << endl;
    } else { // a Laurent fit
      tp[0] = t0coef; tp[1] = t1coef; tp[2] = t2coef; tp[3] = t3coef;
      wp[0] = W0coef; wp[1] = W1coef; wp[2] = W2coef; wp[3] = W3coef;
      double xmax = 50.;
      int kmin = -3;
      int nk = 11;
      for( int o=0; o<order; o++ ) {
        double q=0;
        for( int j=nk-1; j>=0; j-- ) {
          q *= x/xmax;
          q += tp[o][j];
        }
        q *= pow(x/xmax,(double)kmin);
        double rsq = rHermite[o]; rsq *= rsq;
        double d = rsq/x + exp(-x)*q;
        errplot << " " << d-u[o] << " " << u[o];
      }
      for( int o=0; o<order; o++ ) {
        double q=0;
        for( int j=nk-1; j>=0; j-- ) {
          q *= x/xmax;
          q += wp[o][j];
        }
        q *= pow(x/xmax,(double)kmin);
        double d = wHermite[o]/sqrt(x) + exp(-x)*q;
        errplot << " " << d-W[o] << " " << W[o];
      }
      errplot << endl;
    }
  }
  errplot.close();
}
#endif

int main()
{
  fit();
#ifdef TEST
  test();
#endif
}
