// G. Miller, Numerical Analysis for Engineers and Scientists,
// Cambridge University Press, Cambridge, 2014.
// http://www.cambridge.org/9781107021082

#include <iostream>
#include <fstream>
#include <iomanip>
#include <math.h>
#include "Matrix.H"
#include "QR.H"
#include "AdaptiveSimpson.H"
#include "Householder.H"
#include "HQsort.H"
using namespace std;

fstream io;

// Stieltjes values are global for benefit of integrator
Vector<double> s_delta;
Vector<double> s_gammasq;

// more global info for integrator
int    s_pwr;
bool   s_extra_x;
double g_alpha;

// evaluate polynomial degree "deg" with Stieltjes recursion
double polyfn( double x, int deg )
{
  double pm2=0, pm1=0, p=1;
  for( int i=1; i<=deg; i++ ) {
    pm2 = pm1;
    pm1 = p;
    p *= x-s_delta[i-1];
    if( i!=1 ) p -= s_gammasq[i-2]*pm2;
  }
  return p;
}

// the function to be evaluated for Stieltjes form
double orth_rys_f( double x )
{
  double pi = polyfn(x*x,s_pwr);
  double retval = pi*pi*exp(-g_alpha*x*x);
  if( s_extra_x ) retval *= x*x;
  return retval;
}

void coefficients_from_Stieltjes( int maxorder, double& p0p0 )
{
  s_delta.resize(maxorder);       // in case
  s_gammasq.resize(maxorder-1);

  double oldIpipi;
  for( int i=0; i<maxorder; i++ ) {
    s_pwr   = i;      // global variable passes degree to orth_rys_f
    s_extra_x = false;
    double Ipipi = adaptiveSimpson( 0., 1., 1.e-15, orth_rys_f );
    s_extra_x = true;
    double Ixpipi = adaptiveSimpson( 0., 1., 1.e-15, orth_rys_f );

    if( i == 0 ) p0p0 = Ipipi;
    s_delta[i] = Ixpipi/Ipipi;
    if( i != 0 ) s_gammasq[i-1] = Ipipi/oldIpipi;
    oldIpipi = Ipipi;
  }
}

// global power for benefit of moment integrator
int f_pwr;

// integrand for moment calculation
void calcRysData( const int& order, const double& a_alpha,
                  Vector<double>& x, Vector<double>& W )
{
  g_alpha = a_alpha;
  double p0p0;
  Vector<double> delta(order), gamma(order-1);

  coefficients_from_Stieltjes( order, p0p0 );
  // copy from global array
  for( int i=0; i<order-1; i++ ) {
    delta[i] = s_delta[i];
    gamma[i] = sqrt(s_gammasq[i]);
  }
  delta[order-1] = s_delta[order-1];

  // J
  Matrix<double> J(order,order);
  for( int i=0; i<order; i++ ) {
    for( int j=0; j<order; j++ ) J[i][j] = 0;
    J[i][i] = delta[i];
  }
  for( int i=0; i<order-1; i++ )
    J[i][i+1] = J[i+1][i] = gamma[i];

  // eigenvalues and vectors
  Matrix<double> V(order,order);
  QReig( J, V, x );

  // weights
  for( int i=0; i<order; i++ ) {
    double len = 0;
    for( int j=0; j<order; j++ ) len += V[j][i]*V[j][i];
    W[i] = (p0p0/len)*V[0][i]*V[0][i];
  }

  // sort
  Vector<int> perm(order);
  Vector<double> wtmp(order);
  hqsort0( x, perm );
  for( int i=0; i<order; i++ ) wtmp[i] = W[perm[i]];
  for( int i=0; i<order; i++ ) W[i] = wtmp[i];
}

void Chebyfit( const double& a, const double& b, int npts, const string& prefix )
{
  io << showpos << scientific << setprecision(15);
  io << "// a=" << a << " b=" << b << endl;
  int order = 4;
  Vector<double> t(order), W(order), bj(npts);
  Matrix<double> Ws(order,npts), ts(order,npts);

  for( int i=0; i<npts; i++ ) {
    double xk = cos(M_PI*(2.*i+1)/(2.*npts));
    double alpha = (b+a)/2 + xk*(b-a)/2;
    calcRysData( order, alpha, t, W );
    for( int j=0; j<order; j++ ) {
      Ws[j][i] = W[j];
      ts[j][i] = t[j];
    }
  }
  for( int wt=0; wt<2; wt++ ) {
    char c = wt ? 'W' : 't';
    for( int o=0; o<order; o++ ) {
      for( int j=0; j<npts; j++ ) bj[j] = 0;
      for( int k=0; k<npts; k++ ) {
        double xk = cos(M_PI*(2.*k+1)/(2.*npts));
        double Tjold=0, Tj=1;
        double val = (wt ? Ws[o][k] : ts[o][k] );
        for( int j=0; j<npts; j++ ) {
          bj[j] += Tj*val;
          if( j==0 ) {
            Tj = xk;
            Tjold = 1;
          } else {
            double foo = Tj;
            Tj = 2*xk*Tj - Tjold;
            Tjold = foo;
          }
        }
      }
      for( int j=0; j<npts; j++ )
        bj[j] *= ((j==0)?1.:2.)/npts;
        io << "static double " << prefix << c << noshowpos << o
           << "[" << npts << showpos << "]={" << endl;
      for( int i=0; i<npts; i++ ) {
        io << bj[i];
        if( i != npts-1 ) io << ", ";
        if( i%3 == 2 && i != npts-1 ) io << endl;
      }
      io << "};" << endl;
    }
  }
}

void LaurentFit( const double& xmin, const double& xmax,
                 int npts, int kmin, int kmax )
{
  static double rHermite[4] = {
     2.93063742025724401922350270524, 1.98165675669584292585463063976,
     1.15719371244678019472076577906, 0.38118699020732211685471888558 };

   static double wHermite[4] = {
     0.0001996040722113676192060904525, 0.0170779830074134754562030564364,
     0.2078023258148918795432586202857, 0.6611470125582412910304159744958 };

  int order = 4;
  int nk = kmax+1-kmin;
  Vector<double> t(order), W(order), xpts(npts), b(npts), diag(nk);
  Matrix<double> Ws(order,npts), ts(order,npts), A(npts,nk);
  Vector<int> Pl(npts), Pr(nk);

  io << showpos << scientific << setprecision(15);
  io << "// a=" << xmin << " b=" << xmax << endl;

  int n1 = npts/3;
  int n2 = npts*2/3;
  double x1 = xmin + 0.1*(xmax-xmin);
  double x2 = x1 + 0.25*(xmax-x1);
  for( int i=0; i<n1; i++ ) {
    xpts[i] = xmin + i*(x1-xmin)/n1;
  }
  for( int i=n1; i<n2; i++ ) {
    xpts[i] = x1 + (i-n1)*(x2-x1)/(n2-n1);
  }
  for( int i=n2; i<npts; i++ ) {
    xpts[i] = x2 + (i-n2)*(xmax-x2)/(npts - n2 -1);
  }


  for( int i=0; i<npts; i++ ) {
    calcRysData( order, xpts[i], t, W );
    for( int j=0; j<order; j++ ) {
      Ws[j][i] = W[j];
      ts[j][i] = t[j];
    }
  }


  for( int wt=0; wt<2; wt++ ) {
    char c = wt ? 'W' : 't';
    for( int o=0; o<order; o++ ) {

      for( int i=0; i<npts; i++ ) {
        for( int j=0; j<nk; j++ ) A[i][j] = 0;
        b[i] = 0;
      }

      for( int i=0; i<npts; i++ ) {
        double x = xpts[i];
        double xs = x/xmax; // use scaled variable
        double dat = wt ? Ws[o][i] : ts[o][i];
        if( wt ) {
          dat -= wHermite[o]/sqrt(x);
        } else {
          double rsq = rHermite[o]; rsq *= rsq;
          dat -= rsq/x;
        }
        b[i] = dat;
        double xk = pow(xs,(double)kmin);
        for( int j=0; j<nk; j++ ) {
          // weight by exp(-x)
          A[i][j] = xk*exp(-x);
          // A[i][j] = xk;
          xk *= xs;
        }
      }

      Householder( A, Pl, Pr, diag, POWELLREID );
      solveQR( A, Pl, Pr, diag, b );

      io << "static double " << c << noshowpos << o << "coef[" << nk
         << showpos << "]={" << endl;
      for( int i=0; i<nk; i++ ) {
        io << b[i];
        if( i != nk-1 ) io << ", ";
        if( i%3 == 2 && i != nk-1 ) io << endl;
      }
      io << "};" << endl;
    }
  }
}

int main()
{
  io.open("rysfit.H", fstream::out );
  // fit to Laurent series 20,50
  LaurentFit( 20, 50, 150, -3, 7 );
  // fit to Chebyshev in range 0,1
  Chebyfit( 0.0, 1.0, 32, "s1" );
  // fit to Chebyshev in range 1,5
  Chebyfit( 1.0, 5.0, 32, "s2" );
  // fit to Chebyshev in range 5,10
  Chebyfit( 5.0, 10., 32, "s3" );
  // fit to Chebyshev in range 10,15
  Chebyfit( 10., 15., 32, "s4" );
  // fit to Chebyshev in range 15,20
  Chebyfit( 15., 20., 32, "s5" );
  io.close();
}
