// G. Miller, Numerical Analysis for Engineers and Scientists,
// Cambridge University Press, Cambridge, 2014.
// http://www.cambridge.org/9781107021082

#include <iostream>
#include <iomanip>
#include <math.h>
#include "dSFMT.h"
#include "Matrix.H"
using namespace std;
#ifdef USEMPI
#include <mpi.h>
int rank, nproc;
#endif
bool doio;

#ifndef MIN
#define MIN(A,B) (((A)<(B))?(A):(B))
#endif

// initialize the random number generator.  Here, dSFMT
// see http://www.math.sci.hiroshima-u.ac.jp/~m-mat/MT/emt.html

dsfmt_t internal_state;

void setup(uint32_t seed)
{
  dsfmt_init_gen_rand( &internal_state, seed );
}

// Box-Muller transform to make gaussian normal deviates
// from uniform deviates.

double normaldeviate()
{
  static int got_one = 0;
  static double extra_value;
  double v1, v2, tmp, rsq;

  if( got_one ) {
    got_one = 0;
    return extra_value;
  } else {
    do {
      // get 2 uniform deviates on [-1,1]
      // dsfmt_genrand_close_open() returns a number on [0,1)
      v1 = 2.*dsfmt_genrand_close_open(&internal_state)-1.;
      v2 = 2.*dsfmt_genrand_close_open(&internal_state)-1.;
      rsq = v1*v1 + v2*v2;
    } while( rsq==0 || rsq >= 1 );
    tmp = sqrt( - 2.*log(rsq)/rsq );
    extra_value = tmp*v2;
    got_one = 1;
    return tmp*v1;
  }
}


void makeWZ( int nstep, double& h,
             Vector<double>& DW, Vector<double>& DZ )
{
  // make npath*nstep Wiener steps DW, and consistent DZ
  double sqrth = sqrt(h);
  double h32 = h*sqrth;
  double s3    = sqrt(3.);
  double U1, U2;

  for( int n=0; n<nstep; n++ ) {
    U1 = normaldeviate();
    U2 = normaldeviate();
    DW[n] = sqrth*U1;
    DZ[n] = 0.5*h32*(U1 + U2/s3);
  }
}

void coarsenby2( int nstepfine, double& hfine,
                 Vector<double>& DW, Vector<double>& DZ )
{
  int nstepcoarse = nstepfine/2;
  int nfine = 0;
  for( int n=0; n<nstepcoarse; n++ ) {
    // the fine steps in corresponding coarse step
    double dWf1 = DW[nfine];
    double dZf1 = DZ[nfine];
    nfine++;
    double dWf2 = DW[nfine];
    double dZf2 = DZ[nfine];
    nfine++;

    // the coarse step
    DW[n] = dWf1 + dWf2;
    DZ[n] = dZf1 + dZf2 + hfine*dWf1;
  }
}

void integrate( int nstep, double& h,
                Vector<double>& DW, Vector<double>& DZ,
                double& result )
{
  // Ito-Taylor expansion of dx = f*dt + g*dW with
  // f = (x - x^3 + A*cos[2*Pi*t])/eps
  // g = sig/sqrt(eps)
  // after Berglund and Gentz 2002

  double eps = 0.0025;
  double sqe = sqrt(eps);
  double a0  = 0.005;
  double sig = 0.065;
  double x0  = 1.15303;
  double A   = 2./(3.*sqrt(3.)) - a0;


  // IC
  double x = x0;
  double t= 0;
  // integrate
  for( int n=0; n<nstep; n++ ) {
    double ac = A*cos(2.*M_PI*t);
    double dac = -2*A*M_PI*sin(2.*M_PI*t);
    x += (sig/sqe)*DW[n] + (x*(1.-x*x) + ac)*h/eps
      + 0.5*h*h*( x*((1.-3*sig*sig) + x*x*(-4.+3.*x*x))
        + ac*(1.-3.*x*x) + eps*dac )/(eps*eps)
      + (1.-3.*x*x)*sig*DZ[n]/(eps*sqe);
    t += h;
  }
  result = x;
}

void analyze( double hfine, int nstep, int nrefine, int npath, int nmom )
{
  Vector<double> DW(nstep);
  Vector<double> DZ(nstep);
  Vector<double> refres(nrefine);
  Vector<double> strong_err(nrefine-1);
  Vector<double> weak_err(nrefine-1);
  Matrix<double> weak_expect(nrefine,nmom);

  for( int p=0; p<npath; p++ ) {
    makeWZ( nstep, hfine, DW, DZ );
    double resh = hfine;
    int resn = nstep;
    for( int r=0; r<nrefine; r++ ) {
      double result;
      integrate( resn, resh, DW, DZ, result );
      refres[r] = result;
      if( r != nrefine-1 ) {
        coarsenby2( resn, resh, DW, DZ );
        resn /= 2;
        resh *= 2;
      }
    }
    // accumulate strong error
    for( int r=0; r<nrefine-1; r++ ) {
      double e = refres[r] - refres[r+1];
      strong_err[r] += e*e;
    }
    // accumulate weak expectation
    for( int r=0; r<nrefine; r++ ) {
      for( int m=1; m<=nmom; m++ ) {
        weak_expect[r][m-1] += pow( refres[r], (double) m );
      }
    }
  }

  // accumulate across processors
  long long allnpath = npath;
#ifdef USEMPI
  double *mydat = new double[nrefine];
  double *alldat = new double[nrefine];

  // sum strong errors
  for( int i=0; i<nrefine-1; i++ ) mydat[i] = strong_err[i];
  MPI_Allreduce( mydat, alldat, nrefine-1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD );
  for( int i=0; i<nrefine-1; i++ ) strong_err[i] = alldat[i];

  // weak expectations
  for( int m=1; m<=nmom; m++ ) {
    for( int i=0; i<nrefine; i++ ) mydat[i] = weak_expect[i][m-1];
    MPI_Allreduce( mydat, alldat, nrefine, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD );
    for( int i=0; i<nrefine; i++ ) weak_expect[i][m-1] = alldat[i];
  }

  // count
  long long mynpath = npath;
  MPI_Allreduce( &mynpath, &allnpath, 1, MPI_LONG_LONG, MPI_SUM, MPI_COMM_WORLD );

  delete[] alldat;
  delete[] mydat;
#endif

  // and generate convergence results
  for( int r=0; r<nrefine-1; r++ ) {
    strong_err[r] /= allnpath;
    strong_err[r] = sqrt(strong_err[r]);
    if( doio )
      cout << "strong error res=" << r << " " << strong_err[r] << endl;
  }
  if( doio ) {
    for( int r=0; r<nrefine-2; r++ ) {
      cout << "strong order " << log(strong_err[r+1]/strong_err[r])/log(2.) << endl;
    }
  }

  for( int m=1; m<=nmom; m++ ) {
    for( int r=0; r<nrefine; r++ )
      weak_expect[r][m-1] /= allnpath;
    for( int r=0; r<nrefine-1; r++ ) {
      weak_err[r] = fabs(weak_expect[r][m-1] - weak_expect[r+1][m-1]);
      if( doio )
        cout << "m=" << m << " weak error res=" << r << " " << weak_err[r] << endl;
    }
    if( doio ) {
      for( int r=0; r<nrefine-2; r++ ) {
        cout << "m=" << m << " weak order "
             << log(weak_err[r+1]/weak_err[r])/log(2.) << endl;
      }
    }
  }
}

int main(int argc, char *argv[])
{
#ifdef USEMPI
  MPI_Init(&argc,&argv);
  MPI_Comm_rank(MPI_COMM_WORLD,&rank);
  MPI_Comm_size(MPI_COMM_WORLD,&nproc);
  doio = (rank==0);
#else
  rank = 0;
  dioo = true;
#endif
  cout << scientific << setprecision(6) << endl;
  setup(rank);
  int npath = 1000000;
  int nrefine = 8;
  int nstep = 16384;
  double hfine = 1.e-3/nstep;
  int nmom = 5; // necessary to prove 2nd order
  analyze( hfine, nstep, nrefine, npath, nmom );
#ifdef USEMPI
  MPI_Finalize();
#endif
  return 0;
}
