// G. Miller, Numerical Analysis for Engineers and Scientists,
// Cambridge University Press, Cambridge, 2014.
// http://www.cambridge.org/9781107021082

#include <math.h>
#include "Matrix.H"
#include "Givens.H"

#ifndef MAX
#define MAX(A,B) (((A)>(B))?(A):(B))
#endif

// A = U D VT
//   A is m by n
//   U is m by m
//   VT is n by n

//   diag is n
//   supdiag is n-1

void makeBiDiagonal(
  Matrix<double>&A,
  Matrix<double>&U,
  Matrix<double>&VT,
  Vector<double>&diag,
  Vector<double>&supdiag
)
{
  int m = A.rows();
  int n = A.columns();
  assert( m >= n );
  Vector<double> w(m);
  int i,j,l;
  double len, dot, K;

  // begin with U and VT identity matrices
  for( i=0; i<m; i++ ) {
    for( j=0; j<m; j++ ) U[i][j] = 0;
    U[i][i] = 1;
  }
  for( i=0; i<n; i++ ) {
    for( j=0; j<n; j++ ) VT[i][j] = 0;
    VT[i][i] = 1;
  }

  for( j=0; j<n; j++ ) {

    // Householder step to replace column j with k e_1

    len = 0;
    w[j] = A[j][j];
    for( i=j+1; i<m; i++ ) {
      w[i] = A[i][j];
      len += w[i]*w[i];
    }

    if( len == 0 ) {
      // already in correct form -- just change representation
      diag[j] = A[j][j];
    } else {
      // if zero, no need to do anything

      K = sqrt( len + w[j]*w[j]);
      if( w[j] > 0 ) K = -K;

      // K is new diagonal entry.
      diag[j] = K;


      // Replace column of A with w
      w[j] -= K;

      len = 1./sqrt(len + w[j]*w[j]);
      for( i=j; i<m; i++ ) {
        w[i] *= len;
      }

      // use w on subsequent columns of A
      // a := a - 2 w (w.a)  with "a" a column vector
      for( l=j+1; l<n; l++ ) {
        dot = 0.0;
        for( i=j; i<m; i++ )
          dot += A[i][l]*w[i];
        dot *= 2;
        for( i=j; i<m; i++ )
          A[i][l] -= dot*w[i];
      }

      // use w on U, operating from right
      // u := u - 2 (u.w) w^T with "u" a row vector
      for( l=0; l<m; l++ ) {
        dot = 0.0;
        for( i=j; i<m; i++ ) 
          dot += U[l][i]*w[i];
        dot *= 2;
        for( i=j; i<m; i++ )
          U[l][i] -= dot*w[i];
      }
    }

    // Householder step to replace row i with k e_1 (shifted)

    if( j+1 < n ) {

      len = 0;
      w[j+1] = A[j][j+1];
      for( i=j+2; i<n; i++ ) {
        w[i] = A[j][i];
        len += w[i]*w[i];
      }

      if( len == 0 ) {
        // already in correct form -- just change representation
        supdiag[j] = A[j][j+1];
      } else {
        K = sqrt( len + w[j+1]*w[j+1]);
        if( w[j+1] > 0 ) K = -K;
        supdiag[j] = K;

        // Replace row of A with w
        w[j+1] -= K;
        len = 1./sqrt(len + w[j+1]*w[j+1]);

        for( i=j+1; i<n; i++ )
          w[i] *= len;

        // use w on subsequent rows of A
        // a := a - 2 (a.w) w^T with "a" a row vector
        for( l=j+1; l<m; l++ ) {
          dot = 0.0;
          for( i=j+1; i<n; i++ )
            dot += A[l][i]*w[i];
          dot *= 2;
          for( i=j+1; i<n; i++ )
            A[l][i] -= dot*w[i];
        }

        // use w on VT, operating from left
        // v := v - 2 w (w.v) with "v" a column vector
        for( l=0; l<n; l++ ) {
          dot = 0.0;
          for( i=j+1; i<n; i++ )
            dot += w[i]*VT[i][l];
          dot *= 2;
          for( i=j+1; i<n; i++ )
            VT[i][l] -= dot*w[i];
        }
      }
    }
  }
}

// if J is   d0 s0  0  0
//            0 d1 s1  0
//            0  0 d2 s2
//            0  0  0 d3
//
// then J^T J is
//      d0^2  d0*s0     0         0
//      d0*s0 d1^2+s0^2 d1*s1     0
//      0     d1*s1     d2^2+s1^2 d2*s2
//      0     0         d2*s2     d3^2+s2^2
//
// operations are designed to operate on J^T J

// return 1 if a singular value has been found.

int svdstep(
  Vector<double>&diag,
  Vector<double>&supdiag,
  Matrix<double>&U,
  Matrix<double>&VT,
  int  nd )
{
  int m = U.columns();
  int n = VT.rows();
  assert( n >= nd  );
  assert( m >= nd );
  assert( m >= n );
  if( nd == 1 ) return 1; // J is a scalar

  double shift, x1, x2, c, s, v, u1, u2, spare;
  double machine = 1e-16;
  int    j, k;

  // shift:  JT J has lower 2x2 block

  // diag[nd-2]^2 + supdiag[nd-3]^2   diag[nd-2]*supdiag[nd-2]
  // diag[nd-2]*supdiag[nd-2]         diag[nd-1]^2 + supdiag[nd-2]^2 

  int z = nd-1;
  do {
    if( z <= 0 ) break;
    z--;
    double a00 = diag[z]*diag[z];
    if( z > 0 ) a00 += supdiag[z-1]*supdiag[z-1];
    double a10 = diag[z]*supdiag[z];
    double a11 = diag[z+1]*diag[z+1]+supdiag[z]*supdiag[z];

    if( fabs(a10) < machine*(fabs(a00)+fabs(a11)) ) {
      // the matrix is reducible.  set z to be the
      // "zero" position of last partition.
      ++z;
      break;
     }
  } while(1);


  if( z > 0 && (fabs(diag[z-1]) < fabs(supdiag[z-1])) ) {
    //  JT J =   x x 0    means J is  x y 0  or   x x 0
    //           x x 0                0 x 0       0 0 Y
    //  row z -> 0 0 X                0 0 x       0 0 x

    // element "Y" must be eliminated in order that X can be
    // stripped of as the square of a singular value.
    // do this with a succession of Givens operations from left

    spare = supdiag[z-1];
    supdiag[z-1] = 0;
    for( int r=z; r<nd; r++ ) {
         // eliminate spare with diag[r]
         x1 = diag[r];
         x2 = spare;
         Givens( x1, x2, c, s, v );

         u1 = diag[r];
         u2 = spare;
         diag[r] = c*u1 + s*u2;

         if( r<nd-1 ) {
           u1 = supdiag[r];
           u2 = 0;
           supdiag[r] = c*u1 + s*u2;
           spare      = s*u1 - c*u2;
         }

         // and fix U
         for( j=0; j<m; j++ ) {
           u1 = U[j][r  ];
           u2 = U[j][z-1];
           U[j][r  ] = c*u1 + s*u2;
           U[j][z-1] = s*u1 - c*u2;
         }
      }
  }

  if( z >= nd-1 ) return 1;

  shift = 0;

  double a00 = diag[nd-2]*diag[nd-2];
  if( nd > 2 ) a00 += supdiag[nd-3]*supdiag[nd-3];
  double a10 = diag[nd-2]*supdiag[nd-2];
  double a11 = diag[nd-1]*diag[nd-1]+supdiag[nd-2]*supdiag[nd-2];

  double qa = 1;
  double qb = -(a00+a11);
  double qc = a00*a11 - a10*a10;
  double qs = ((qb<0) ? -1 : 1 );

  double qr = qb*qb - 4*qa*qc;

  // block of JT J is positive semidef, so roots should be real
  // in exact math, but ...
  qr = MAX(qr,0.);
  if( !qb && !qr ) {
    shift = 0;
  } else {
    qr = sqrt(qr);
    double mu1 = -qs*( fabs(qb)+qr )/( 2 * qa );
    double mu2 = -qs*( 2 * qc )/( fabs(qb)+qr );
    shift = mu2;
    if( fabs(a11-mu1) < fabs(a11-mu2) ) shift = mu1;
  }

  // first block of  JT J is like
  // d0^2    d0*s0
  // d0*s0   d1^2 + s0^2
  // and if this was QR with implicit shift, then (from right) Givens
  // reflection would seek to zero (d0*s0) against (d0^2 - k)

  x1 = diag[z]*diag[z] - shift;
  x2 = diag[z]*supdiag[z];
  Givens( x1, x2, c, s, v );

  // operate on row z
  u1 = diag[z];                       // old z,z
  u2 = supdiag[z];                    // old z,z+1
  diag[z] = c*u1 + s*u2;              // new z,z
  supdiag[z] = v*(u1 + diag[z]) - u2; // new z,z+1

  // operate on row 1
  u1 = 0;                             // old z+1,z
  u2 = diag[z+1];                     // old z+1,z+1
  spare = c*u1 + s*u2;                // new z+1,z
  diag[z+1] = v*(u1 + spare) - u2;    // new z+1,z+1

  // "spare" is new element z+1,z -- should be zero

  // use this Givens on VT
  for( j=0; j<n; j++ ) {
    u1 = VT[z  ][j];
    u2 = VT[z+1][j];
    VT[z  ][j] = c*u1 + s*u2;
    VT[z+1][j] = v*(u1 + VT[z][j]) - u2;
  }

  // now from left.  use element z,z to zero z+1,z (spare)
  x1 = diag[z];
  x2 = spare;
  Givens( x1, x2, c, s, v );

  // column z
  u1 = diag[z];                       // old z,z
  u2 = spare;                         // old z+1,z
  diag[z] = c*u1 + s*u2;              // new z,z

  // column z+1
  u1 = supdiag[z];                    // old z,z+1
  u2 = diag[z+1];                     // old z+1,z+1
  supdiag[z] = c*u1 + s*u2;           // new z,z+1
  diag[z+1] = v*(u1 + supdiag[z]) - u2; // new z+1,z+1

  // column 2?
  if( nd > z+2 ) {
    u1 = 0;                             // old z,z+2
    u2 = supdiag[z+1];                  // old z+1,z+2
    spare = c*u1 + s*u2;                // new z,z+2
    supdiag[z+1] = v*(u1 + spare) - u2; // new z+1,z+2

    // spare is new element z,z+2
  }

  // use this Givens on U
  for( j=0; j<m; j++ ) {
    u1 = U[j][z  ];
    u2 = U[j][z+1];
    U[j][z  ] = c*u1 + s*u2;
    U[j][z+1] = v*(u1 + U[j][z]) - u2;
  }

  for( k=z; k<nd-2; k++ ) {

    // from right.  zero spare against supdiag[k]
    x1 = supdiag[k];
    x2 = spare;
    Givens( x1, x2, c, s, v );

    // row k
    u1 = supdiag[k];                  // old k,k+1
    u2 = spare;                       // old k,k+2
    supdiag[k] = c*u1 + s*u2;         // new k,k+1

    // row k+1
    u1 = diag[k+1];                         // old k+1,k+1
    u2 = supdiag[k+1];                      // old k+1,k+2
    diag[k+1] = c*u1 + s*u2;                // new k+1,k+1
    supdiag[k+1] = v*(u1 + diag[k+1]) - u2; // new k+1,k+2

    // row k+2
    u1 = 0;                                 // old k+2,k+1
    u2 = diag[k+2];                         // old k+2,k+2
    spare = c*u1 + s*u2;                    // new k+2,k+1
    diag[k+2] = v*(u1 + spare) - u2;        // new k+2,k+2

    // use this Givens on VT
    for( j=0; j<n; j++ ) {
      u1 = VT[k+1][j];
      u2 = VT[k+2][j];
      VT[k+1][j] = c*u1 + s*u2;
      VT[k+2][j] = v*(u1 + VT[k+1][j]) - u2;
    }

    // zero spare agains diag[k+1]
    x1 = diag[k+1];
    x2 = spare;
    Givens( x1, x2, c, s, v );

    u1 = diag[k+1];                      // old k+1,k+1
    u2 = spare;                          // old k+2,k+1
    diag[k+1] = c*u1 + s*u2;             // new k+1,k+1

    u1 = supdiag[k+1];                      // old k+1,k+2
    u2 = diag[k+2];                         // old k+2,k+2
    supdiag[k+1] = c*u1 + s*u2;             // new k+1,k+2
    diag[k+2] = v*(u1 + supdiag[k+1]) - u2; // new k+2,k+2

    if( k+2 < nd-1 ) {
      u1 = 0;                               // old k+1,k+3
      u2 = supdiag[k+2];                    // old k+2,k+3
      spare = c*u1 + s*u2;                  // new k+1,k+3
      supdiag[k+2] = v*(u1 + spare) - u2;   // new k+2,k+3
    }

    // use this Givens on U
    for( j=0; j<m; j++ ) {
      u1 = U[j][k+1];
      u2 = U[j][k+2];
      U[j][k+1] = c*u1 + s*u2;
      U[j][k+2] = v*(u1 + U[j][k+1]) - u2;
    }
  }

  return 0;
}


// input:  "A" is (m x n) with m >= n (more rows than columns).
// output, U,VT, and sigma such that A = U*Diag(sigma)*VT
// with U and VT unitary and sigma positive.
// A is modified

int SVD( Matrix<double>& A,
         Matrix<double>& U,
         Matrix<double>& VT,
         Vector<double>& sigma )
{
  int i,j;
  int m = A.rows();
  int n = A.columns();
  assert( m >= n );
  Vector<double>supdiag(n-1);
  int d = sigma.size();

  // make bi-diagonal.  Store results in "sigma" and "supdiag",
  // and save tranformation matrices "U" and "VT"
  makeBiDiagonal(A,U,VT,sigma,supdiag);

  // limit on loop prevents accidentally going forever
  for( int step=0; step<10*m; step++ ) {

    int got = svdstep( sigma, supdiag, U, VT, d );

    if( got ) {
      d--;
      if( d == 0 ) break;
    }
  }

  // respect sign convention
  for( i=0; i<n; i++ ) {
    if( sigma[i] < 0 ) {
      sigma[i] *= -1;
      for( j=0; j<n; j++ ) VT[i][j] *= -1;
    }
  }

  return 1;
}
