// G. Miller, Numerical Analysis for Engineers and Scientists,
// Cambridge University Press, Cambridge, 2014.
// http://www.cambridge.org/9781107021082

#include <iostream>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
using namespace std;

#include "Matrix.H"
#include "Householder.H"

// if "MODIFY" is defined, then in Phase 2 matrix modification
// will be used instead of reduction.
#define MODIFY

#define MAX(A,B) (((A)>(B))?(A):(B))

// anything this big or smaller is close enough to zero.

// phase 2 of the simplex method.
// return 0 on success, not zero if algorithm fails.
// x_J is the solution

int phase2( const Matrix<double>& A,
            const Vector<double>& b,
            Vector<double>& x_J,
            Vector<int>& J_list,
            Vector<int>& K_list,
            const Vector<int>& is_restricted,
            int t_var,
            double simplex_small )
{
  int i, j, k, r, s, all_neg;
  double c_k;
  double r_val, x_J_over_sigma_alpha, sigma;
  int t_in_J=-1;

  int retval = 0;

  // a bunch of allocated memory
  int n_var = A.columns();
  int n_eqn = A.rows();
  Vector<double> tmp(n_eqn);
  Vector<double> d(n_eqn);
  Vector<double> a_bar(n_eqn);
  Matrix<double> AJ(n_eqn,n_eqn);
  Matrix<double> AJiAK(n_eqn,n_var-n_eqn);
  Vector<int> Pl(n_eqn), Pr(n_eqn);

  // index of t in J
  for( i=0; i<n_eqn; i++ )
    if( J_list[i] == t_var ) t_in_J = i;

  // first step, compute x_J = Inverse[AJ]*b
  for( i=0; i<n_eqn; i++ ) {
    for( j=0; j<n_eqn; j++ )
      AJ[i][j] = A[i][J_list[j]];
    x_J[i] = b[i];
  }
  retval = Householder( AJ, Pl, Pr, d, NOTHING );
  // if there's a problem, return error code.
  if( retval != 0 ) return retval;
  solveQR( AJ, Pl, Pr, d, x_J );

  // second, compute Inverse[AJ]*AK
  for( k=0; k<n_var-n_eqn; k++ ) {
    for( i=0; i<n_eqn; i++ )
      tmp[i] = A[i][K_list[k]];
    solveQR( AJ, Pl, Pr, d, tmp );
    for( i=0; i<n_eqn; i++ )
      AJiAK[i][k] = tmp[i];
  }
  
  // invalid indices prevent mod/reduction on first pass.
  r = s = -1;

  do {

    if( r >= 0 && s >= 0 ) {

      // there has been a change to J/K lists.

#ifdef MODIFY

      // Recompute Inverse(AJ)*AK and Inverse(AJ)*b using MODIFICATION

      // s is position in K where echange happened,
      // r is position in J where echange happened.

      // "m" is "s" column of AjiAk
      for( i=0; i<n_eqn; i++ ) tmp[i] = AJiAK[i][s];

      // replace "s" column of AJiAK with e_r
      for( i=0; i<n_eqn; i++ ) AJiAK[i][s] = 0;
      AJiAK[r][s] = 1;

      // multiply by M inverse:
      x_J[r] /= tmp[r];
      for( i=0; i<n_eqn; i++ )
        if( i != r ) x_J[i] -= tmp[i]*x_J[r];

      for( j=0; j<n_var-n_eqn; j++ ) {
        AJiAK[r][j] /= tmp[r];
        for( i=0; i<n_eqn; i++ )
          if( i != r ) AJiAK[i][j] -= tmp[i]*AJiAK[r][j];
      }

#else

      // Recompute Inverse(AJ)*AK and Inverse(AJ)*b using REDUCTION

      // first step, compute x_J = Inverse[AJ]*b
      for( i=0; i<n_eqn; i++ ) {
        for( j=0; j<n_eqn; j++ )
          AJ[i][j] = A[i][J_list[j]];
        x_J[i] = b[i];
      }
      retval = Householder( AJ, Pl, Pr, d, NOTHING );
      // if there's a problem, return error code.
      if( retval != 0 ) return retval;
      solveQR( AJ, Pl, Pr, d, x_J );
    
      // second, compute Inverse[AJ]*AK
      for( k=0; k<n_var-n_eqn; k++ ) {
        for( i=0; i<n_eqn; i++ )
          tmp[i] = A[i][K_list[k]];
        solveQR( AJ, Pl, Pr, d, tmp );
        for( i=0; i<n_eqn; i++ )
          AJiAK[i][k] = tmp[i];
      }

#endif

    }

    for( i=0; i<n_eqn; i++ ) {
      if( is_restricted[J_list[i]] && x_J[i] < -simplex_small ) {
        cout << "restricted variable #" << i << " " << J_list[i] <<
                " has illegal value " << x_J[i] << endl;
        retval = -100;
        return retval;
      }
    }

    // find the coefficients c_k of step 2, p. 260
    // and do test of step 3.
    s = -1;
    for( k=0; k<n_var-n_eqn; k++ ) {
      c_k = AJiAK[t_in_J][k];
      if( ( is_restricted[K_list[k]] && c_k < 0 ) || 
          ( !is_restricted[K_list[k]] && fabs(c_k) > 0) ) {
         // got a candidate s
         s = k;
         sigma = 1;
         if( c_k > 0 ) sigma = -1;
         break;
      }
    }

    if( s == -1 ) {
      break; // stop.  normal exit.
    }

    // vector a_bar, of step 4
    all_neg = 1;
    for( i=0; i<n_eqn; i++ ) {
      a_bar[i] = AJiAK[i][s];
      if( sigma*a_bar[i] > 0 && is_restricted[J_list[i]] )
        all_neg = 0;
    }

    if( all_neg ) {
      cout << "all sigma*a_bar are negative" << endl;
      retval = -10;
      break; // abnormal exit.
    }

    r = -1;
    r_val = 0;
    for( i=0; i<n_eqn; i++ ) {
      if( is_restricted[J_list[i]] ) {
        if( sigma*a_bar[i] > 0 ) {
          x_J_over_sigma_alpha = x_J[i]/(sigma*a_bar[i]);
          if( r < 0 || r_val > x_J_over_sigma_alpha ) {
            r_val = x_J_over_sigma_alpha;
            r = i;
          }
        }
      }
    }

    // swap r and s in J/K
    k = K_list[s];
    j = J_list[r];
    K_list[s] = j;
    J_list[r] = k;

  } while(1);

  return retval;
}

// phase 1 of the simplex method.
// return 0 on success, not zero if algorithm fails.
// x_J is the solution
// dimensions of all arrays may be changed.

int phase1( Matrix<double>& A,
            Vector<double>& b,
            Vector<double>& x_J,
            Vector<int>& J_list,
            Vector<int>& K_list,
            Vector<int>& is_restricted,
            int t_var,
            int t_eqn,
            double simplex_small )
{
  int i,i2,j2,j;
  int retval=0;

  // augment:
  //   rewrite objective equation
  //   add artifical variables for each eqn other than objective

  int n_eqn = A.rows();
  int n_var = A.columns();

  // first, all "b" assumed positive
  for( int i=0; i<n_eqn; i++ ) {
    if( b[i] < 0 ) {
      b[i] *= -1;
      for( int j=0; j<n_var; j++ ) A[i][j] *= -1;
    }
  }

  // spare memory to save copy of objective, and for manipulation of lists.
  Vector<double> objective(n_var);
  Vector<int>    itmp(n_var+n_eqn);

  // save the objective
  for( i=0; i<n_var; i++ ) objective[i] = A[t_eqn][i];
  double b_obj = b[t_eqn];

  // resize A to accommodate new artificial variables
  A.resize( n_eqn, n_var + n_eqn - 1 );

  // modify A to incorporate artificial variables
  j2 = n_var;
  b[t_eqn] = 0;
  for( i=0; i<n_eqn; i++ ) {
    if( i == t_eqn ) {
      // rewrite objective
      for( j=0; j<n_var; j++ ) A[i][j] = 0;
      for( j=n_var; j<n_var+n_eqn-1; j++ ) A[i][j] = 1;
      A[i][t_var] = 1;
    } else {
      b[t_eqn] += 2*b[i];
      for( j=n_var; j<n_var+n_eqn-1; j++ ) A[i][j] = 0;
      A[i][j2++] = 1;
    }
  }

  // modify restricted list
  is_restricted.resize(n_var+n_eqn-1);
  for( i=n_var; i<n_var+n_eqn-1; i++ ) is_restricted[i] = 1;

  // populate J list -- size is OK
  for( i=1; i<n_eqn; i++ )
    J_list[i] = n_var-1+i;
  J_list[0] = t_var;

  // K list
  K_list.resize( n_var - 1 );
  i2 = 0;
  for( i=0; i<n_var; i++ )
    if( i != t_var ) K_list[i2++] = i;

  // call phase 2
  retval = phase2( A, b, x_J, J_list, K_list, is_restricted, t_var, simplex_small );

  // return error code if there's a problem
  if( retval != 0 ) return retval;

  // are any augmented variables active?
  int num_aug_active = 0;
  double max_aug_value = 0;
  for( i=0; i<n_eqn; i++ ) {
     // look for indices that signify augmentation, but omit
     // augmented objective.
     if( J_list[i] >= n_var ) {
       num_aug_active++;
       max_aug_value = MAX( max_aug_value, fabs(x_J[i]) );
     }
  }

  if( num_aug_active ) {

    // we have active augmented variables.
    if( max_aug_value > simplex_small ) {
      // value not zero -- failure.
      cout << "phase 1:  augmented variables active and nonzero."
           << max_aug_value << endl;
      retval = -20;
      return retval;
    }

    // before resizing "A" etc, we need to reorder the artificial variables.
    Vector<int> newvar(num_aug_active);
    j = 0;
    for( i=0; i<n_eqn; i++ )
      if( J_list[i] >= n_var )
         newvar[j++] = J_list[i];

    // do the reording
    for( i=0; i<n_eqn; i++ )
      for( j=0; j<num_aug_active; j++ )
         A[i][n_var+j] = A[i][newvar[j]];

    // do the resize
    A.resize( n_eqn+1, n_var+num_aug_active+1 );
    b.resize( n_eqn+1 );
    x_J.resize( n_eqn+1 );

    // restore the objective
    b[t_eqn] = b_obj;
    for( i=0; i<n_var; i++ ) A[t_eqn][i] = objective[i];
    for( i=n_var; i<n_var+num_aug_active+1; i++ ) A[t_eqn][i] = 0;

    // add the new equation that "pins" artificial variables to zero
    b[n_eqn] = 0;
    for( i=0; i<n_var; i++ ) A[n_eqn][i] = 0;
    for( i=n_var; i<n_var+num_aug_active+1; i++ ) A[n_eqn][i] = 1;

    // make coefficient of new "pinning" variable zero except in last eqn
    for( i=0; i<n_eqn; i++ ) A[i][n_var+num_aug_active] = 0;

    // J list -- need to renumber the artificial variables to be
    // compatible with reordering in "A".  And, add "pinning" variable.
    J_list.resize( n_eqn+1 );
    j = n_var;
    for( i=0; i<n_eqn; i++ )
      if( J_list[i] >= n_var ) J_list[i] = j++;
    J_list[n_eqn] = n_var+num_aug_active;

    // K list -- reorder, keeping only original variables.
    j = 0;
    for( i=0; i<n_var-1; i++ )
      if( K_list[i] < n_var ) K_list[j++] = K_list[i];
    K_list.resize( (n_var+num_aug_active+1) - (n_eqn+1) );

    // and the restricted list
    is_restricted.resize( n_var+num_aug_active+1 );
    is_restricted[n_var+num_aug_active] = 1;

    return retval;
  }
    
  // there are no augmented variables in the J list.
  // they can be removed everywhere.


  // the K list
  j = 0;
  for( i=0; i<n_var-1; i++ )
    if( K_list[i] < n_var )
      K_list[j++] = K_list[i];
  K_list.resize( n_var - n_eqn );

  // restricted
  is_restricted.resize( n_var );

  // restore objective
  b[t_eqn] = b_obj;
  for( i=0; i<n_var; i++ ) A[t_eqn][i] = objective[i];
  for( i=n_var; i<n_var+num_aug_active+1; i++ ) A[t_eqn][i] = 0;

  // resize A
  A.resize( n_eqn, n_var );

  
  return retval;
}
