// G. Miller, Numerical Analysis for Engineers and Scientists,
// Cambridge University Press, Cambridge, 2014.
// http://www.cambridge.org/9781107021082

#include "Matrix.H"
#include <math.h>
#include "VMM.H"
#include "GoldenSection.H"

// leave the Hessian as a global variable to facilitate
// interrogating it after an optimization

void Psi( Matrix<double>& Hnew, const Matrix<double>& Hold,
          const Vector<double>& y, const Vector<double>& sigma,
          double beta, double gamma )
{
  int N = Hold.rows();
  Vector<double> Hy(N);
  Matrix<double> HysT(N,N);

  // H.y and
  for( int i=0; i<N; i++ ) {
    Hy[i] = 0;
    for( int j=0; j<N; j++ ) Hy[i] += Hold[i][j]*y[j];
  }

  // H y sigma^T
  for( int i=0; i<N; i++ ) {
    for( int j=0; j<N; j++ ) {
      HysT[i][j] = 0;
      for( int k=0; k<N; k++ ) HysT[i][j] += Hold[i][k]*y[k];
      HysT[i][j] *= sigma[j];
    }
  }

  // some inner products
  double yHy = 0, sigy = 0;
  for( int i=0; i<N; i++ ) {
    yHy  += y[i]*Hy[i];
    sigy += sigma[i]*y[i];
  }

  // some constants
  double c1 = (1.+gamma*beta*yHy/sigy)/sigy;
  double c2 = - gamma*(1.-beta)/yHy;
  double c3 = - gamma*beta/sigy;

  // the update
  for( int i=0; i<N; i++ ) {
    for( int j=0; j<N; j++ ) {
      Hnew[i][j] = gamma*Hold[i][j] + c1*sigma[i]*sigma[j]
                   + c2*Hy[i]*Hy[j] + c3*(HysT[j][i] + HysT[i][j]);
    }
  }
}

int VMM( double (*funcp)(const Vector<double>&),
         void   (*gradp)(const Vector<double>&,Vector<double>&),
         Vector<double>& x0,
         VMMTYPE t,
         int &iteration,
         const double& thresh,
         const double& gsthresh )
{
  int N = x0.size();
  Matrix<double> H(N,N);
  return VMM0( H, funcp, gradp, x0, t, iteration, thresh, gsthresh );
}

int VMM0( Matrix<double>& H,
          double (*funcp)(const Vector<double>&),
          void   (*gradp)(const Vector<double>&,Vector<double>&),
          Vector<double>& x0,
          VMMTYPE t,
          int &iteration,
          const double& thresh,
          const double& gsthresh )
{
  double gold = (1.+sqrt(5.))/2.;
  double beta, gamma, ep, tau, s;
  int N = x0.size();;
  Vector<double> g0(N), x1(N), g1(N), p(N), y(N), sigma(N), tmp1(N), tmp2(N);

  Matrix<double> Hi(N,N);
  for( int i=0; i<N; i++ ) {
    for( int j=0; j<N; j++ ) H[i][j] = Hi[i][j] = 0;
    H[i][i] = Hi[i][i] = 1;
  }


  iteration = 0;
  do {

    double f0 = (*funcp)(x0);
    (*gradp)(x0,g0);
    for( int i=0; i<N; i++ ) {
      p[i] = 0;
      for( int j=0; j<N; j++ ) p[i] -= H[i][j]*g0[j];
    }

    // test convergence.  df will be theoretical improvement possible.
    double df=0;
    for( int i=0; i<N; i++ ) df -= g0[i]*p[i];
    df /= 2.;
    if( df < thresh ) return 0;
    
    // bracket
    double lambdam, lambda1 = 1;
    double fm, f1;
    for( int i=0; i<N; i++ ) x1[i] = x0[i] + lambda1*p[i];
    f1 = (*funcp)( x1 );


    if( f1 < f0 ) {
      // look for lambda1,lambdam with greater values
      do {
        lambdam = lambda1;
        fm      = f1;
        lambda1 *= gold;
        for( int i=0; i<N; i++ ) x1[i] = x0[i] + lambda1*p[i];
        f1 = (*funcp)( x1 );
      } while( fm>f0 || fm>f1 );
    } else {
      // look for lambda1,lambdam with lesser values
      do {
        lambdam = lambda1/gold;
        for( int i=0; i<N; i++ ) x1[i] = x0[i] + lambdam*p[i];
        fm = (*funcp)( x1 );
        if( fm<f0 && fm<f1 ) break;
        // lambda ridiculously small is a problem
        if( lambdam < 1.e-16 ) break;
        lambda1 = lambdam;
        f1      = fm;
      } while(1);
    }

    if( lambdam < 1.e-16 ) {
      // failed to find improvement; inconsistent with df>thresh
      return -2;
    }

    double lambda = lambdam;
    fm = (*funcp)( x1 );
    double a = 0;
    goldensection( x0, p, a, lambda, lambda1, f0, fm, f1, gsthresh, funcp );

    // new x, grad, make y & sigma
    for( int i=0; i<N; i++ ) {
       sigma[i] = p[i]*lambda;
       x0[i] += sigma[i];
    }

    (*gradp)( x0, g1 );
    for( int i=0; i<N; i++ ) {
      y[i] = g1[i] - g0[i];
      g0[i]   = g1[i];
    }

    switch (t) {
    case OS:
    
      // Hy and Hiy
      for( int i=0; i<N; i++ ) {
      tmp1[i] = tmp2[i] = 0;
        for( int j=0; j<N; j++ ) {
          tmp1[i] += H[i][j]*y[j];
          tmp2[i] += Hi[i][j]*sigma[j];
        }
      }

      s=0, tau=0, ep=0;
      for( int i=0; i<N; i++ ) {
        s   += y[i]*sigma[i];
        tau += y[i]*tmp1[i];
        ep  += sigma[i]*tmp2[i];
      }

      if( (s>0 && ep<=s) || (s<0 && ep>=s) ) {
         gamma = ep/s;
         beta  = 0;
      } else if( s >= tau ) {
         gamma = s/tau;
         beta = 1;
      } else {
         gamma = 1;
         beta = s*(ep-s)/(ep*tau-s*s);
      }
      break;

    case BFGS:
      gamma = 1;
      beta  = 1;
      break;

    default: // DFP
      gamma = 1;
      beta  = 0;
    }
  
    Psi( H, H, y, sigma, beta, gamma ); // DFP

    if( t == OS ) {
      double barb = s*s*(1.-beta);
      barb = barb/(barb + beta*ep*tau);
      Psi( Hi, Hi, sigma, y, barb, 1./gamma ); // DFP
    }

    ++iteration;
  }while(iteration<1000);
  return -1;
}
