// G. Miller, Numerical Analysis for Engineers and Scientists,
// Cambridge University Press, Cambridge, 2014.
// http://www.cambridge.org/9781107021082

#include <iostream>
#include <iomanip>
#include <math.h>
#include "Matrix.H"
#include "SVD.H"
using namespace std;

template <class T> void qsort( Vector<T>& list, int left, int right )
{
  if( left < right ) {
    int    pivot = (left+right)/2;
    T      pval  = list[pivot];
    int lo=left;
    int hi=right;
    while( lo <= hi ) {
      while( list[lo] > pval ) lo++;
      while( list[hi] < pval ) hi--;
      if( lo <= hi ) {
        double d=list[lo]; list[lo]=list[hi]; list[hi]=d;
        lo++;
        hi--;
      }
    };
    if( left < hi  ) qsort<T>( list, left, hi );
    if( lo < right ) qsort<T>( list, lo, right );
  }
}

void makeVandermonde( int n, Matrix<double>& A, Vector<double>& b )
{
  for( int i=0; i<n; i++ ) {
    b[i] = 0;
    double xi = (n-1.-i)/(n-1.);
    double xij = 1;
    for( int j=0; j<n; j++ ) {
      b[i] += xij;
      A[i][j] = xij;
      xij *= xi;
    }
  }
}

int main()
{
  cout << scientific << setprecision(10);
  int n = 20;
  Matrix<double> A(n,n), Acpy(n,n), U(n,n), VT(n,n);
  Vector<double> b(n),    bcpy(n), sigma(n), orderedlist(n), UTb(n);

  // make test matrix and rhs, make a copy of rhs
  makeVandermonde(n, A, b);
  makeVandermonde(n, Acpy, bcpy);

  // perform SVD
  SVD( A, U, VT, sigma );
  for( int i=0; i<n; i++ ) orderedlist[i]=sigma[i];
  qsort<double>( orderedlist, 0, n-1 );
  cout << "singular values" << endl << orderedlist << endl;

  // min and max singular values; condition number
  cout << "condition number estimate in L2 " << orderedlist[0]/orderedlist[n-1] << endl;

  for( int k=0; k<=n; k++ ) {
    // include k singular values

    // solve U sigma VT x = b;   x = V 1/sigma UT b
    for( int i=0; i<n; i++ ) {
      UTb[i] = 0;
      for( int j=0; j<n; j++ ) UTb[i] += U[j][i]*bcpy[j];
    }
    // divide by sigma only where sigma is greater than cutoff
    for( int i=0; i<n; i++ ) {
      if( k && sigma[i] >= orderedlist[k-1] ) {
        UTb[i] /= sigma[i];
      } else {
        UTb[i] = 0;
      }
    }
    for( int i=0; i<n; i++ ) {
       b[i] = 0;
       for( int j=0; j<n; j++ ) b[i] += VT[j][i]*UTb[j];
    }
    // solution error norm, residual norm, solution norm
    double snorm=0, rnorm = 0, xnorm=0;
    for( int i=0; i<n; i++ ) {
      double r=bcpy[i];
      for( int j=0; j<n; j++ ) r -= Acpy[i][j]*b[j];
      rnorm += r*r;
      snorm += (b[i]-1.)*(b[i]-1.);
      xnorm += b[i]*b[i];
    }
    rnorm = sqrt(rnorm);
    snorm = sqrt(snorm);
    xnorm = sqrt(xnorm);
    cout << xnorm << " " << rnorm << " " << snorm << " \"" << k << "\"" <<endl;
  }
}
