// G. Miller, Numerical Analysis for Engineers and Scientists,
// Cambridge University Press, Cambridge, 2014.
// http://www.cambridge.org/9781107021082

#include <iostream>
#include <math.h>
#include "Matrix.H"
using std::cout;
using std::endl;

void prolong( Vector<double>& coarse, Vector<double>& fine )
{
  int nh  = fine.size();
  int n2h = coarse.size();
  for( int ih=0; ih<nh; ih++ ) {
     int i2h = ih/2;
     if( ih % 2 ) { // odd
       fine[ih] = coarse[i2h];
     } else {
       fine[ih] = 0;
       if( i2h <  n2h ) fine[ih] += 0.5*coarse[i2h  ];
       if( i2h >= 1   ) fine[ih] += 0.5*coarse[i2h-1];
     }
  }
}

void restrict( Vector<double>& fine, Vector<double>& coarse )
{
  int n2h = coarse.size();
  for( int i2h=0; i2h<n2h; i2h++ ) {
    int ih = 2*i2h;
    coarse[i2h] = 0.25*( fine[ih] + 2.*fine[ih+1] + fine[ih+2] );
  }
}

void smooth( int nu, const double& omega, const double& scale,
             Vector<double>& b, Vector<double>& x )
{
  // matrix is scale*( -2, 1, 0, 0, ...  ) etc
  //                 (  1,-2, 1, 0, ...  )
  //                 (  0, 1,-2, 1, ...  )

  // GS:
  // (-2  0  0 ... )          ( 0 -1       ... )
  // ( 1 -2  0 ... ) * xnew = ( 0  0 -1    ... ) *xold + b/s
  // ( 0  1 -2 ... )          ( 0  0  0 -1 ... )

  int n = x.size();

  if( omega < 0 ) {
    // assume Gauss Seidel
    for( int pass=0; pass<nu; pass++ ) {
      for( int i=0; i<n; i++ ) {
        double t = b[i]/scale;
        if( i < n-1 ) t -= x[i+1];
        if( i > 0   ) t -= x[i-1];
        x[i] = -0.5*t;
      }
    }
    return;
  } 

  // Damped Jacobi:
  // (-2/w    0    0 ...)       (2*(w-1)/w                   -1  ...)
  // (   0 -2/w    0 ...)*xnew =(       -1 2*(w-1)/w         -1  ...)*xold+b/s
  // (   0    0 -2/w ...)       (        0        -1  2*(w-1)/w  ...)


  // assume damped Jacobi
  double oldxm1;
  for( int pass=0; pass<nu; pass++ ) {
    for( int i=0; i<n; i++ ) {
      double oldx = x[i];
      double t = b[i]/scale + 2*(omega-1)*oldx/omega;
      if( i < n-1 ) t -= x[i+1];
      if( i > 0   ) t -= oldxm1;
      x[i] = -0.5*t*omega;
      oldxm1 = oldx;
    }
  }
}

double residual( const double& scale, Vector<double>& x,
                 Vector<double>& b, Vector<double> r )
{
  double r2 = 0;
  int n = x.size();
  for( int i=0; i<n; i++ ) {
    double t = b[i] + 2*scale*x[i];
    if( i > 0   ) t -= scale*x[i-1];
    if( i < n-1 ) t -= scale*x[i+1];
    r[i] = t;
    r2 += t*t;
  }
  return sqrt(r2);
}

double VCycle( int nu, const double& omega, const double& scale,
               Vector<double>&b, Vector<double>&x )
{
  int n = x.size();

  if( n == 1 ) {
    // bottom solver!
    x[0] = b[0]/(-2*scale);
    return 0;
  }

  // do nu cycles of smoothing
  smooth( nu, omega, scale, b, x );

  // residual
  Vector<double> temp1(n);
  double r2 = residual( scale, x, b, temp1 );

  // restrict residual
  int nc = ((n+1)/2)-1;
  Vector<double> temp2(nc);
  restrict( temp1, temp2 );

  // solve on coarser grid
  Vector<double> temp3(nc);
  for( int i=0; i<nc; i++ ) temp3[i] = 0;
  VCycle( nu, omega, scale/4., temp2, temp3 );

  // prolong
  prolong( temp3, temp1 );

  // correct
  for( int i=0; i<n; i++ ) x[i] += temp1[i];

  // and smooth again
  smooth( nu, omega, scale, b, x );

  r2 = residual( scale, x, b, temp1 );
  return r2;
}

int main()
{
  double goal = 1.e-4;

  int n = 512-1;
  Vector<double> b(n);
  Vector<double> x(n);
  Vector<double> r(n);
  for( int i=0; i<n; i++ ) {
    b[i] = 1;
    x[i] = 0;
  }

  cout << "MULTIGRID WITH DAMPED JACOBI" << endl;
  int iteration = 0;
  double r2;
  do {
    ++iteration;
    r2 = VCycle( 3, 2./3., 1., b, x );
    cout << "iteration " << iteration << "  resid " << r2 << endl;
  }while( r2 > goal );

  for( int i=0; i<n; i++ ) {
    b[i] = 1;
    x[i] = 0;
  }

  cout << "MULTIGRID WITH GAUSS SEIDEL " << endl;
  iteration = 0;
  do {
    ++iteration;
    r2 = VCycle( 3, -10., 1., b, x );
    cout << "iteration " << iteration << "  resid " << r2 << endl;
  }while( r2 > goal );

  for( int i=0; i<n; i++ ) {
    b[i] = 1;
    x[i] = 0;
  }

  cout << "JUST JACOBI" << endl;
  iteration = 0;
  do {
    ++iteration;
    smooth( 1, 1., 1., b, x );
    r2 = residual( 1., x, b, r );
    if( r2 <= goal ) cout << "iteration " << iteration << "  resid " << r2 << endl;
  }while( r2 > goal );

  for( int i=0; i<n; i++ ) {
    b[i] = 1;
    x[i] = 0;
  }

  cout << "JUST GAUSS SEIDEL" << endl;
  iteration = 0;
  do {
    ++iteration;
    smooth( 1, -10., 1., b, x );
    r2 = residual( 1., x, b, r );
    if( r2 <=  goal ) cout << "iteration " << iteration << "  resid " << r2 << endl;
  }while( r2 > goal );

  return 1;
}
