// G. Miller, Numerical Analysis for Engineers and Scientists,
// Cambridge University Press, Cambridge, 2014.
// http://www.cambridge.org/9781107021082

#include <iostream>
#include <iomanip>
#include <math.h>
using namespace std;

// ambient
double Pt = 101.325;

// gas constant
double R = 8.314472;      // J/K.mol

// saturation pressure parameters
double A1 = 7.16879, B1 = 1552.601, C1 = -50.731;
double A2 = 7.23255, B2 = 1750.286, C2 = -38.15;

// NRTL parapeters
double alpha12 = -0.351, alpha21 = -0.351;
double dg12 = 3159.86,   dg21 = -618.37;

void model( double T, double x1, double& y1, double& y2 )
{
  double x2 = 1. - x1;

  double tau12 = dg12/(R*T);
  double tau21 = dg21/(R*T);
  double G12 = exp(-alpha12*tau12);
  double G21 = exp(-alpha21*tau21);

  double d1 = x1 + x2*G21;
  double d2 = x2 + x1*G12;
  // activity coef ethanol
  double gamma1 = exp( x2*x2*( tau21*G21*G21/(d1*d1) + tau12*G12/(d2*d2) ) );
  // activity coef water
  double gamma2 = exp( x1*x1*( tau12*G12*G12/(d2*d2) + tau21*G21/(d1*d1) ) );

  // saturation pressure
  double Psat1 = pow(10.,A1 - B1/(T+C1));
  double Psat2 = pow(10.,A2 - B2/(T+C2));

  y1 = gamma1*x1*Psat1/Pt;
  y2 = gamma2*x2*Psat2/Pt;
}

void solve( double x1, double& T, double& y1)
{
  double T0,T1,e0,e1,e2,y2;
  double Tmin = -C1 + B1/(A1 - log10(Pt));
  double Tmax = -C2 + B2/(A2 - log10(Pt));

  T0 = Tmin;
  model( T0, x1, y1, y2 );
  e0 = y1+y2-1.;

  T1 = Tmax;
  model( T1, x1, y1, y2 );
  e1 = y1+y2-1.;

  int iter=0;
  do {
    // secant
    T = (T0*e1-T1*e0)/(e1-e0);
    model( T, x1, y1, y2 );
    e2 = y1+y2-1.;

    T0=T1;   T1=T;
    e0=e1;   e1=e2;
  }while( (fabs(e2)>1.e-12) && (++iter<1000));
}


int main()
{
  cout << scientific << setprecision(15);
  double Tmin = -C1 + B1/(A1 - log10(Pt));
  double Tmax = -C2 + B2/(A2 - log10(Pt));
  double x1, y1, T;

  // data for plot
  cout << Tmax << " " << 0. << " " << 0. << endl;
  for( int ix=1; ix<1000; ix++ ) {
    x1 = 0.001*ix;
    solve( x1, T, y1 );
    cout << T << " " << x1 << " " << y1 << endl;
  }
  cout << Tmin << " " << 1. << " " << 1. << endl;

  // find azeotrope.  error = y1-x1;
  double x10 = 0.1, x11=0.9, x12;
  double e0, e1, e2;

  solve( x10, T, y1 );
  e0 = x10-y1;

  solve( x11, T, y1 );
  e1 = x11-y1;

  int iter=0;
  do {
    // secant
    x12 = (x10*e1-x11*e0)/(e1-e0);
    solve( x12, T, y1 );
    e2 = x12-y1;

    e0=e1;   e1=e2;
    x10=x11; x11=x12;
  }while( (fabs(e2)>1.e-12) && (++iter<1000));
  cout << "azeotrope T=" << T << " x1=" << x12 << endl;
}
