// G. Miller, Numerical Analysis for Engineers and Scientists,
// Cambridge University Press, Cambridge, 2014.
// http://www.cambridge.org/9781107021082

#include <math.h>
#include <iostream>
#include <iomanip>
#include "Matrix.H"
using namespace std;

// global parameters of problem
double omegasq;

// 4th order Runge Kutta

void rk4( double h, double x, Vector<double>& y,
        void (*f)(double, Vector<double>&, Vector<double>&) )
{
  // make static to avoid re-allocating each call
  static Vector<double> k1, k2, k3, k4, tmp;

  int n = y.size();
  if( k1.size() != n) {
    k1.resize( n );
    k2.resize( n );
    k3.resize( n );
    k4.resize( n );
    tmp.resize( n );
  }


  (*f)(x,y,k1);
  for( int i=0; i<n; i++ ) {
    k1[i] *= h;
    tmp[i] = y[i] + k1[i]/2.;
  }

  (*f)(x+h/2.,tmp,k2);
  for( int i=0; i<n; i++ ) {
    k2[i] *= h;
    tmp[i] = y[i] + k2[i]/2.;
  }

  (*f)(x+h/2.,tmp,k3);
  for( int i=0; i<n; i++ ) {
    k3[i] *= h;
    tmp[i] = y[i] + k3[i];
  }

  (*f)(x+h,tmp,k4);
  for( int i=0; i<n; i++ ) {
    k4[i] *= h;
    y[i] += (k1[i]+2.*(k2[i]+k3[i])+k4[i])/6.;
  }
}

// the derivative function called by rk4
// problem-specific
void deriv( double x, Vector<double>&y, Vector<double>&dy )
{
  extern double omegasq;
  double c = 3 + x*x;
  double csq = c*c;

  dy[0] = y[1];
  dy[1] = -omegasq*y[0]/csq;
  dy[2] = y[3];
  dy[3] = -(y[0]+omegasq*y[2])/csq;
}

int main()
{
  cout << scientific << setprecision(15);

  extern double omegasq;
  Vector<double> y(4);
  double x;

  int N = 100;
  double h = 1./N;

  omegasq = 100;
  int iterations =0;
  do {
    // initial value
    y[0] = 0;
    y[1] = 1;
    y[2] = 0;
    y[3] = 0;
    x = 0;

    // integrate to x=1
    for( int i=0; i<N; i++ ) {
      rk4( h, x,  y, deriv );
      x += h;
    }

    // Newton-Raphson adjustment of omegasq
    omegasq -= y[0]/y[2];
    ++iterations;

    cout << "after iteration " << iterations << " omega = " << sqrt(omegasq) << endl;
  }while(fabs(y[0])>1.e-15 && iterations<10 );
  if( iterations >= 10 ) {
    cout << "shooting method failed to converge" << endl;
  }
  cout << "omega " << sqrt(omegasq) << endl;
}
