// G. Miller, Numerical Analysis for Engineers and Scientists,
// Cambridge University Press, Cambridge, 2014.
// http://www.cambridge.org/9781107021082

#include <Matrix.H>
#include <iostream>
#include <iomanip>
#include <math.h>
using namespace std;

// compute p = Ax without storing A
void Ax( Vector<double>& p, const Vector<double>& x, int N, double& hsq )
{
  int nv = x.size();
  for( int k=0; k<nv; k++ ) {
    int j = k/(N-1) + 1;
    int i = k%(N-1) + 1;
    p[k] = -4.*x[k]/hsq;
    if( i!=1   ) p[k] += x[k-1]/hsq;
    if( i!=N-1 ) p[k] += x[k+1]/hsq;
    if( j!=1   ) p[k] += x[k-(N-1)]/hsq;
    if( j!=N-1 ) p[k] += x[k+(N-1)]/hsq;
  }
}

int main()
{
  int N = 64;
  int nv = (N-1)*(N-1);
  double h = 1./N;
  double hsq = h*h;

  Vector<double> x(nv);
  Vector<double> b(nv), p(nv), r(nv), v(nv);

  // build vector b
  int k = 0;
  for( int i=1; i<N; i++ ) {
    double xc = i*h;
    for( int j=1; j<N; j++ ) {
      double yc = j*h;
      b[k] = 0;
      if( i==1 ) {
        // left edge -- include -u(0,y)/h^2 in RHS
        b[k] -= yc*yc/hsq;
      } else if( i ==N-1 ) {
        // right edge -- include -u(1,y)/h^2 in RHS
        b[k] -= (yc*yc-1.)/hsq;
      }
      if( j==1 ) {
        // bottom edge -- include -u(x,0)/h^2 in RHS
        b[k] -= xc*xc/hsq;
      } else if( j==N-1 ) {
        // top edge -- include -u(x,1)/h^2 in RHS
        b[k] -= (xc*xc-1)/hsq;
      }
      k++;
    }
  }

  // zero x to start
  for( int i=0; i<nv; i++ ) {
    x[i] = 0;
    r[i] = p[i] = b[i];
  }
  
  do {
    // evaluate 2-norm of p to decide if convergence has been achieved
    double dot = 0;
    for( int i=0; i<nv; i++ ) dot += p[i]*p[i];
    dot = sqrt(dot);

    cout << "dot " << dot << endl;
    if( dot < 0.001 ) break;

    // v = A*p
    Ax( v, p, N, hsq );

    double oldrdr=0;
    dot=0;
    for( int i=0; i<nv; i++ ) {
      oldrdr += r[i]*r[i];
      dot += p[i]*v[i];
    }

    double alpha = oldrdr/dot;

    for( int i=0; i<nv; i++ ) x[i] += alpha*p[i];

    dot = 0;
    for( int i=0; i<nv; i++ ) {
      r[i] -= alpha*v[i];
      dot += r[i]*r[i];
    }

    double gamma = - dot/oldrdr;

    for( int i=0; i<nv; i++ ) p[i] = r[i] - gamma*p[i];

  } while(1);
        
}
