// G. Miller, Numerical Analysis for Engineers and Scientists,
// Cambridge University Press, Cambridge, 2014.
// http://www.cambridge.org/9781107021082

#include <iostream>
#include <sstream>
#include <iomanip>
#include "Matrix.H"
#include "Householder.H"
#include "ConvexQP.H"
using namespace std;

int main()
{
  cout << setprecision(14);
  Matrix<double> A(4,3), Ce, Ci(3,4);
  Vector<double> b(4), diag(3), ce, ci(4);
  Vector<int> Pl(4), Pr(3);

  stringstream ss( stringstream::in | stringstream::out );
  ss << "1 1 1    1 4 2    1 1 6    1 3 4    1.985 5.990 7.970 6.982";
  ss >> A >> b;

  // 4 inequality constraints in form Ci^T x + ci >= 0
  ss.clear();
  ss << "1 0 0 0    0 1 0 1    0 0 1 -1    0 0 0 0";
  ss >> Ci >> ci;

  // least squares solution without constraints
  Householder( A, Pl, Pr, diag, POWELLREID );
  double r2 = solveQR( A, Pl, Pr, diag, b );
  // at this point, no longer overspecified.  Resize so L is square.
  A.resize(3,3);
  b.resize(3);
  cout << "initial ls solution x=" << endl << b << endl;
  cout << "with L2 residual " << r2 << endl;

  // A contains part of R, and representation of Q.  Convert A to L = R^T
  for( int i=0; i<3; i++ ) {
    for( int j=i+1; j<3; j++ ) {
      A[j][i] = A[i][j];
      A[i][j] = 0;
    }
    A[i][i] = diag[i];
  }

  // However, depending on the Householder pivoting scheme,
  // R may also include a right permutation Pr.  To keep the form of R,
  // we must permute solution "b", and constraints.

  for( int i=b.size()-1; i>=0; i-- ) {
    int k = Pr[i];
    while( k > i ) k = Pr[k];
    if( k != i ) {
      double d=b[k]; b[k]=b[i]; b[i]=d;
      for( int j=0; j<Ce.columns(); j++ ) {
        d = Ce[k][j]; Ce[k][j]=Ce[i][j]; Ce[i][j]=d;
      }
      for( int j=0; j<Ci.columns(); j++ ) {
        d = Ci[k][j]; Ci[k][j]=Ci[i][j]; Ci[i][j]=d;
      }
    }
  }

  // compute impact of constraints with Goldfarb Idnani method
  double df = 0;
  ConvexQP( A, b, Ce, ce, Ci, ci, df, 1.e-16 );

  // and, undo permutation on solution vector to express in natural order

  for( int i=b.size()-1; i>=0; i-- ) {
    int k = Pr[i];
    while( k < i ) k = Pr[k];
    if( k != i ) {
      double d=b[k]; b[k]=b[i]; b[i]=d;
    }
  }

  cout << "Convex Quadratic Solution x=" << endl << b << endl;
  cout << "with additional L2 residual " << df << endl;
}
