// G. Miller, Numerical Analysis for Engineers and Scientists,
// Cambridge University Press, Cambridge, 2014.
// http://www.cambridge.org/9781107021082

#include <iostream>
#include <iomanip>
#include "Matrix.H"
#include "Givens.H"
using namespace std;

double solve( int N )
{
  double u1, u2, c, s, v, delta, lambda=1.e6;
  int iterations;

  // matrix by row, original and triangular
  Vector<double>diag(N);
  Vector<double>subdiag(N);
  Vector<double>supdiag(N);
  Vector<double>supsupdiag(N);

  // eigenvector
  Vector<double>z(N);

  // transformation
  Vector<double>givS(N);
  Vector<double>givC(N);
  Vector<double>givV(N);

  // create representation of matrix
  double L = 1;
  double h = L/(N+1.);
  double hsq = h*h;

  for( int row=0; row<N; row++ ) {
    double x = (row+1)*h;
    double c = 3 + x*x/(L*L);
    double csq = c*c;
    double f = csq/hsq;
    diag[row] = 2.*f;
    if( row > 0 ) subdiag[row] = -f;
    if( row < N-1 ) supdiag[row] = -f;
    if( row < N-2 ) supsupdiag[row] = 0;
  }

  // find c,s Givens pairs to make triangular.
  for( int row=1; row<N; row++ ) {
    Givens( diag[row-1], subdiag[row], c, s, v );
    givC[row] = c;
    givS[row] = s;
    givV[row] = v;
    u1 = diag[row-1];
    u2 = subdiag[row];
    diag[row-1] = c*u1 + s*u2;
    subdiag[row] = 0;
    u1 = supdiag[row-1];
    u2 = diag[row];
    supdiag[row-1] = c*u1 + s*u2;
    diag[row] = v*(u1 + supdiag[row-1]) - u2;
    u1 = supsupdiag[row-1];
    u2 = supdiag[row];
    supsupdiag[row-1] = c*u1 + s*u2;
    supdiag[row] = v*(u1 + supsupdiag[row-1]) - u2;
  }

  // "random" eigenvector
  for( int row=0; row<N; row++ ) z[row] = 1/sqrt((double)N);

  // iterate until relative change in lambda is a few times machine precision
  iterations = 0;
  do {
    // z -> G*z
    for( int row=1; row<N; row++ ) {
      c = givC[row];
      s = givS[row];
      v = givV[row];
      u1 = z[row-1];
      u2 = z[row];
      z[row-1] = c*u1 + s*u2;
      z[row]   = v*(u1 + z[row-1]) - u2;
    }

    // back substitution.  assemble length for renormalization
    double len = 0;
    for( int row=N-1; row>=0; row-- ) {
      if( row<N-1 ) z[row] -= supdiag[row]*z[row+1];
      if( row<N-2 ) z[row] -= supsupdiag[row]*z[row+2];
      z[row] /= diag[row];
      len += z[row]*z[row];
    }
    // z grows by 1/lambda so len proportional to 1/lambda^2
    len = 1./sqrt(len);
    // now proportional to lambda
    delta = (len - lambda)/lambda;
    lambda = len;
    for( int row=0; row<N; row++ ) z[row] *= len;

    ++iterations;
  } while(delta < 1e-15 && iterations < 10);

  if( iterations >= 10 ) {
    cout << "inverse vector iteration failed to converge" << endl;
  }

  return lambda;
}

int main()
{
  cout << scientific << setprecision(15);

  int N = 10;
  for( int i=0; i<6; i++ )  {
    double lambda = solve(N);
    cout << "N=" << N << " omega = " << sqrt(lambda) << endl;
    N *= 10;
  }
  return 1;
}
