// G. Miller, Numerical Analysis for Engineers and Scientists,
// Cambridge University Press, Cambridge, 2014.
// http://www.cambridge.org/9781107021082

#include <Matrix.H>
#include <iostream>
#include <iomanip>
#include <math.h>
#include <stdlib.h>
using namespace std;

void prolong( const Vector<double>& coarse, Vector<double>& fine, int Nh )
{
  int nvh = fine.size();
  int N2h  = Nh/2;
  for( int kh=0; kh<nvh; kh++ ) {
    int jh = kh/(Nh-1);
    int ih = kh%(Nh-1);
    // node i2h,j2h may be fictitious
    // h:   0,2  1,2  2,2   c:  .   .    .
    //      0,1  1,1  2,1       .   0,0  .
    //      0,0  1,0  2,0       .   .    .
    int i2h = ih/2;
    int j2h = jh/2;
    int k2h = (N2h-1)*j2h + i2h;
    fine[kh] = 0;
    if( ih%2 && jh%2 ) {
      // odd in both indices -- fine lies on top of coarse
      fine[kh] = coarse[k2h];
    } else if( ih%2 ) {
      // aligned in x direction, not y
      if( j2h > 0     ) fine[kh] += 0.5*coarse[k2h-N2h+1];    // below
      if( j2h < N2h-1 ) fine[kh] += 0.5*coarse[k2h      ];    // above
    } else if( jh%2 ) {
      // aligned in y direction, not x
      if( i2h > 0     ) fine[kh] += 0.5*coarse[k2h-1];        // left
      if( i2h < N2h-1 ) fine[kh] += 0.5*coarse[k2h  ];        // right
    } else {
      // not aligned
      if( i2h > 0 ) {
        if( j2h > 0     ) fine[kh] += 0.25*coarse[k2h-N2h];   // lower left
        if( j2h < N2h-1 ) fine[kh] += 0.25*coarse[k2h-1  ];   // upper left
      }
      if( i2h < N2h-1 ) {
        if( j2h > 0     ) fine[kh] += 0.25*coarse[k2h-N2h+1]; // lower right
        if( j2h < N2h-1 ) fine[kh] += 0.25*coarse[k2h      ]; // upper right
      }
    }
  }
}

void restrict( const Vector<double> fine, Vector<double>& coarse, int N2h )
{
  int nv2h = coarse.size();
  int Nh = N2h*2;

  for( int k2h=0; k2h<nv2h; k2h++ ) {
    int j2h = k2h/(N2h-1);
    int i2h = k2h%(N2h-1);
    // this is the fine cell under the coarse cell
    int ih = i2h*2+1;
    int jh = j2h*2+1;
    int kh = (Nh-1)*jh + ih;
    // fine cell beneath coarse
    coarse[k2h] = fine[kh]/4.;
    // upper/lower left/right
    coarse[k2h] += ( fine[kh-1-(Nh-1)] + fine[kh-1+(Nh-1)] +
                     fine[kh+1-(Nh-1)] + fine[kh+1+(Nh-1)] )/16.;
    // lower, upper, left, right
    coarse[k2h] += ( fine[kh-(Nh-1)]   + fine[kh+(Nh-1)] +
                     fine[kh-1     ]   + fine[kh+1     ] )/8.;
  }
}

void smooth( int nu, const double& scale, int N,
             const Vector<double>& b, Vector<double>& x )
{
  int nv = x.size();
  // for Jacobi need a spare copy
  Vector<double> s(nv);
  for( int c=0; c<nu; c++ ) {
    for( int k=0; k<nv; k++ ) s[k] = x[k];
    for( int k=0; k<nv; k++ ) {
      x[k] = b[k]/scale;
      // -2 is (-4/(2/3))+4
      x[k] -= 2*s[k];
      int j = k/(N-1);
      int i = k%(N-1);
      if( i!=0   ) x[k] -= s[k-1];
      if( i!=N-2 ) x[k] -= s[k+1];
      if( j!=0   ) x[k] -= s[k-(N-1)];
      if( j!=N-2 ) x[k] -= s[k+(N-1)];
      // -6 is -4/(2/3))
      x[k] /= -6;
    }
  }
}

double residual( const double& scale, int N, const Vector<double>&x,
                 const Vector<double>&b, Vector<double>&r )
{
  int nv = x.size();
  double r2 = 0;
  for( int k=0; k<nv; k++ ) {
    r[k] = b[k] + 4*scale*x[k];
    int j = k/(N-1);
    int i = k%(N-1);
    if( i!=0   ) r[k] -= scale*x[k-1];
    if( i!=N-2 ) r[k] -= scale*x[k+1];
    if( j!=0   ) r[k] -= scale*x[k-(N-1)];
    if( j!=N-2 ) r[k] -= scale*x[k+(N-1)];
    r2 += r[k]*r[k];
  }
  return sqrt(r2);
}

double Vcycle( int nu, int Nh, const double& scale, Vector<double>&b,
               Vector<double>&x )
{
  int nvh = (Nh-1)*(Nh-1);
  if( Nh==2 ) {
    // one point ... trivial bottom solver
    x[0] = b[0]/(-4*scale);
    return 0;
  }

  // do nu cycles of smoothing
  smooth( nu, scale, Nh, b, x );

  // residual
  Vector<double> tmp1( nvh );
  double r2 = residual( scale, Nh, x, b, tmp1 );

  // restrict
  int N2h = Nh/2;
  int nv2h = (N2h-1)*(N2h-1);
  Vector<double> tmp2( nv2h );
  restrict( tmp1, tmp2, N2h );

  // solve on coarser grid
  Vector<double> tmp3( nv2h );
  for( int i=0; i<nv2h; i++ ) tmp3[i] = 0;
  Vcycle( nu, N2h, scale/4., tmp2, tmp3 );

  // prolong
  prolong( tmp3, tmp1, Nh );

  // correct
  for( int i=0; i<nvh; i++ ) x[i] += tmp1[i];

  // smooth again
  smooth( nu, scale, Nh, b, x );

  r2 = residual( scale, Nh, x, b, tmp1 );
  return r2;
}

// compute p = Ax without storing A
void Ax( Vector<double>& p, const Vector<double>& x, int N, double& hsq )
{
  int nv = x.size();
  for( int k=0; k<nv; k++ ) {
    int j = k/(N-1);
    int i = k%(N-1);
    p[k] = -4.*x[k]/hsq;
    if( i!=0   ) p[k] += x[k-1]/hsq;
    if( i!=N-2 ) p[k] += x[k+1]/hsq;
    if( j!=0   ) p[k] += x[k-(N-1)]/hsq;
    if( j!=N-2 ) p[k] += x[k+(N-1)]/hsq;
  }
}

int main()
{
  int N = 64;
  int nv = (N-1)*(N-1);
  double h = 1./N;
  double hsq = h*h;

  Vector<double> x(nv), q(nv), b(nv), p(nv), r(nv), v(nv);

  // build vector b
  for( int i=0; i<N-1; i++ ) {
    double xc = (i+1)*h;
    for( int j=0; j<N-1; j++ ) {
      double yc = (j+1)*h;
      int k = j*(N-1)+i;
      b[k] = 0;
      if( i==0 ) {
        // left edge -- include -u(0,y)/h^2 in RHS
        b[k] -= -yc*yc/hsq;
      } else if( i==N-2 ) {
        // right edge -- include -u(1,y)/h^2 in RHS
        b[k] -= (1.-yc*yc)/hsq;
      }
      if( j==0 ) {
        // bottom edge -- include -u(x,0)/h^2 in RHS
        b[k] -= xc*xc/hsq;
      } else if( j==N-2 ) {
        // top edge -- include -u(x,1)/h^2 in RHS
        b[k] -= (xc*xc-1)/hsq;
      }
    }
  }

  // ----------------------------------------------------------------------
  // first -- Conjugate gradients alone

  // zero x to start
  for( int i=0; i<nv; i++ ) {
    x[i] = 0;
    r[i] = p[i] = b[i];
  }
  
  int iter=0;
  do {
    iter++;

    // evaluate 2-norm of p to decide if convergence has been achieved
    double dot = 0;
    for( int i=0; i<nv; i++ ) dot += p[i]*p[i];
    dot = sqrt(dot);

    cout << iter << " CG L2(p) " << dot << endl;
    if( dot < 0.001 ) break;

    // v = A*p
    Ax( v, p, N, hsq );

    double oldrdr=0;
    dot=0;
    for( int i=0; i<nv; i++ ) {
      oldrdr += r[i]*r[i];
      dot += p[i]*v[i];
    }

    double alpha = oldrdr/dot;

    for( int i=0; i<nv; i++ ) x[i] += alpha*p[i];

    dot = 0;
    for( int i=0; i<nv; i++ ) {
      r[i] -= alpha*v[i];
      dot += r[i]*r[i];
    }

    double gamma = - dot/oldrdr;

    for( int i=0; i<nv; i++ ) p[i] = r[i] - gamma*p[i];

  } while(1);

  // ----------------------------------------------------------------------
  // second -- Multigrid alone

  for( int i=0; i<nv; i++ ) x[i] = 0;

  iter = 0;
  do {
    iter++;
    double r2 = Vcycle( 3, N, 1./hsq, b, x );
    cout << iter << " MG L2(r) " << r2 << endl;
    if( r2 < 0.001 ) break;
  } while(1);

  // ----------------------------------------------------------------------
  // third -- Multigrid preconditioned CG

  int pcc = 1; // Vcycles per preconditioner

  // zero x to start
  for( int i=0; i<nv; i++ ) {
    q[i] = x[i] = 0;
    r[i] = b[i];
  }

  // p,q from approximate multigrid
  for( int c=0; c<pcc; c++ ) Vcycle( 3, N, 1./hsq, r, q );
  for( int i=0; i<nv; i++ ) p[i] = q[i];

  iter=0;
  do {
    iter++;

    // evaluate 2-norm of p to decide if convergence has been achieved
    double dot = 0;
    for( int i=0; i<nv; i++ ) {
      dot += p[i]*p[i];
    }
    dot = sqrt(dot);
    cout << iter << " MGCG L2(p) " << dot << endl;

    if( dot < 0.001 ) break;

    // v = A*p
    Ax( v, p, N, hsq );

    double oldrdq=0;
    dot=0;
    for( int i=0; i<nv; i++ ) {
      oldrdq += r[i]*q[i];
      dot    += p[i]*v[i];
    }

    double alpha = oldrdq/dot;

    for( int i=0; i<nv; i++ ) x[i] += alpha*p[i];

    for( int i=0; i<nv; i++ ) r[i] -= alpha*v[i];

    // precondition with multigrid
    for( int i=0; i<nv; i++ ) q[i] = 0;
    for( int c=0; c<pcc; c++ ) Vcycle( 3, N, 1./hsq, r, q );

    dot = 0;
    for( int i=0; i<nv; i++ ) dot += r[i]*q[i];

    double beta = dot/oldrdq;
    for( int i=0; i<nv; i++ ) p[i] = q[i] + beta*p[i];
  } while(1);
}
