// G. Miller, Numerical Analysis for Engineers and Scientists,
// Cambridge University Press, Cambridge, 2014.
// http://www.cambridge.org/9781107021082

#include <iostream>
#include <stdlib.h>
#include <iomanip>
#include <math.h>
#include <Matrix.H>
#include <GE.H>
using namespace std;

#define MAX(A,B) (((A)>(B))?(A):(B))

// evaluate the function and return the square of its 2-norm

double func( Vector<double>& f, const Vector<double>& x, const Vector<double> params )
{
    double Rab = params[0];
    double Rbc = params[1];
    double Rac = params[2];
    double sum = x[0] + x[1] + x[2];
    f[0] = 1./Rab - 1./x[2] - 1./(x[0]+x[1]);
    f[1] = 1./Rbc - 1./x[0] - 1./(x[1]+x[2]);
    f[2] = 1./Rac - 1./x[1] - 1./(x[0]+x[2]);
    double fdf = f[0]*f[0] + f[1]*f[1] + f[2]*f[2];
    return fdf;
}

int main()
{
  cout << scientific << setprecision(15);
  Vector<double> x(3), y(3);
  Vector<double> f(3), g(3), params(3);
  Matrix<double> Df(3,3);
  Vector<int> Pl(3), Pr(3);

  // a starting value
  x[0] = x[1] = x[2] = 500.;

  double Rab = params[0] = 116.68;
  double Rbc = params[1] = 107.14;
  double Rac = params[2] = 123.81;


  double infnorm;  // infinity norm of f(x)

  int k = 0;
  do {
    Df[0][0] = 1./((x[0]+x[1])*(x[0]+x[1]));
    Df[0][1] = 1./((x[0]+x[1])*(x[0]+x[1]));
    Df[0][2] = 1./(x[2]*x[2]);

    Df[1][0] = 1./(x[0]*x[0]);
    Df[1][1] = 1./((x[1]+x[2])*(x[1]+x[2]));
    Df[1][2] = 1./((x[1]+x[2])*(x[1]+x[2]));

    Df[2][0] = 1./((x[0]+x[2])*(x[0]+x[2]));
    Df[2][1] = 1./(x[1]*x[1]);
    Df[2][2] = 1./((x[0]+x[2])*(x[0]+x[2]));


    double fdf = func( f, x, params );

    infnorm = MAX(fabs(f[0]),fabs(f[1]));
    infnorm = MAX(infnorm,fabs(f[2]));
    // stop if converged
    if( infnorm < 1.e-9 ) break;
   
    // replace f with Df^-1 f
    ge( Df, Pl, Pr, COMPLETE );
    solveLR( Df, Pl, Pr, f );

    double lambda = 1;
    do {
      // a trial solution of modified Newton-Raphson
      for( int i=0; i<3; i++ ) y[i] = x[i] - lambda*f[i];
      // if any value is negative, then we need a different lambda.
      // if all are positive, check size of f^T f
      if( y[0] > 0 && y[1] > 0 && y[2] > 0 ) {
        double fdf2 = func( g, y, params );
        if( fdf2 < fdf ) {
          // f^T f has diminished, so we have new x iterate.
          for( int i=0; i<3; i++ ) x[i] = y[i];
          break;
        }
      }
      // try a smaller lambda
      lambda /= 2;
      // if lambda is too small, we are probably not converging
      if( lambda < 1./1024 ) {
         cout << "Stop iteration because lambda too small " << lambda << endl;
         exit(1);
      }

    } while (1);
      
    ++k;
  } while( infnorm > 1.e-9 );

  cout << "after " << k << " iterations, the solution is" << endl << x << endl;
}
