// G. Miller, Numerical Analysis for Engineers and Scientists,
// Cambridge University Press, Cambridge, 2014.
// http://www.cambridge.org/9781107021082

#include <iostream>
#include <iomanip>
#include <math.h>
#include <fstream>
#include "dSFMT.h"
#include "Matrix.H"
using namespace std;

#ifndef MIN
#define MIN(A,B) (((A)<(B))?(A):(B))
#endif

// initialize the random number generator.  Here, dSFMT
// see http://www.math.sci.hiroshima-u.ac.jp/~m-mat/MT/emt.html

dsfmt_t internal_state;

void setup()
{
  uint32_t seed = 0;
  dsfmt_init_gen_rand( &internal_state, seed );
}

// Box-Muller transform to make gaussian normal deviates
// from uniform deviates.

double normaldeviate()
{
  static int got_one = 0;
  static double extra_value;
  double v1, v2, tmp, rsq;

  if( got_one ) {
    got_one = 0;
    return extra_value;
  } else {
    do {
      // get 2 uniform deviates on [-1,1]
      // dsfmt_genrand_close_open() returns a number on [0,1)
      v1 = 2.*dsfmt_genrand_close_open(&internal_state)-1.;
      v2 = 2.*dsfmt_genrand_close_open(&internal_state)-1.;
      rsq = v1*v1 + v2*v2;
    } while( rsq==0 || rsq >= 1 );
    tmp = sqrt( - 2.*log(rsq)/rsq );
    extra_value = tmp*v2;
    got_one = 1;
    return tmp*v1;
  }
}


void makeWZ( int npath, int nstep, double& h,
             Matrix<double>& DW, Matrix<double>& DZ )
{
  // make npath*nstep Wiener steps DW, and consistent DZ
  double sqrth = sqrt(h);
  double h32 = h*sqrth;
  double s3    = sqrt(3.);
  double U1, U2;

  for( int p=0; p<npath; p++ ) {
    for( int n=0; n<nstep; n++ ) {
      U1 = normaldeviate();
      U2 = normaldeviate();
      DW[p][n] = sqrth*U1;
      DZ[p][n] = 0.5*h32*(U1 + U2/s3);
    }
  }
}

void coarsenby2( int npath, int nstepfine, double& hfine,
                 Matrix<double>& DW, Matrix<double>& DZ )
{
  int nstepcoarse = nstepfine/2;
  for( int p=0; p<npath; p++ ) {
    int nfine = 0;
    for( int n=0; n<nstepcoarse; n++ ) {
      // the fine steps in corresponding coarse step
      double dWf1 = DW[p][nfine];
      double dZf1 = DZ[p][nfine];
      nfine++;
      double dWf2 = DW[p][nfine];
      double dZf2 = DZ[p][nfine];
      nfine++;

      // the coarse step
      DW[p][n] = dWf1 + dWf2;
      DZ[p][n] = dZf1 + dZf2 + hfine*dWf1;
    }
  }
}

void simulateOne( int npath, int nstep, double& h,
                  Matrix<double>& DW, Matrix<double>& DZ,
                  Vector<double>& results )
{

  // Black Scholes
  // dx = a x0 dt + b x dW, a and b constant

  double a = 4;
  double b = 1.;
  double x0 = 1;
  static int fileno = 0;

  fstream pathout;
  char name[100];
  for( int p=0; p<npath; p++ ) {
    sprintf(name,"sderef-p-%d-r-%d.output",p,fileno);
    pathout.open(name, fstream::out );


    // IC
    double x = x0;
    double t= 0;
pathout << " " << t << " " << x << endl;
    // integrate
    for( int n=0; n<nstep; n++ ) {
      x += h*a*x + x*b*DW[p][n];
      t += h;
//      if( (n+1)%1000 == 0 )
pathout << " " << t << " " << x << endl;
    }
    results[p] = x;
    pathout.close();
  }
  fileno++;
}

void simulateTwo( int npath, int nrefine, int nstep, double& h,
                  Matrix<double>& DW, Matrix<double>& DZ,
                  Matrix<double>& allresults )
{
  // first, make stochastic paths
  makeWZ( npath, nstep, h, DW, DZ );

  for( int n=0; n<nrefine; n++ ) {
    simulateOne( npath, nstep, h, DW, DZ, allresults[n] );

    if( n != nrefine-1 ) {
      coarsenby2( npath, nstep, h, DW, DZ );
      h     *= 2;
      nstep /= 2;
    }
  }
}

void strongRichardsonAnalysis( int npath, int nrefine,
                               Matrix<double>& allresults )
{
  Vector<double> er(nrefine-1);
  for( int r=0; r<nrefine-1; r++ ) {
    // compute strong error estimate given two levels of refinement, same paths.
    double e = 0;
    for( int p=0; p<npath; p++ ) {
      double e1 = allresults[r][p] - allresults[r+1][p];
      e += e1*e1;
    }
    er[r] = sqrt( e/(double(npath)) );
    cout << "strong L2 error w resolutions ("
         << r << "," << r+1 << ") " << er[r] << endl;
  }

  for( int r=0; r<nrefine-2; r++ ) {
    double o = log(er[r+1]/er[r])/log(2.);
    cout << "order (" << r << "," << r+1 << "," << r+2 << ") "
         << o << endl;
  }
}

void weakRichardsonAnalysis( int nmom, int npath, int nrefine,
                             Matrix<double>& allresults )
{
  Vector<double> er(nrefine-1);
  Vector<double> order(nrefine-2);
  for( int m=1; m<=nmom; m++ ) {
    cout << "moment = " << m << endl;
    for( int r=0; r<nrefine-1; r++ ) {
      double sf=0, sc=0;
      for( int p=0; p<npath; p++ ) {
        sf += pow( allresults[r][p],  (double)m);
        sc += pow( allresults[r+1][p],(double)m);
      }
      sf /= (double)npath;
      sc /= (double)npath;
      er[r] = fabs( sf - sc );
      cout << "weak error w resolutions ("
           << r << "," << r+1 << ") " << er[r] << endl;
    }

    for( int r=0; r<nrefine-2; r++ ) {
      double o = log(er[r+1]/er[r])/log(2.);
      cout << "order (" << r << "," << r+1 << "," << r+2 << ") "
           << o << endl;
      if( m == 1 ) {
        order[r] = o;
      } else {
        order[r] = MIN(order[r],o);
      }
    }
  }

  for( int r=0; r<nrefine-2; r++ )
    cout << "minimum order over moments ("
         << r << "," << r+1 << "," << r+2 << ") " << order[r] << endl;
}

main()
{
  cout << scientific << setprecision(6) << endl;
  int npath = 8;
  int nrefine = 10;
  int nstep = 1024;
  double h = 1./nstep;
  Matrix<double>DW(npath,nstep);
  Matrix<double>DZ(npath,nstep);
  Matrix<double>allresults(nrefine,npath);
  setup();
  simulateTwo( npath, nrefine, nstep, h, DW, DZ, allresults );
}
