// G. Miller, Numerical Analysis for Engineers and Scientists,
// Cambridge University Press, Cambridge, 2014.
// http://www.cambridge.org/9781107021082

#include <iostream>
#include <iomanip>
#include <math.h>
#include <fstream>
#include "dSFMT.h"
#include "Matrix.H"
using namespace std;

#ifndef MIN
#define MIN(A,B) (((A)<(B))?(A):(B))
#endif

// initialize the random number generator.  Here, dSFMT
// see http://www.math.sci.hiroshima-u.ac.jp/~m-mat/MT/emt.html

dsfmt_t internal_state;

void setup()
{
  uint32_t seed = 0;
  dsfmt_init_gen_rand( &internal_state, seed );
}

// Box-Muller transform to make gaussian normal deviates
// from uniform deviates.

double normaldeviate()
{
  static int got_one = 0;
  static double extra_value;
  double v1, v2, tmp, rsq;

  if( got_one ) {
    got_one = 0;
    return extra_value;
  } else {
    do {
      // get 2 uniform deviates on [-1,1]
      // dsfmt_genrand_close_open() returns a number on [0,1)
      v1 = 2.*dsfmt_genrand_close_open(&internal_state)-1.;
      v2 = 2.*dsfmt_genrand_close_open(&internal_state)-1.;
      rsq = v1*v1 + v2*v2;
    } while( rsq==0 || rsq >= 1 );
    tmp = sqrt( - 2.*log(rsq)/rsq );
    extra_value = tmp*v2;
    got_one = 1;
    return tmp*v1;
  }
}


void makeWZ( int npath, int nstep, double& h,
             Matrix<double>& DW, Matrix<double>& DZ )
{
  // make npath*nstep Wiener steps DW, and consistent DZ
  double sqrth = sqrt(h);
  double h32 = h*sqrth;
  double s3    = sqrt(3.);
  double U1, U2;

  for( int p=0; p<npath; p++ ) {
    for( int n=0; n<nstep; n++ ) {
      U1 = normaldeviate();
      U2 = normaldeviate();
      DW[p][n] = sqrth*U1;
      DZ[p][n] = 0.5*h32*(U1 + U2/s3);
    }
  }
}

void coarsenby2( int npath, int nstepfine, double& hfine,
                 Matrix<double>& DW, Matrix<double>& DZ )
{
  int nstepcoarse = nstepfine/2;
  for( int p=0; p<npath; p++ ) {
    int nfine = 0;
    for( int n=0; n<nstepcoarse; n++ ) {
      // the fine steps in corresponding coarse step
      double dWf1 = DW[p][nfine];
      double dZf1 = DZ[p][nfine];
      nfine++;
      double dWf2 = DW[p][nfine];
      double dZf2 = DZ[p][nfine];
      nfine++;

      // the coarse step
      DW[p][n] = dWf1 + dWf2;
      DZ[p][n] = dZf1 + dZf2 + hfine*dWf1;
    }
  }
}

void simulateOne( int npath, int nstep, double& h,
                  Matrix<double>& DW, Matrix<double>& DZ,
                  Vector<double>& results )
{
  // Ito-Taylor expansion of dx = f*dt + g*dW with
  // f = (x - x^3 + A*cos[2*Pi*t])/eps
  // g = sig/sqrt(eps)
  // after Berglund and Gentz 2002

  double eps = 0.0025;
  double sqe = sqrt(eps);
  double sig = 0.065;
  double x0  = 1.15303;
  double a0  = 0.005;
  double A   = 2./(3.*sqrt(3.)) - a0;

  static int fileno = 0;

  fstream pathout;
  pathout.open("2state-path.output", fstream::out );

  for( int p=0; p<npath; p++ ) {

    // IC
    double x = x0;
    double t= 0;
pathout << " " << t << " " << x << endl;
    // integrate
    for( int n=0; n<nstep; n++ ) {
      double ac = A*cos(2.*M_PI*t);
      double dac = -2*A*M_PI*sin(2.*M_PI*t);
      x += (sig/sqe)*DW[p][n] + (x*(1.-x*x) + ac)*h/eps
        + 0.5*h*h*( x*((1.-3*sig*sig) + x*x*(-4.+3.*x*x)) + ac*(1.-3.*x*x)
                    + eps*dac )/(eps*eps)
        + (1.-3.*x*x)*sig*DZ[p][n]/(eps*sqe);
      t += h;
      if( (n+1)%1000 == 0 )
pathout << " " << t << " " << x << endl;
    }
    results[p] = x;
  }
pathout.close();
}

void simulateTwo( int npath, int nrefine, int nstep, double& h,
                  Matrix<double>& DW, Matrix<double>& DZ,
                  Matrix<double>& allresults )
{
  // first, make stochastic paths
  makeWZ( npath, nstep, h, DW, DZ );

  for( int n=0; n<nrefine; n++ ) {
    simulateOne( npath, nstep, h, DW, DZ, allresults[n] );

    if( n != nrefine-1 ) {
      coarsenby2( npath, nstep, h, DW, DZ );
      h     *= 2;
      nstep /= 2;
    }
  }
}

void strongRichardsonAnalysis( int npath, int nrefine,
                               Matrix<double>& allresults )
{
  Vector<double> er(nrefine-1);
  for( int r=0; r<nrefine-1; r++ ) {
    // compute strong error estimate given two levels of refinement, same paths.
    double e = 0;
    for( int p=0; p<npath; p++ ) {
      double e1 = allresults[r][p] - allresults[r+1][p];
      e += e1*e1;
    }
    er[r] = sqrt( e/(double(npath)) );
    cout << "strong L2 error w resolutions ("
         << r << "," << r+1 << ") " << er[r] << endl;
  }

  for( int r=0; r<nrefine-2; r++ ) {
    double o = log(er[r+1]/er[r])/log(2.);
    cout << "order (" << r << "," << r+1 << "," << r+2 << ") "
         << o << endl;
  }
}

void weakRichardsonAnalysis( int nmom, int npath, int nrefine,
                             Matrix<double>& allresults )
{
  Vector<double> er(nrefine-1);
  Vector<double> order(nrefine-2);
  for( int m=1; m<=nmom; m++ ) {
    cout << "moment = " << m << endl;
    for( int r=0; r<nrefine-1; r++ ) {
      double sf=0, sc=0;
      for( int p=0; p<npath; p++ ) {
        sf += pow( allresults[r][p],  (double)m);
        sc += pow( allresults[r+1][p],(double)m);
      }
      sf /= (double)npath;
      sc /= (double)npath;
      er[r] = fabs( sf - sc );
      cout << "weak error w resolutions ("
           << r << "," << r+1 << ") " << er[r] << endl;
    }

    for( int r=0; r<nrefine-2; r++ ) {
      double o = log(er[r+1]/er[r])/log(2.);
      cout << "order (" << r << "," << r+1 << "," << r+2 << ") "
           << o << endl;
      if( m == 1 ) {
        order[r] = o;
      } else {
        order[r] = MIN(order[r],o);
      }
    }
  }

  for( int r=0; r<nrefine-2; r++ )
    cout << "minimum order over moments ("
         << r << "," << r+1 << "," << r+2 << ") " << order[r] << endl;
}

void roots( double &a, double &b, double &c, double &d,
            double& x1, double& x2, double &x3 )
{
// ax^3 + bx^2 + cx + d = 0
double q = (3.*a*c-b*b)/(9.*a*a);
double r = (9.*a*b*c-27.*a*a*d-2.*b*b*b)/(54.*a*a*a);
double del = q*q*q + r*r*r;
double rho = sqrt(-q*q*q);
double theta = acos(r/rho);
double tr = pow(rho,1./3.);
double sr = tr;
double si = sr*sin(theta/3.);
double ti = -si;
sr *= cos(theta/3.);
tr = sr;
x1 = sr+tr-b/(3.*a);
x2 = -0.5*(sr+tr) - b/(3.*a) - sqrt(3.)*(si-ti)/2.;
x3 = -0.5*(sr+tr) - b/(3.*a) + sqrt(3.)*(si-ti)/2.;
}

void curves( const double& h, int nstep )
{
  fstream str;
  str.open("2state-refcurve.data", fstream::out);

  // f = (x - x^3 + A*cos[2*Pi*t])/eps
  double a0  = 0.005;
  double A   = 2./(3.*sqrt(3.)) - a0;
  for( int n=0; n<nstep; n++ ) {
    double t = n*h;
    double a = -1.;
    double b = 0.;
    double c = 1.;
    double d = A*cos(2.*M_PI*t);
    double x1,x2,x3;
    roots( a, b, c, d, x1, x2, x3 );
    // sort
    double tmp;
    if( x2 < x1 ) {
      tmp = x2;
      x2 = x1;
      x1 = tmp;
    }
    if( x3 < x2 ) {
      if( x1 < x3 ) {
        // 1 3 2
        tmp = x2;
        x2 = x3;
        x3 = tmp;
      } else {
        // 3 1 2
        tmp = x3;
        x3 = x2;
        x2 = x1;
        x1 = tmp;
      }
    } else {
     // 1 2 3 ... noop
    }
    str << t << " " << x1 << " " << x2 << " " << x3 << endl;
  }
  str.close();
}

main()
{
  cout << scientific << setprecision(6) << endl;
  int npath = 1;
  int nrefine = 1;
  int nstep = 1000000;
  double h = 10.0/nstep;
  curves( h*1000, nstep/1000 );
  Matrix<double>DW(npath,nstep);
  Matrix<double>DZ(npath,nstep);
  Matrix<double>allresults(nrefine,npath);
  setup();
  simulateTwo( npath, nrefine, nstep, h, DW, DZ, allresults );
}
