// G. Miller, Numerical Analysis for Engineers and Scientists,
// Cambridge University Press, Cambridge, 2014.
// http://www.cambridge.org/9781107021082

//#define VERBOSE
#include <iostream>
#include <sstream>
#include <math.h>
#include "Matrix.H"
#include "SVD.H"
#include "QR.H"
#include "HQsort.H"
using namespace std;

void integrals( Vector<double>& fourcenter, 
                Matrix<double>& H1,
                Matrix<double>& S,
                Matrix<double>& nucleix,
                Vector<int>& nucleiZ,
                Matrix<double>& basisx,
                Vector<double>& basisa )
{
  double small = 1.e-9;
  int Nbasis = basisa.size();
  int N2 = (Nbasis*(Nbasis+1))/2;
  int N4 = (N2*(N2+1))/2;
  int Nnuclei = nucleiZ.size();

  // the H1 terms and overlap
  Vector<double> cij(3);
  for( int i=0; i<Nbasis; i++ ) {
    double ai = basisa[i];
    const Vector<double>& xi = basisx[i];
    for( int j=0; j<=i; j++ ) {
      double aj = basisa[j];
      const Vector<double>& xj = basisx[j];

      double asum = ai+aj;
      double aij = ai*aj/asum;
      double dxsq=0;
      for( int d=0; d<3; d++ ) {
        double tmp = xi[d]-xj[d];
        dxsq += tmp*tmp;
        cij[d] = (ai*xi[d] + aj*xj[d])/asum;
      }

      // overlap
      double olap = M_PI/asum;
      olap *= sqrt(olap)*exp(-aij*dxsq);
      S[i][j] = S[j][i] = olap;

      // KE
      H1[i][j] = H1[j][i] = olap*aij*(3 - 2*aij*dxsq);

      // nuclear
      for( int k=0; k<Nnuclei; k++ ) {
        int Z = nucleiZ[k];
        const Vector<double>& xn = nucleix[k];

        double tmp, r = 0;
        for( int d=0; d<3; d++ ) {
          tmp = xn[d] - cij[d];
          r += tmp*tmp;
        }
        tmp = asum*r/3;
        if( tmp < small ) {
          // treat erf(x)/x in limit x-> 0
          tmp = 1.-tmp;
          tmp *= 2.*sqrt(asum/M_PI);
        } else {
          r = sqrt(r);
          tmp = erf(sqrt(asum)*r)/r;
        }
        H1[i][j] -= Z*olap*tmp;
        if( i!=j ) H1[j][i] -= Z*olap*tmp;
      }
    }
  }

  // the four-center terms
  // packed in funny way to exploit symmetry
  double p3 = M_PI*M_PI*M_PI;

  int l1=0, l2=0, r1=0, r2=0, eone=0, etwo=0;

  for( int indx=0; indx<N4; indx++ ) {
    const Vector<double>& xl1 = basisx[l1];
    const Vector<double>& xl2 = basisx[l2];
    const Vector<double>& xr1 = basisx[r1];
    const Vector<double>& xr2 = basisx[r2];

    double al1 = basisa[l1];
    double al2 = basisa[l2];
    double ar1 = basisa[r1];
    double ar2 = basisa[r2];
    double a1sum = al1+ar1;
    double a2sum = al2+ar2;

    double tmp, dx11sq=0, dx22sq=0, dx12=0;
    for( int d=0; d<3; d++ ) {
      tmp = xl1[d] - xr1[d];
      dx11sq += tmp*tmp;
      tmp = xl2[d] - xr2[d];
      dx22sq += tmp*tmp;
      tmp = (al1*xl1[d] + ar1*xr1[d])/a1sum - (al2*xl2[d] + ar2*xr2[d])/a2sum;
      dx12 += tmp*tmp;
    }
    double c = a1sum*a2sum/(a1sum+a2sum);
    tmp = dx12*c/3.;
    if( tmp < small ) {
      tmp = 1.-tmp;
      tmp *= 2*sqrt(c/M_PI);
    } else {
     dx12 = sqrt(dx12);
     tmp = erf(sqrt(c)*dx12)/dx12;
    }
    tmp *= p3*exp( - al1*ar1*dx11sq/a1sum - al2*ar2*dx22sq/a2sum );
    tmp /= a1sum*a2sum*sqrt(a1sum*a2sum);
    fourcenter[indx] = tmp;

    if( eone == etwo ) {
      etwo = l2 = r2 = 0;
      if( r1 == l1 ) {
         l1++;
         r1=0;
       } else {
         r1++;
       }
       eone++;
    } else {
      if( r2 == l2 ) {
        l2++;
        r2=0;
      } else {
        r2++;
      }
      etwo++;
    }
  }
}

void orthogonalize( Matrix<double>& occOrbCoef, Matrix<double>& U, Vector<double>& sigma )
{
  int Nbasis = occOrbCoef.columns();
  int Npairs = occOrbCoef.rows();
  Matrix<double> tmp(Npairs,Nbasis);

  for( int i=0; i<Npairs; i++ ) {
    for( int j=0; j<Nbasis; j++ ) {
      tmp[i][j] = 0;
      for( int k=0; k<Nbasis; k++ )
        tmp[i][j] += sigma[j]*U[k][j]*occOrbCoef[i][k];
    }
  }

  double d;
  for( int i=0; i<Npairs; i++ ) {
    for( int j=0; j<i; j++ ) {
      d = 0;
      for( int k=0; k<Nbasis; k++ )
        d += tmp[i][k]*tmp[j][k];
      for( int k=0; k<Nbasis; k++ )
        tmp[i][k] -= d*tmp[j][k];
    }
    d = 0;
    for( int k=0; k<Nbasis; k++ )
      d += tmp[i][k]*tmp[i][k];
    d = 1./sqrt(d);
    for( int k=0; k<Nbasis; k++ )
      tmp[i][k] *= d;
  }

  for( int i=0; i<Npairs; i++ ) {
    for( int j=0; j<Nbasis; j++ ) {
      occOrbCoef[i][j] = 0;
      for( int k=0; k<Nbasis; k++ )
        occOrbCoef[i][j] += U[j][k]*tmp[i][k]/sigma[k];
    }
  }
}

void solveH2( double radius)
{
  int Nbasis  = 6;
  int Nnuclei = 2;
  int Npairs = 1;
  int N2 = (Nbasis*(Nbasis+1))/2;
  int N4 = (N2*(N2+1))/2;
  Vector<double> fourcenter(N4);
  Matrix<double> H1(Nbasis,Nbasis);
  Matrix<double> S(Nbasis,Nbasis);
  Matrix<double> nucleix(Nnuclei,3);
  Vector<int>    nucleiZ(Nnuclei);
  Matrix<double> basisx(Nbasis,3);
  Vector<double> basisa(Nbasis);

  Matrix<double> J(Nbasis,Nbasis);
  Matrix<double> K(Nbasis,Nbasis);

  double Etot=0, deltaE, deltaC;

  stringstream ss( stringstream::in | stringstream::out );
  ss << "0 0 0  1.4 0 0  1 1" << endl;
  ss >> nucleix;
  nucleix[1][0] = radius;
  ss >> nucleiZ;

  ss.clear();
  ss << "0 0 0  0 0 0  0 0 0 1.4 0 0  1.4 0 0 1.4 0 0 0.16885540 0.62391373 3.42525091 0.16885540 0.62391373 3.42525091";
  ss >> basisx;
  basisx[3][0] = basisx[4][0] = basisx[5][0] = radius;

  ss >> basisa;

  // initialize occupied Orbital coeffients
  Matrix<double> occOrbCoef(Npairs,Nbasis);
  Matrix<double> newOccOrbCoef(Npairs,Nbasis);
  ss.clear();
//  ss << "0 0 1 0 0 1";
  ss << ".059025 .18939 .195798 .059025 .18939 .195798";
  ss >> occOrbCoef;

  integrals( fourcenter, H1, S,
             nucleix, nucleiZ, basisx, basisa );

  Matrix<double> U(Nbasis,Nbasis);
  Matrix<double> VT(Nbasis,Nbasis);
  Vector<double> sigma(Nbasis);
  Matrix<double> Scpy(Nbasis,Nbasis);
  for( int i=0; i<Nbasis; i++ ) {
    for( int j=0; j<Nbasis; j++ ) Scpy[i][j] = S[i][j];
  }
  SVD( S, U, VT, sigma );
  for( int i=0; i<Nbasis; i++ ) sigma[i] = sqrt(sigma[i]);

  cout << "Lowdin test" << endl;
  for( int i=0; i<Nbasis; i++ ) {
    for( int j=0; j<Nbasis; j++ ) {
      cout << " " << U[i][j] - VT[j][i];
    }
    cout << endl;
  }

  Matrix<double> P(Nbasis,Nbasis);
  Matrix<double> F(Nbasis,Nbasis);
  Matrix<double> X(Nbasis,Nbasis);
  Vector<double> lambda(Nbasis);
  Vector<int> perm(Nbasis);

  int iteration = 0;
  do {
    ++iteration;
    cout << "iteration " << iteration << endl;

    cout << "before " << occOrbCoef << endl;
    orthogonalize( occOrbCoef, U, sigma );
    cout << "after " << occOrbCoef << endl;

    // P matrix
    for( int i=0; i<Nbasis; i++ ) {
      for( int j=0; j<=i; j++ ) {
        double  t = 0;
        for( int k=0; k<Npairs; k++ ) {
          t += occOrbCoef[k][i]*occOrbCoef[k][j];
        }
        P[i][j] = P[j][i] = t;
      }
    }

    cout << "P" << endl << P << endl;

    // initialize the Fock matrix
    for( int i=0; i<Nbasis; i++ )
      for( int j=0; j<Nbasis; j++ ) F[i][j] = H1[i][j];

    for( int i=0; i<Nbasis; i++ )
      for( int j=0; j<Nbasis; j++ ) J[i][j] = K[i][j] = 0;

    int l1=0, l2=0, r1=0, r2=0, eone=0, etwo=0;
    for( int indx=0; indx<N4; indx++ ) {
      double fc = fourcenter[indx];

      F[l1][r1] += 2*fc*P[l2][r2];
      F[l1][r2] -=   fc*P[l2][r1];

      J[l1][r1] += fc*P[l2][r2];
      K[l1][r2] += fc*P[l2][r1];

      if( l1!=r1 ) {
        F[r1][l1] += 2*fc*P[l2][r2];
        F[r1][r2] -=   fc*P[l2][l1];

        J[r1][l1] += fc*P[l2][r2];
        K[r1][r2] += fc*P[l2][l1];
      }

      if( l2!=r2 ) {
        F[l1][r1] += 2*fc*P[r2][l2];
        F[l1][l2] -=   fc*P[r2][r1];

        J[l1][r1] += fc*P[r2][l2];
        K[l1][l2] += fc*P[r2][r1];
      }

      if( l1!=r1 && l2!=r2 ) {
        F[r1][l1] += 2*fc*P[r2][l2];
        F[r1][l2] -=   fc*P[r2][l1];

        J[r1][l1] += fc*P[r2][l2];
        K[r1][l2] += fc*P[r2][l1];
      }

      if( eone != etwo ) {

        F[l2][r2] += 2*fc*P[l1][r1];
        F[l2][r1] -=   fc*P[l1][r2];

        J[l2][r2] += fc*P[l1][r1];
        K[l2][r1] += fc*P[l1][r2];

        if( l2!=r2 ) {
          F[r2][l2] += 2*fc*P[l1][r1];
          F[r2][r1] -=   fc*P[l1][l2];

          J[r2][l2] += fc*P[l1][r1];
          K[r2][r1] += fc*P[l1][l2];
        }

        if( l1!=r1 ) {
          F[l2][r2] += 2*fc*P[r1][l1];
          F[l2][l1] -=   fc*P[r1][r2];

          J[l2][r2] += fc*P[r1][l1];
          K[l2][l1] += fc*P[r1][r2];
        }

        if( l2!=r2 && l1!=r1 ) {
          F[r2][l2] += 2*fc*P[r1][l1];
          F[r2][l1] -=   fc*P[r1][l2];

          J[r2][l2] += fc*P[r1][l1];
          K[r2][l1] += fc*P[r1][l2];
        }
      }

      if( eone == etwo ) {
        etwo = l2 = r2 = 0;
        if( r1 == l1 ) {
           l1++;
           r1=0;
         } else {
           r1++;
         }
         eone++;
      } else {
        if( r2 == l2 ) {
          l2++;
          r2=0;
        } else {
          r2++;
        }
        etwo++;
      }
    }

    // evaluate total energy
    double Ee = 0, Enn = 0;
    for( int i=0; i<Nbasis; i++ ) {
      for( int j=0; j<Nbasis; j++ ) {
         Ee += P[i][j]*(F[i][j]+H1[i][j]);
      }
    }
    Ee /= 2;
    for( int i=0; i<Nnuclei; i++ ) {
      const Vector<double>& xi = nucleix[i];
      int zi = nucleiZ[i];
      for( int j=0; j<i; j++ ) {
        const Vector<double>& xj = nucleix[j];
        int zj = nucleiZ[j];
        double r = 0;
        for( int dir=0; dir<3; dir++ ) 
          r += (xi[dir]-xj[dir])*(xi[dir]-xj[dir]);
        Enn += zi*zj/sqrt(r);
      }
    }
    cout << "electron energy   " << Ee << endl;
    cout << "nuclear repulsion " << Enn << endl;
    cout << "total energy      " << Ee + Enn << endl;

    // F C = E S C
    // F C = E U sigma UT C
    // 1/sqrt(sigma) UT F U 1/sqrt(sigma) sqrt(sigma) UT C = E sqrt(sigma) UT C

    // replace VT with F U/sqrt(sigma)
    for( int i=0; i<Nbasis; i++ ) {
      for( int j=0; j<Nbasis; j++ ) {
        VT[i][j] = 0;
        for( int k=0; k<Nbasis; k++ ) VT[i][j] += F[i][k]*U[k][j]/sigma[j];
      }
    }

    // replace F with 1/sqrt(sigma) UT VT
    for( int i=0; i<Nbasis; i++ ) {
      for( int j=0; j<Nbasis; j++ ) {
        F[i][j] = 0;
        for( int k=0; k<Nbasis; k++ ) F[i][j] += U[k][i]*VT[k][j]/sigma[i];
      }
    }

    // eigenvalue
    QReig( F, X, lambda );
    cout << "lambda " << endl << lambda << endl;
    cout << "X " << endl << X << endl;

    cout << "X transformed" << endl;
    for( int i=0; i<Nbasis; i++ ) {
      for( int j=0; j<Nbasis; j++ ) {
        double t= 0;
        for( int k=0; k<Nbasis; k++ ) {
          t += U[i][k]*X[k][j]/sigma[k];
        }
        cout << " " << t;
      }
      cout << endl;
    }

    // X = sqrt(sigma) UT C
    // C = U 1/sqrt(sigma) X

    // but collect only lowest energy pairs
    cout << "lambda before" << lambda << endl;
    hqsort0( lambda, perm );

    cout << "perm " << endl << perm << endl;
    cout << "lambda after" << lambda << endl;

    double Eorb = 0;
    for( int i=0; i<Npairs; i++ ) Eorb += 2*lambda[Nbasis-1-i];
    double Ej = 0, Ek = 0;
    for( int p=0; p<Nbasis; p++ ) {
       for( int q=0; q<Nbasis; q++ ) {
          Ej  += P[p][q]*J[p][q];
          Ek  += P[p][q]*K[p][q];
       }
    }
    cout << "Eorb " << Eorb << endl;
    cout << "Enuc " << Enn  << endl;
    cout << "EJ   " << Ej  << endl;
    cout << "EK   " << Ek  << endl;
    cout << radius << " test  " << Eorb-(2*Ej-Ek) + Enn  << endl;

    double Etmp = Eorb-(2*Ej-Ek) + Enn;
    deltaE = fabs(Etmp-Etot);
    Etot = Etmp;
    
  cout << "grab column " << perm[Nbasis-1-0] << endl;

    for( int i=0; i<Npairs; i++ ) {
      for( int j=0; j<Nbasis; j++ ) {
        newOccOrbCoef[i][j] = 0;
        for( int k=0; k<Nbasis; k++ )
          newOccOrbCoef[i][j] += U[j][k]*X[k][perm[Nbasis-1-i]]/sigma[k];
      }
    }

    cout << "newOccOrbCoef" << endl << newOccOrbCoef << endl;

    for( int i=0; i<Npairs; i++ ) {
      for( int j=0; j<Npairs; j++ ) {
        double t = 0;
        for( int k=0; k<Nbasis; k++ ) {
          for( int l=0; l<Nbasis; l++ ) {
            t += newOccOrbCoef[i][k]*Scpy[k][l]*newOccOrbCoef[j][l];
          }
        }
        cout << "orb orth " << i << " " << j << " " << t << endl;
      }
    }

    // relax
    double f = 0.0;
    deltaC = 0;
    for( int i=0; i<Npairs; i++ ) {
      for( int j=0; j<Nbasis; j++ ) {
        double tmp = (1-f)*newOccOrbCoef[i][j] + f*occOrbCoef[i][j];
        double diff = occOrbCoef[i][j] - tmp;
        deltaC += diff*diff;
        occOrbCoef[i][j] = tmp;
      }
    }
    deltaC = sqrt(deltaC);
    cout << "occOrbCoef" << endl << occOrbCoef << endl;

  } while( (iteration < 2) || deltaE > 1.e-4  || deltaC > 1.e-4 );
}

int main()
{
  for( double r=0.8; r<=3.21; r+= 0.1 ) solveH2(r);
}
